@@ -16,9 +16,18 @@
 /* */
 #undef HAVE_MAXINT_CONVERSION_BUG
 
+/* */
+#undef HAVE_STDLIB_H_X
+
 /* */
 #undef HAVE_SOCKLEN_T
 
+/* */
+#undef HAVE_VOLATILE
+
+/* */
+#undef HAVE_ERRNO
+
 /* */
 #undef C_BIG_ENDIAN
 
@@ -46,21 +55,45 @@
 /* */
 #undef HAVE_STRNCMP
 
+/* */
+#undef HAVE_STRCASECMP
+
+/* */
+#undef HAVE_STRNCASECMP
+
 /* */
 #undef HAVE_STRCSPN
 
 /* */
 #undef HAVE_STRSTR
 
+/* */
+#undef HAVE_MEMCMP
+
+/* */
+#undef HAVE_MEMCHR
+
+/* */
+#undef HAVE_MEMCPY
+
 /* */
 #undef HAVE_MEMMOVE
 
+/* */
+#undef HAVE_MEMSET
+
 /* */
 #undef HAVE_STRERROR
 
 /* */
 #undef HAVE_SIGSETJMP
 
+/* */
+#undef DEBUGLEVEL
+
+/* */
+#undef HAVE_CLOCK_GETTIME
+
 /* */
 #undef HAVE_GETHOSTBYNAME
 
@@ -68,7 +101,16 @@
 #undef HAVE_GETHOSTBYNAME_BUG
 
 /* */
-#undef DEBUGLEVEL
+#undef SUPPORT_IPV6
+
+/* */
+#undef SUPPORT_IPV6_SCOPE
+
+/* */
+#undef HAVE_POW
+
+/* */
+#undef HAVE_POWF
 
 /* */
 #undef JS
@@ -94,6 +136,9 @@
 /* */
 #undef X2
 
+/* */
+#undef HAVE_XSETLOCALE
+
 /* */
 #undef HAVE_SSL
 
@@ -136,6 +181,9 @@
 /* */
 #undef GRDRV_ATHEOS
 
+/* */
+#undef GRDRV_GRX
+
 /* */
 #undef DONT_INCLUDE_SETJMP
 
@@ -144,3 +192,6 @@
 
 /* Tiff by Brain */
 #undef HAVE_TIFF
+
+/* Gpm_Event has wdx and wdy */
+#undef HAVE_WDX_WDY
