@@ -64,7 +64,7 @@ class LinksView : public View {
 	virtual void MouseUp(const Point &p, uint32 b, Message *m);
 	virtual void MouseMove(const Point &p, int c, uint32 b, Message *m);
 	virtual void KeyDown(const char *s, const char *rs, uint32 q);
-	virtual void WheelMoved(const Point &d);
+	virtual void WheelMove(const point &d);
 	LinksWindow *win;
 	struct graphics_device *dev;
 	void d_flush();
@@ -246,7 +246,7 @@ void LinksView::MouseMove(const Point &p, int c, uint32 b, Message *m)
 	write(wpipe, " ", 1);
 }
 
-void LinksView::WheelMoved(const Point &d)
+void LinksView::WheelMove(const point &d)
 {
 	win->Unlock();
 	ath_lock->Lock();
@@ -291,6 +291,11 @@ void LinksView::KeyDown(const char *s, const char *rs, uint32 q)
 	/*debug2("key: :%s: :%s: %d %d\n", s, rs, q, c);*/
 }
 
+unsigned char *ath_get_driver_param(void)
+{
+	return NULL;
+}
+
 uint32 ath_app_thread(void *p)
 {
 	ath_links_app->Run();
@@ -317,15 +322,8 @@ unsigned char *ath_init_driver(unsigned char *param, unsigned char *display)
 	}
 	fcntl(rpipe, F_SETFL, O_NONBLOCK);
 	fcntl(wpipe, F_SETFL, O_NONBLOCK);
-	set_handlers(rpipe, ath_get_event, NULL, NULL);
+	set_handlers(rpipe, ath_get_event, NULL, NULL, NULL);
 	ath_app_thread_id = spawn_thread("links_app", (void *)ath_app_thread, 0, 0, NULL);
-	if (ath_app_thread_id == -1) {
-		close(rpipe);
-		close(wpipe);
-		delete ath_lock; ath_lock = NULL;
-		delete ath_links_app;
-		return stracpy((unsigned char *)"Could not spawn thread.\n");
-	}
 	resume_thread(ath_app_thread_id);
 	if ((d = new Desktop)) {
 		ath_cs_desktop = d->GetColorSpace();
@@ -350,7 +348,7 @@ unsigned char *ath_init_driver(unsigned char *param, unsigned char *display)
 	*/
 	ath_win_y_pos = (ath_y_size - ath_y_panel - ath_win_y_size) / 2 + ath_y_panel;
 	ath_win_x_pos = ath_x_size - ath_win_x_size - ath_win_y_pos;
-	if (/*ath_cs_desktop == CS_RGB32 ||*/ ath_cs_desktop == CS_RGB24 || ath_cs_desktop == CS_RGB16 || ath_cs_desktop == CS_RGB15)
+	if (/*ath_cs_desktop == CS_RGB32 ||*/ ath_cs_desktop == CS_RGB24 || ath_cs_desktop == CS_RGB16 || ath_cs_desktop == CS_RGB15) 
 		ath_cs_bmp = ath_cs_desktop;
 	else if (ath_cs_desktop == CS_RGB32 || ath_cs_desktop == CS_RGBA32) ath_cs_bmp = CS_RGB24;
 	else ath_cs_bmp = CS_RGB15;
@@ -365,7 +363,7 @@ unsigned char *ath_init_driver(unsigned char *param, unsigned char *display)
 			atheos_driver.depth = 0x7a;
 			break;
 		default:
-			internal("unknown depth");
+			internal((unsigned char *)"unknown depth");
 	}
 	return NULL;
 }
@@ -373,9 +371,9 @@ unsigned char *ath_init_driver(unsigned char *param, unsigned char *display)
 void ath_shutdown_driver()
 {
 	debug((unsigned char *)"D");
-	set_handlers(rpipe, NULL, NULL, NULL);
 	close(rpipe);
 	close(wpipe);
+	set_handlers(rpipe, NULL, NULL, NULL, NULL);
 	ath_lock->Unlock();
 	debug((unsigned char *)"DD");
 	ath_links_app->PostMessage(M_TERMINATE);
@@ -384,12 +382,6 @@ void ath_shutdown_driver()
 	debug((unsigned char *)"F");
 }
 
-void ath_after_fork()
-{
-	close(rpipe);
-	close(wpipe);
-}
-
 struct graphics_device *ath_init_device()
 {
 	LinksView *view;
@@ -401,7 +393,7 @@ struct graphics_device *ath_init_device()
 	win = new LinksWindow(Rect(ath_win_x_pos, ath_win_y_pos, ath_win_x_pos + ath_win_x_size, ath_win_y_pos + ath_win_y_size));
 	debug((unsigned char *)"2");
 	if (!win) {
-		if (out_of_memory(0, NULL, 0))
+		if (out_of_memory(NULL, 0))
 			goto retry;
 		mem_free(dev);
 		return NULL;
@@ -410,7 +402,7 @@ struct graphics_device *ath_init_device()
 	retry2:
 	view = new LinksView(win);
 	if (!view) {
-		if (out_of_memory(0, NULL, 0))
+		if (out_of_memory(NULL, 0))
 			goto retry2;
 		delete win;
 		mem_free(dev);
@@ -460,7 +452,7 @@ int ath_get_empty_bitmap(struct bitmap *bmp)
 	retry:
 	b = new Bitmap(bmp->x, bmp->y, ath_cs_bmp, Bitmap::SHARE_FRAMEBUFFER);
 	if (!b) {
-		if (out_of_memory(0, NULL, 0))
+		if (out_of_memory(NULL, 0))
 			goto retry;
 		bmp->data = NULL;
 		bmp->flags = NULL;
@@ -509,13 +501,28 @@ void ath_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int x, int
 	debug2("drawb\n");
 	Bitmap *b = (Bitmap *)bmp->flags;
 	if (!b) return;
-	CLIP_DRAW_BITMAP
 	lock_dev(dev);
 	lv(dev)->DrawBitmap(b, b->GetBounds(), Rect(x, y, x + bmp->x - 1, y + bmp->y - 1));
 	lv(dev)->d_flush();
 	unlock_dev(dev);
 }
 
+#if 0
+void ath_draw_bitmaps(struct graphics_device *dev, struct bitmap **bmp, int n, int x, int y)
+{
+	LinksView *lvv = lv(dev);
+	lock_dev(dev);
+	while (n--) {
+		Bitmap *b = (Bitmap *)(*bmp)->flags;
+		lvv->DrawBitmap(b, b->GetBounds(), Rect(x, y, x + (*bmp)->x, y + (*bmp)->y));
+		x += (*bmp)->x;
+		bmp++;
+	}
+	lv(dev)->d_flush();
+	unlock_dev(dev);
+}
+#endif
+
 long ath_get_color(int rgb)
 {
 	if (small_color) {
@@ -527,7 +534,7 @@ long ath_get_color(int rgb)
 void ath_fill_area(struct graphics_device *dev, int x1, int y1, int x2, int y2, long color)
 {
 	debug2("fill\n");
-	CLIP_FILL_AREA
+	if (x1 >= x2 || y1 >= y2) return;
 	lock_dev(dev);
 	if (small_color)
 		lv(dev)->FillRect(Rect(x1, y1, x2 - 1, y2 - 1), *(Color32_s *)(void *)&color);
@@ -540,7 +547,7 @@ void ath_fill_area(struct graphics_device *dev, int x1, int y1, int x2, int y2,
 void ath_draw_hline(struct graphics_device *dev, int x1, int y, int x2, long color)
 {
 	debug2("hline\n");
-	CLIP_DRAW_HLINE
+	if (x1 >= x2) return;
 	lock_dev(dev);
 	if (small_color)
 		lv(dev)->SetFgColor(*(Color32_s *)(void *)&color);
@@ -554,7 +561,7 @@ void ath_draw_hline(struct graphics_device *dev, int x1, int y, int x2, long col
 void ath_draw_vline(struct graphics_device *dev, int x, int y1, int y2, long color)
 {
 	debug2("vline\n");
-	CLIP_DRAW_VLINE
+	if (y1 >= y2) return;
 	lock_dev(dev);
 	if (small_color)
 		lv(dev)->SetFgColor(*(Color32_s *)(void *)&color);
@@ -565,76 +572,79 @@ void ath_draw_vline(struct graphics_device *dev, int x, int y1, int y2, long col
 	unlock_dev(dev);
 }
 
-int ath_scroll(struct graphics_device *dev, struct rect_set **ignore, int scx, int scy)
+int ath_hscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
 {
 	debug2("hscroll\n");
+	if (dev->clip.x1 >= dev->clip.x2 || dev->clip.y1 >= dev->clip.y2) return 0;
+	if (sc <= dev->clip.x1 - dev->clip.x2) return 1;
+	if (sc >= dev->clip.x2 - dev->clip.x1) return 1;
 	lock_dev0(dev);
-	lv(dev)->ScrollRect(
-		Rect(	dev->clip.x1 - (scx < 0 ? scx : 0),
-			dev->clip.y1 - (scy < 0 ? scy : 0),
-			dev->clip.x2 - (scx >= 0 ? scx : 0) - 1,
-			dev->clip.y2 - (scy >= 0 ? scy : 0) - 1),
-		Rect(	dev->clip.x1 + (scx >= 0 ? scx : 0),
-			dev->clip.y1 + (scy >= 0 ? scy : 0),
-			dev->clip.x2 + (scx < 0 ? scx : 0) - 1,
-			dev->clip.y2 + (scy < 0 ? scy : 0) - 1)
-	);
+	if (sc > 0) lv(dev)->ScrollRect(Rect(dev->clip.x1, dev->clip.y1, dev->clip.x2 - sc - 1, dev->clip.y2 - 1), Rect(dev->clip.x1 + sc, dev->clip.y1, dev->clip.x2 - 1, dev->clip.y2 - 1));
+	else lv(dev)->ScrollRect(Rect(dev->clip.x1 - sc, dev->clip.y1, dev->clip.x2 - 1, dev->clip.y2 - 1), Rect(dev->clip.x1, dev->clip.y1, dev->clip.x2 + sc - 1, dev->clip.y2 - 1));
 	lv(dev)->d_flush();
 	unlock_dev(dev);
 	return 1;
 }
 
-void ath_set_clip_area(struct graphics_device *dev)
+int ath_vscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
+{
+	debug2("vscroll\n");
+	if (!sc || dev->clip.x1 >= dev->clip.x2 || dev->clip.y1 >= dev->clip.y2) return 0;
+	if (sc <= dev->clip.y1 - dev->clip.y2) return 1;
+	if (sc >= dev->clip.y2 - dev->clip.y1) return 1;
+	lock_dev0(dev);
+	if (sc > 0) lv(dev)->ScrollRect(Rect(dev->clip.x1, dev->clip.y1, dev->clip.x2 - 1, dev->clip.y2 - sc - 1), Rect(dev->clip.x1, dev->clip.y1 + sc, dev->clip.x2 - 1, dev->clip.y2 - 1));
+	else lv(dev)->ScrollRect(Rect(dev->clip.x1, dev->clip.y1 - sc, dev->clip.x2 - 1, dev->clip.y2 - 1), Rect(dev->clip.x1, dev->clip.y1, dev->clip.x2 - 1, dev->clip.y2 + sc - 1));
+	lv(dev)->d_flush();
+	unlock_dev(dev);
+	return 1;
+}
+
+void ath_set_clip_area(struct graphics_device *dev, struct rect *r)
 {
 	debug2("setc\n");
+	generic_set_clip_area(dev, r);
 	lock_dev(dev);
 	lv(dev)->SetDrawingRegion(Region(IRect(dev->clip.x1, dev->clip.y1, dev->clip.x2 - 1, dev->clip.y2 - 1)));
 	unlock_dev(dev);
 }
 
-void ath_flush(struct graphics_device *dev)
-{
-	unregister_bottom_half(do_flush, dev);
-	do_flush(dev);
-}
-
 struct graphics_driver atheos_driver = {
 	(unsigned char *)"atheos",
 	ath_init_driver,
 	ath_init_device,
 	ath_shutdown_device,
 	ath_shutdown_driver,
-	NULL,
-	ath_after_fork,
-	NULL,
-	NULL,
+	dummy_emergency_shutdown,
+	ath_get_driver_param,
 	NULL,
 	NULL,
 	ath_get_empty_bitmap,
+	/*ath_get_filled_bitmap,*/
 	ath_register_bitmap,
 	ath_prepare_strip,
 	ath_commit_strip,
 	ath_unregister_bitmap,
 	ath_draw_bitmap,
+	/*ath_draw_bitmaps,*/
 	ath_get_color,
 	ath_fill_area,
 	ath_draw_hline,
 	ath_draw_vline,
-	ath_scroll,
+	ath_hscroll,
+	ath_vscroll,
 	ath_set_clip_area,
-	ath_flush,
 	dummy_block,
 	dummy_unblock,
-	NULL,				/* set_palette */
-	NULL,				/* get_real_colors */
 	ath_set_title,
 	NULL,				/* exec */
 	NULL,				/* set_clipboard_text */
 	NULL,				/* get_clipboard_text */
 	0,				/* depth */
 	0, 0,				/* size */
-	GD_UNICODE_KEYS | GD_NO_OS_SHELL | GD_NO_LIBEVENT,/* flags */
-	NULL,				/* param */
+	GD_NO_OS_SHELL,			/* flags */
+	0,				/* codepage */
+	NULL,				/* shell */
 };
 
 #endif
