@@ -8,100 +8,61 @@
 /*
 #define LOG_TRANSFER	"/tmp/log"
 #define LOG_SSL
-#define LOG_TO_STDERR
 */
 
-#if defined(LOG_TRANSFER) || defined(LOG_TO_STDERR)
+#ifdef LOG_TRANSFER
 static void log_data(unsigned char *data, int len)
 {
-#ifndef LOG_TO_STDERR
 	static int hlaseno = 0;
 	int fd;
 	if (!hlaseno) {
 		printf("\n"ANSI_SET_BOLD"WARNING -- LOGGING NETWORK TRANSFERS !!!"ANSI_CLEAR_BOLD ANSI_BELL"\n");
 		fflush(stdout);
-		portable_sleep(1000);
+		sleep(1);
 		hlaseno = 1;
 	}
-	fd = c_open3(cast_uchar LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0600);
+	EINTRLOOP(fd, open(LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0600));
 	if (fd != -1) {
 		int rw;
+		set_bin(fd);
 		EINTRLOOP(rw, write(fd, data, len));
 		EINTRLOOP(rw, close(fd));
 	}
-#else
-	int rw;
-	EINTRLOOP(rw, write(2, data, len));
-#endif
-}
-
-static void log_string(unsigned char *data)
-{
-	log_data(data, (int)strlen(cast_const_char data));
-}
-
-static void log_number(int number)
-{
-	unsigned char n[64];
-	snprintf(cast_char n, sizeof n, "%d", number);
-	log_string(n);
 }
 
 #else
-#define log_data(x, y)		do { } while (0)
-#define log_string(x)		do { } while (0)
-#define log_number(x)		do { } while (0)
+#define log_data(x, y)
 #endif
 
-#ifdef HAVE_SSL
-static void log_ssl_error(unsigned char *url, int line, int ret1, int ret2)
+#ifdef LOG_SSL
+#include <openssl/err.h>
+#endif
+
+static inline void log_ssl_error(int line, int ret1, int ret2)
 {
-#ifndef LOG_SSL
-	unsigned long err;
-	while ((err = ERR_get_error())) ;
-#else
-	unsigned char *u, *uu;
-	u = stracpy(url);
-	if ((uu = cast_uchar strchr(cast_const_char u, POST_CHAR))) *uu = 0;
-#if defined(HAVE_SSL_LOAD_ERROR_STRINGS) || defined(SSL_load_error_strings)
+#ifdef LOG_SSL
 	SSL_load_error_strings();
-#endif
-#if defined(HAVE_OPENSSL) && !defined(OPENSSL_NO_STDIO)
 	ERR_print_errors_fp(stderr);
-#else
-	{
-		unsigned long err;
-		while ((err = ERR_get_error())) {
-			unsigned char buf[1024];
-			ERR_error_string_n(err, cast_char buf, sizeof buf);
-			fprintf(stderr, "%s\n", buf);
-		}
-	}
-#endif
-	fprintf(stderr, "ssl error at %d: %d, %d, %d (%s), url (%s)\n", line, ret1, ret2, errno, strerror(errno), u);
-	mem_free(u);
+	debug("ssl error at %d: %d, %d, %d (%s)", line, ret1, ret2, errno, strerror(errno));
 #endif
 }
 
-void clear_ssl_errors(int line)
-{
-	if (ERR_peek_error())
-		log_ssl_error(cast_uchar "", line, 0, 0);
-}
-#endif
-
-static void connected(void *);
-static void update_dns_priority(struct connection *);
-static void connected_callback(struct connection *);
-static void dns_found(void *, int);
+static void connected(struct connection *);
+static void dns_found(struct connection *, int);
 static void try_connect(struct connection *);
-static void handle_socks_reply(void *);
+static void handle_socks_reply(struct connection *);
+
+static void exception(struct connection *c)
+{
+	setcstate(c, S_EXCEPT);
+	retry_connection(c);
+}
 
 int socket_and_bind(int pf, unsigned char *address)
 {
 	int s;
 	int rs;
-	s = c_socket(pf, SOCK_STREAM, IPPROTO_TCP);
+	EINTRLOOP(s, socket(pf, SOCK_STREAM, IPPROTO_TCP));
 	if (s == -1)
 		return -1;
 	if (address && *address) {
@@ -110,7 +71,9 @@ int socket_and_bind(int pf, unsigned char *address)
 			struct sockaddr_in sa;
 			unsigned char addr[4];
 			if (numeric_ip_address(address, addr) == -1) {
-				goto bind_to_iface;
+				EINTRLOOP(rs, close(s));
+				errno = EINVAL;
+				return -1;
 			}
 			memset(&sa, 0, sizeof sa);
 			sa.sin_family = AF_INET;
@@ -131,7 +94,9 @@ int socket_and_bind(int pf, unsigned char *address)
 			unsigned char addr[16];
 			unsigned scope;
 			if (numeric_ipv6_address(address, addr, &scope) == -1) {
-				goto bind_to_iface;
+				EINTRLOOP(rs, close(s));
+				errno = EINVAL;
+				return -1;
 			}
 			memset(&sa, 0, sizeof sa);
 			sa.sin6_family = AF_INET6;
@@ -158,103 +123,84 @@ int socket_and_bind(int pf, unsigned char *address)
 		}
 	}
 	return s;
-bind_to_iface:
-	errno = EINVAL;
-#ifdef SO_BINDTODEVICE
-	EINTRLOOP(rs, setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, address, (socklen_t)(strlen(cast_const_char address) + 1)));
-	if (!rs) {
-		return s;
-	}
-#endif
-	EINTRLOOP(rs, close(s));
-	return -1;
 }
 
 void close_socket(int *s)
 {
 	int rs;
 	if (*s == -1) return;
-	set_handlers(*s, NULL, NULL, NULL);
 	EINTRLOOP(rs, close(*s));
+	set_handlers(*s, NULL, NULL, NULL, NULL);
 	*s = -1;
 }
 
 struct conn_info {
 	void (*func)(struct connection *);
-	struct lookup_state l;
+	struct lookup_result addr;
+	int addr_index;
 	int first_error;
+	int port;
 	int *sock;
+	int real_port;
 	int socks_byte_count;
-	int socks_handled;
 	unsigned char socks_reply[8];
-	unsigned char host[1];
+	unsigned char *host;
+	unsigned char *dns_append;
 };
 
 void make_connection(struct connection *c, int port, int *sock, void (*func)(struct connection *))
 {
-	int socks_port = -1;
+	int real_port = -1;
+	int as;
+	unsigned char *dns_append = cast_uchar "";
 	unsigned char *host;
-	size_t sl;
 	struct conn_info *b;
 	if (*c->socks_proxy) {
-		/* pasing duplicated in proxy_ok_dialog */
 		unsigned char *p = cast_uchar strchr(cast_const_char c->socks_proxy, '@');
 		if (p) p++;
 		else p = c->socks_proxy;
 		host = stracpy(p);
-		socks_port = 1080;
+		real_port = port;
+		port = 1080;
 		if ((p = cast_uchar strchr(cast_const_char host, ':'))) {
 			long lp;
-			char *end;
 			*p++ = 0;
 			if (!*p) goto badu;
-			lp = strtol(cast_const_char p, &end, 10);
-			if (*end || lp <= 0 || lp >= 65536) {
+			lp = strtol(cast_const_char p, (char **)(void *)&p, 10);
+			if (*p || lp <= 0 || lp >= 65536) {
 				badu:
 				mem_free(host);
 				setcstate(c, S_BAD_URL);
 				abort_connection(c);
 				return;
 			}
-			socks_port = (int)lp;
+			port = (int)lp;
 		}
+		dns_append = proxies.dns_append;
 	} else if (!(host = get_host_name(c->url))) {
 		setcstate(c, S_INTERNAL);
 		abort_connection(c);
 		return;
 	}
 	if (c->newconn)
-		internal_error("already making a connection");
-	sl = strlen(cast_const_char host);
-	if (sl > MAXINT - sizeof(struct conn_info)) overalloc();
-	b = mem_calloc(sizeof(struct conn_info) + sl);
+		internal("already making a connection");
+	b = mem_calloc(sizeof(struct conn_info) + strlen(cast_const_char host) + 1 + strlen(cast_const_char dns_append) + 1);
 	b->func = func;
 	b->sock = sock;
-	b->l.socks_port = socks_port;
-	b->l.target_port = port;
+	b->port = port;
+	b->real_port = real_port;
+	b->host = (unsigned char *)(b + 1);
 	strcpy(cast_char b->host, cast_const_char host);
+	b->dns_append = cast_uchar strchr(cast_const_char b->host, 0) + 1;
+	strcpy(cast_char b->dns_append, cast_const_char dns_append);
 	c->newconn = b;
-	log_string(cast_uchar "\nCONNECTION: ");
+	log_data(cast_uchar "\nCONNECTION: ", 13);
 	log_data(host, strlen(cast_const_char host));
-	log_string(cast_uchar ":");
-	log_number(socks_port != -1 ? socks_port : port);
-	log_string(cast_uchar "\n");
-	if (c->last_lookup_state.addr_index < c->last_lookup_state.addr.n) {
-		b->l.addr = c->last_lookup_state.addr;
-		b->l.addr_index = c->last_lookup_state.addr_index;
-		b->l.dont_try_more_servers = 1;
-		dns_found(c, 0);
-	} else {
-		int q = -1;
-		if (c->no_cache < NC_RELOAD) {
-			q = find_host_in_cache(host, &b->l.addr, &c->dnsquery, dns_found, c);
-		}
-		if (q) {
-			setcstate(c, S_DNS);
-			find_host_no_cache(host, c->doh, &b->l.addr, &c->dnsquery, dns_found, c);
-		}
-	}
+	log_data(cast_uchar "\n", 1);
+	if (c->no_cache >= NC_RELOAD) as = find_host_no_cache(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
+	else as = find_host(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
 	mem_free(host);
+	if (as) setcstate(c, S_DNS);
 }
 
 int is_ipv6(int h)
@@ -291,7 +237,7 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 		errno = EINVAL;
 		goto e;
 	}
-	s = c_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+	EINTRLOOP(s, socket(PF_INET, SOCK_STREAM, IPPROTO_TCP));
 	if (s == -1) goto e;
 	*sock = s;
 	set_nonblock(s);
@@ -315,6 +261,7 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 }
 
 #ifdef SUPPORT_IPV6
+
 int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char *result)
 {
 	int s;
@@ -330,7 +277,7 @@ int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char
 		errno = EINVAL;
 		goto e;
 	}
-	s = c_socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
+	EINTRLOOP(s, socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP));
 	if (s == -1) goto e;
 	*sock = s;
 	set_nonblock(s);
@@ -368,132 +315,44 @@ int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char
 	retry_connection(c);
 	return -1;
 }
+
 #endif
 
 #ifdef HAVE_SSL
-static void ssl_setup_downgrade(struct connection *c)
+static void ssl_want_read(struct connection *c)
 {
-#if !defined(HAVE_NSS)
-	int send_scsv = 0;
-	int dd = c->no_tls;
-	/* avoid warnings */
-	send_scsv++; send_scsv--;
-	dd++; dd--;
-	/*debug("no tls: %d", dd);*/
-#ifdef HAVE_SSL_SET_SECURITY_LEVEL
-	if (dd) {
-		SSL_set_security_level(c->ssl->ssl, 0);
-		dd--;
-	}
-#endif
-#ifdef SSL_OP_NO_TLSv1_3
-	if (dd) {
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_TLSv1_3);
-		send_scsv = 1;
-		dd--;
-	}
-#endif
-#ifdef SSL_OP_NO_TLSv1_2
-	if (dd) {
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_TLSv1_2);
-		send_scsv = 1;
-		dd--;
-	}
-#endif
-#ifdef SSL_OP_NO_TLSv1_1
-	if (dd) {
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_TLSv1_1);
-		send_scsv = 1;
-		dd--;
-	}
-#endif
-#if defined(SSL_OP_NO_TLSv1) && !defined(OPENSSL_NO_SSL3)
-	if (dd) {
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_TLSv1);
-		send_scsv = 1;
-		dd--;
-	}
-#endif
-#ifdef SSL_MODE_SEND_FALLBACK_SCSV
-	if (send_scsv) {
-		SSL_set_mode(c->ssl->ssl, SSL_MODE_SEND_FALLBACK_SCSV);
-	}
-#endif
-	if (SCRUB_HEADERS) {
-#ifdef SSL_OP_NO_SSLv2
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_SSLv2);
-#endif
-#ifdef SSL_OP_NO_SSLv3
-		SSL_set_options(c->ssl->ssl, SSL_OP_NO_SSLv3);
-#endif
-	}
-#endif
-}
-
-static void ssl_downgrade_dance(struct connection *c)
-{
-#if !defined(HAVE_NSS)
-	int downgrades = 0;
-	c->no_ssl_session = 1;
-	if (c->ssl && c->ssl->session_set) {
-		retry_connect(c, S_SSL_ERROR, 1);
-		return;
-	}
-#ifdef HAVE_SSL_SET_SECURITY_LEVEL
-	downgrades++;
-#endif
-#ifdef SSL_OP_NO_TLSv1_3
-	downgrades++;
-#endif
-#ifdef SSL_OP_NO_TLSv1_2
-	downgrades++;
-#endif
-#ifdef SSL_OP_NO_TLSv1_1
-	downgrades++;
-#endif
-#if defined(SSL_OP_NO_TLSv1) && !defined(OPENSSL_NO_SSL3)
-	downgrades++;
-#endif
-	if (++c->no_tls <= downgrades) {
-		retry_connect(c, S_SSL_ERROR, 1);
-	} else {
-		c->no_tls = 0;
-		retry_connect(c, S_SSL_ERROR, 0);
-	}
-#else
-	retry_connect(c, S_SSL_ERROR, 0);
-#endif
-}
-
-static void ssl_want_io(void *c_)
-{
-	struct connection *c = (struct connection *)c_;
 	int ret1, ret2;
 	struct conn_info *b = c->newconn;
 
 	set_connection_timeout(c);
 
-	switch ((ret2 = SSL_get_error(c->ssl->ssl, ret1 = SSL_connect(c->ssl->ssl)))) {
+#ifndef HAVE_NSS
+	if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
+#endif
+	switch ((ret2 = SSL_get_error(c->ssl, ret1 = SSL_connect(c->ssl)))) {
 		case SSL_ERROR_NONE:
-			connected_callback(c);
+			c->newconn = NULL;
+			b->func(c);
+			mem_free(b);
 			break;
 		case SSL_ERROR_WANT_READ:
-			set_handlers(*b->sock, ssl_want_io, NULL, c);
+			set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
 			break;
 		case SSL_ERROR_WANT_WRITE:
-			set_handlers(*b->sock, NULL, ssl_want_io, c);
+			set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
 			break;
 		default:
-			log_ssl_error(c->url, __LINE__, ret1, ret2);
-			ssl_downgrade_dance(c);
+			log_ssl_error(__LINE__, ret1, ret2);
+			c->no_tsl++;
+			setcstate(c, S_SSL_ERROR);
+			retry_connection(c);
 			break;
 	}
 }
 #endif
 
-static void handle_socks(void *c_)
+static void handle_socks(struct connection *c)
 {
-	struct connection *c = (struct connection *)c_;
 	struct conn_info *b = c->newconn;
 	unsigned char *command = init_str();
 	int len = 0;
@@ -502,8 +361,8 @@ static void handle_socks(void *c_)
 	setcstate(c, S_SOCKS_NEG);
 	set_connection_timeout(c);
 	add_bytes_to_str(&command, &len, cast_uchar "\004\001", 2);
-	add_chr_to_str(&command, &len, b->l.target_port >> 8);
-	add_chr_to_str(&command, &len, b->l.target_port);
+	add_chr_to_str(&command, &len, b->real_port >> 8);
+	add_chr_to_str(&command, &len, b->real_port);
 	add_bytes_to_str(&command, &len, cast_uchar "\000\000\000\001", 4);
 	if (strchr(cast_const_char c->socks_proxy, '@'))
 		add_bytes_to_str(&command, &len, c->socks_proxy, strcspn(cast_const_char c->socks_proxy, "@"));
@@ -515,7 +374,7 @@ static void handle_socks(void *c_)
 		return;
 	}
 	add_to_str(&command, &len, host);
-	add_to_str(&command, &len, c->dns_append);
+	add_to_str(&command, &len, b->dns_append);
 	add_chr_to_str(&command, &len, 0);
 	mem_free(host);
 	if (b->socks_byte_count >= len) {
@@ -524,10 +383,6 @@ static void handle_socks(void *c_)
 		retry_connection(c);
 		return;
 	}
-	if (!b->socks_byte_count) {
-		log_data(command, len);
-		log_string(cast_uchar "\n");
-	}
 	EINTRLOOP(wr, (int)write(*b->sock, command + b->socks_byte_count, len - b->socks_byte_count));
 	mem_free(command);
 	if (wr <= 0) {
@@ -537,18 +392,17 @@ static void handle_socks(void *c_)
 	}
 	b->socks_byte_count += wr;
 	if (b->socks_byte_count < len) {
-		set_handlers(*b->sock, NULL, handle_socks, c);
+		set_handlers(*b->sock, NULL, (void(*)(void *))handle_socks, (void(*)(void *))exception, c);
 		return;
 	} else {
 		b->socks_byte_count = 0;
-		set_handlers(*b->sock, handle_socks_reply, NULL, c);
+		set_handlers(*b->sock, (void(*)(void *))handle_socks_reply, NULL, (void(*)(void *))exception, c);
 		return;
 	}
 }
 
-static void handle_socks_reply(void *c_)
+static void handle_socks_reply(struct connection *c)
 {
-	struct connection *c = (struct connection *)c_;
 	struct conn_info *b = c->newconn;
 	int rd;
 	set_connection_timeout(c);
@@ -560,7 +414,7 @@ static void handle_socks_reply(void *c_)
 	}
 	b->socks_byte_count += rd;
 	if (b->socks_byte_count < (int)sizeof b->socks_reply) return;
-	/*debug("%x %x %x %x %x %x %x %x", b->socks_reply[0], b->socks_reply[1], b->socks_reply[2], b->socks_reply[3], b->socks_reply[4], b->socks_reply[5], b->socks_reply[6], b->socks_reply[7]);*/
+	/* debug("%x %x %x %x %x %x %x %x", b->socks_reply[0], b->socks_reply[1], b->socks_reply[2], b->socks_reply[3], b->socks_reply[4], b->socks_reply[5], b->socks_reply[6], b->socks_reply[7]); */
 	if (b->socks_reply[0]) {
 		setcstate(c, S_BAD_SOCKS_VERSION);
 		abort_connection(c);
@@ -586,46 +440,29 @@ static void handle_socks_reply(void *c_)
 		case 90:
 			break;
 	}
+	b->real_port = -1;
 	connected(c);
 }
 
-static void dns_found(void *c_, int state)
+static void dns_found(struct connection *c, int state)
 {
-	struct connection *c = (struct connection *)c_;
 	if (state) {
-		setcstate(c, *c->socks_proxy || is_proxy_url(c->url) ? S_NO_PROXY_DNS : S_NO_DNS);
+		setcstate(c, S_NO_DNS);
 		abort_connection(c);
 		return;
 	}
 	try_connect(c);
 }
 
-void retry_connect(struct connection *c, int err, int ssl_downgrade)
+static void retry_connect(struct connection *c, int err)
 {
 	struct conn_info *b = c->newconn;
-	if (!b->l.addr_index) b->first_error = err;
-#ifdef HAVE_SSL
-	if (c->ssl) {
-		freeSSL(c->ssl);
-		if (is_proxy_url(c->url)) c->ssl = NULL;
-		else c->ssl = DUMMY;
-	}
-#endif
-	if (ssl_downgrade) {
-		log_string(cast_uchar "\nSSL DOWNGRADE\n");
-		close_socket(b->sock);
-		try_connect(c);
-		return;
-	}
-	b->l.addr_index++;
-	if (b->l.addr_index < b->l.addr.n && !b->l.dont_try_more_servers) {
-		if (b->l.addr_index == 1)
-			rotate_addresses(&b->l.addr);
-		log_string(cast_uchar "\nNEXT ADDRESS\n");
+	if (!b->addr_index) b->first_error = err;
+	b->addr_index++;
+	if (b->addr_index < b->addr.n) {
 		close_socket(b->sock);
 		try_connect(c);
 	} else {
-		dns_clear_host(b->host);
 		setcstate(c, b->first_error);
 		retry_connection(c);
 	}
@@ -633,13 +470,10 @@ void retry_connect(struct connection *c, int err, int ssl_downgrade)
 
 static void try_connect(struct connection *c)
 {
-	int i;
 	int s;
 	int rs;
-	unsigned short p;
 	struct conn_info *b = c->newconn;
-	struct host_address *addr = &b->l.addr.a[b->l.addr_index];
-	/*debug("%p: %p %d %d\n", b, addr, b->l.addr_index, addr->af);*/
+	struct host_address *addr = &b->addr.a[b->addr_index];
 	if (addr->af == AF_INET) {
 		s = socket_and_bind(PF_INET, bind_ip_address);
 #ifdef SUPPORT_IPV6
@@ -652,31 +486,17 @@ static void try_connect(struct connection *c)
 		return;
 	}
 	if (s == -1) {
-		retry_connect(c, get_error_from_errno(errno), 0);
+		retry_connect(c, get_error_from_errno(errno));
 		return;
 	}
 	set_nonblock(s);
 	*b->sock = s;
-	b->socks_handled = 0;
-	b->socks_byte_count = 0;
-	p = b->l.socks_port != -1 ? b->l.socks_port : b->l.target_port;
-	log_string(cast_uchar "\nADDRESS: ");
-	for (i = 0; i < (addr->af == AF_INET ? 4 : 16); i++) {
-		if (i) log_string(cast_uchar ".");
-		log_number(addr->addr[i]);
-	}
-	log_string(cast_uchar ":");
-	log_number(p);
-	log_string(cast_uchar "\n");
 	if (addr->af == AF_INET) {
 		struct sockaddr_in sa;
 		memset(&sa, 0, sizeof sa);
 		sa.sin_family = AF_INET;
 		memcpy(&sa.sin_addr.s_addr, addr->addr, 4);
-		sa.sin_port = htons(p);
-#if defined(__aarch64__) && defined(__ILP32__)
-		errno = EINPROGRESS;	/* arm64 ilp32 bug */
-#endif
+		sa.sin_port = htons(b->port);
 		EINTRLOOP(rs, connect(s, (struct sockaddr *)(void *)&sa, sizeof sa));
 #ifdef SUPPORT_IPV6
 	} else if (addr->af == AF_INET6) {
@@ -687,10 +507,7 @@ static void try_connect(struct connection *c)
 #ifdef SUPPORT_IPV6_SCOPE
 		sa.sin6_scope_id = addr->scope_id;
 #endif
-		sa.sin6_port = htons(p);
-#if defined(__aarch64__) && defined(__ILP32__)
-		errno = EINPROGRESS;	/* arm64 ilp32 bug */
-#endif
+		sa.sin6_port = htons(b->port);
 		EINTRLOOP(rs, connect(s, (struct sockaddr *)(void *)&sa, sizeof sa));
 #endif
 	} else {
@@ -702,56 +519,44 @@ static void try_connect(struct connection *c)
 #ifdef BEOS
 			if (errno == EWOULDBLOCK) errno = ETIMEDOUT;
 #endif
-			retry_connect(c, get_error_from_errno(errno), 0);
+			retry_connect(c, get_error_from_errno(errno));
 			return;
 		}
-		set_handlers(s, NULL, connected, c);
-		setcstate(c, !b->l.addr_index ? S_CONN : S_CONN_ANOTHER);
-		if (b->l.addr.n > 1 && (is_connection_restartable(c) || max_tries == 1)) {
-			set_connection_timeout(c);
-		}
+		set_handlers(s, NULL, (void(*)(void *))connected, (void(*)(void *))exception, c);
+		setcstate(c, !b->addr_index ? S_CONN : S_CONN_ANOTHER);
 	} else {
 		connected(c);
 	}
 }
 
-#ifdef HAVE_SSL
 void continue_connection(struct connection *c, int *sock, void (*func)(struct connection *))
 {
 	struct conn_info *b;
 	if (c->newconn)
-		internal_error("already making a connection");
+		internal("already making a connection");
 	b = mem_calloc(sizeof(struct conn_info));
 	b->func = func;
 	b->sock = sock;
-	b->l = c->last_lookup_state;
-	b->socks_handled = 1;
+	b->real_port = -1;
 	c->newconn = b;
-	log_string(cast_uchar "\nCONTINUE CONNECTION\n");
+	log_data(cast_uchar "\nCONTINUE CONNECTION\n", 21);
 	connected(c);
 }
-#endif
 
-static void connected(void *c_)
+static void connected(struct connection *c)
 {
-	struct connection *c = (struct connection *)c_;
 	struct conn_info *b = c->newconn;
-#ifdef SO_ERROR
 	int err = 0;
 	socklen_t len = sizeof(int);
 	int rs;
-#endif
-	clear_connection_timeout(c);
 #ifdef SO_ERROR
 	errno = 0;
 	EINTRLOOP(rs, getsockopt(*b->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &len));
 	if (!rs) {
-#ifdef OS2
 		if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
-#endif
 	} else {
 		if (!(err = errno)) {
-			retry_connect(c, S_STATE, 0);
+			retry_connect(c, S_STATE);
 			return;
 		}
 	}
@@ -763,168 +568,53 @@ static void connected(void *c_)
 #ifdef DOS
 		if (err == EALREADY) err = ETIMEDOUT;
 #endif
-		retry_connect(c, get_error_from_errno(err), 0);
+		retry_connect(c, get_error_from_errno(err));
 		return;
 	}
 #endif
+	if (b->addr_index) {
+		int i;
+		for (i = 0; i < b->addr_index; i++)
+			dns_set_priority(b->host, &b->addr.a[i], 0);
+		dns_set_priority(b->host, &b->addr.a[i], 1);
+	}
 	set_connection_timeout(c);
-	if (b->l.socks_port != -1 && !b->socks_handled) {
-		b->socks_handled = 1;
-		update_dns_priority(c);
+	if (b->real_port != -1) {
 		handle_socks(c);
 		return;
 	}
-	log_string(cast_uchar "\nCONNECTED\n");
 #ifdef HAVE_SSL
 	if (c->ssl) {
 		int ret1, ret2;
-		unsigned char *orig_url = remove_proxy_prefix(c->url);
-		unsigned char *h = get_host_name(orig_url);
-		log_string(cast_uchar "\nSSL\n");
-		if (*h && h[strlen(cast_const_char h) - 1] == '.') {
-			h[strlen(cast_const_char h) - 1] = 0;
-		}
 		c->ssl = getSSL();
 		if (!c->ssl) {
-			c->ssl = DUMMY;
-			ret1 = ret2 = 0;
-			mem_free(h);
 			goto ssl_error;
 		}
-#ifdef SSL_SESSION_RESUME
-		if (!proxies.only_proxies && !c->no_ssl_session) {
-			unsigned char *h = get_host_name(orig_url);
-			int p = get_port(orig_url);
-			SSL_SESSION *ses = get_session_cache_entry(c->ssl->ctx, h, p);
-			if (ses) {
-				if (SSL_set_session(c->ssl->ssl, ses) == 1)
-					c->ssl->session_set = 1;
-			}
-			mem_free(h);
-		}
+		SSL_set_fd(c->ssl, *b->sock);
+#ifndef HAVE_NSS
+		if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
 #endif
-#if defined(HAVE_SSL_CERTIFICATES) && !defined(OPENSSL_NO_STDIO)
-		if (!proxies.only_proxies) {
-			if (ssl_options.client_cert_key[0]) {
-				SSL_use_PrivateKey_file(c->ssl->ssl, cast_const_char ssl_options.client_cert_key, SSL_FILETYPE_PEM);
-			}
-			if (ssl_options.client_cert_crt[0]) {
-				SSL_use_certificate_file(c->ssl->ssl, cast_const_char ssl_options.client_cert_crt, SSL_FILETYPE_PEM);
-			}
-		}
-#endif
-		SSL_set_fd(c->ssl->ssl, *b->sock);
-		ssl_setup_downgrade(c);
-#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
-		if (h[0] == '[' || !numeric_ip_address(h, NULL)
-#ifdef SUPPORT_IPV6
-		    || !numeric_ipv6_address(h, NULL, NULL)
-#endif
-		    ) goto skip_numeric_address;
-		SSL_set_tlsext_host_name(c->ssl->ssl, cast_const_char h);
-skip_numeric_address:
-#endif
-		mem_free(h);
-		switch ((ret2 = SSL_get_error(c->ssl->ssl, ret1 = SSL_connect(c->ssl->ssl)))) {
+		switch ((ret2 = SSL_get_error(c->ssl, ret1 = SSL_connect(c->ssl)))) {
 			case SSL_ERROR_WANT_READ:
 				setcstate(c, S_SSL_NEG);
-				set_handlers(*b->sock, ssl_want_io, NULL, c);
+				set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
 				return;
 			case SSL_ERROR_WANT_WRITE:
 				setcstate(c, S_SSL_NEG);
-				set_handlers(*b->sock, NULL, ssl_want_io, c);
+				set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
 				return;
 			case SSL_ERROR_NONE:
 				break;
 			default:
+				log_ssl_error(__LINE__, ret1, ret2);
 			ssl_error:
-				log_ssl_error(c->url, __LINE__, ret1, ret2);
-				ssl_downgrade_dance(c);
+				c->no_tsl++;
+				setcstate(c, S_SSL_ERROR);
+				retry_connection(c);
 				return;
 		}
 	}
 #endif
-	connected_callback(c);
-}
-
-static void update_dns_priority(struct connection *c)
-{
-	struct conn_info *b = c->newconn;
-	if (!b->l.dont_try_more_servers && b->host[0]) {
-		if (b->l.addr_index) {
-			int i;
-			for (i = 0; i < b->l.addr_index; i++)
-				dns_set_priority(b->host, &b->l.addr.a[i], 0);
-			dns_set_priority(b->host, &b->l.addr.a[i], 1);
-		}
-		b->l.dont_try_more_servers = 1;
-	}
-}
-
-static void connected_callback(struct connection *c)
-{
-	struct conn_info *b = c->newconn;
-	update_dns_priority(c);
-#ifdef HAVE_SSL_CERTIFICATES
-	if (c->ssl) {
-		if (ssl_options.certificates != SSL_ACCEPT_INVALID_CERTIFICATE) {
-			unsigned char *h = get_host_name(remove_proxy_prefix(c->url));
-			int err = verify_ssl_certificate(c->ssl, h);
-			if (err) {
-				if (ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
-					int flags = get_blacklist_flags(h);
-					if (flags & BL_IGNORE_CERTIFICATE)
-						goto ignore_cert;
-				}
-				mem_free(h);
-				setcstate(c, err);
-				abort_connection(c);
-				return;
-			}
-			ignore_cert:
-
-			if (c->no_tls) {
-#ifdef HAVE_SSL_SET_SECURITY_LEVEL
-				if (c->no_tls == 1) {
-					err = S_INSECURE_CIPHER;
-					goto weak_cipher;
-				}
-#endif
-				if (ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
-					int flags = get_blacklist_flags(h);
-					if (flags & BL_IGNORE_DOWNGRADE)
-						goto ignore_downgrade;
-				}
-				mem_free(h);
-				setcstate(c, S_DOWNGRADED_METHOD);
-				abort_connection(c);
-				return;
-			}
-			ignore_downgrade:
-
-			err = verify_ssl_cipher(c->ssl);
-			if (err) {
-#ifdef HAVE_SSL_SET_SECURITY_LEVEL
-				weak_cipher:
-#endif
-				if (ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
-					int flags = get_blacklist_flags(h);
-					if (flags & BL_IGNORE_CIPHER)
-						goto ignore_cipher;
-				}
-				mem_free(h);
-				setcstate(c, err);
-				abort_connection(c);
-				return;
-			}
-			ignore_cipher:
-
-			mem_free(h);
-		}
-	}
-#endif
-	retrieve_ssl_session(c);
-	c->last_lookup_state = b->l;
 	c->newconn = NULL;
 	b->func(c);
 	mem_free(b);
@@ -938,45 +628,35 @@ struct write_buffer {
 	unsigned char data[1];
 };
 
-static void write_select(void *c_)
+static void write_select(struct connection *c)
 {
-	struct connection *c = (struct connection *)c_;
 	struct write_buffer *wb;
 	int wr;
 	if (!(wb = c->buffer)) {
-		internal_error("write socket has no buffer");
+		internal("write socket has no buffer");
 		setcstate(c, S_INTERNAL);
 		abort_connection(c);
 		return;
 	}
-	if (wb->pos)
-		set_connection_timeout(c);
-	else
-		set_connection_timeout_keepal(c);
+	set_connection_timeout(c);
 	/*printf("ws: %d\n",wb->len-wb->pos);
 	for (wr = wb->pos; wr < wb->len; wr++) printf("%c", wb->data[wr]);
 	printf("-\n");*/
 
 #ifdef HAVE_SSL
-	if (c->ssl) {
-		set_handlers(wb->sock, NULL, write_select, c);
-		if ((wr = SSL_write(c->ssl->ssl, (void *)(wb->data + wb->pos), wb->len - wb->pos)) <= 0) {
+	if(c->ssl) {
+		if ((wr = SSL_write(c->ssl, wb->data + wb->pos, wb->len - wb->pos)) <= 0) {
 			int err;
-			err = SSL_get_error(c->ssl->ssl, wr);
-			if (err == SSL_ERROR_WANT_WRITE) {
+			if ((err = SSL_get_error(c->ssl, wr)) != SSL_ERROR_WANT_WRITE) {
+				setcstate(c, wr ? (err == SSL_ERROR_SYSCALL ? get_error_from_errno(errno) : S_SSL_ERROR) : S_CANT_WRITE);
+				if (c->state == S_SSL_ERROR)
+					log_ssl_error(__LINE__, wr, err);
+				if (!wr || err == SSL_ERROR_SYSCALL) retry_connection(c);
+				else abort_connection(c);
 				return;
 			}
-			if (err == SSL_ERROR_WANT_READ) {
-				set_handlers(wb->sock, write_select, NULL, c);
-				return;
-			}
-			setcstate(c, wr ? (err == SSL_ERROR_SYSCALL ? get_error_from_errno(errno) : S_SSL_ERROR) : S_CANT_WRITE);
-			log_ssl_error(c->url, __LINE__, wr, err);
-			if (!wr || err == SSL_ERROR_SYSCALL) retry_connection(c);
-			else abort_connection(c);
-			return;
+			else return;
 		}
-		c->ssl->bytes_written += wr;
 	} else
 #endif
 	{
@@ -997,7 +677,7 @@ static void write_select(void *c_)
 	if ((wb->pos += wr) == wb->len) {
 		void (*f)(struct connection *) = wb->done;
 		c->buffer = NULL;
-		set_handlers(wb->sock, NULL, NULL, NULL);
+		set_handlers(wb->sock, NULL, NULL, NULL, NULL);
 		mem_free(wb);
 		f(c);
 	}
@@ -1016,25 +696,24 @@ void write_to_socket(struct connection *c, int s, unsigned char *data, int len,
 	memcpy(wb->data, data, len);
 	if (c->buffer) mem_free(c->buffer);
 	c->buffer = wb;
-	set_handlers(s, NULL, write_select, c);
+	set_handlers(s, NULL, (void (*)(void *))write_select, (void (*)(void *))exception, c);
 }
 
 #define READ_SIZE	64240
 #define TOTAL_READ	(4193008 - READ_SIZE)
 
-static void read_select(void *c_)
+static void read_select(struct connection *c)
 {
-	struct connection *c = (struct connection *)c_;
 	int total_read = 0;
 	struct read_buffer *rb;
 	int rd;
 	if (!(rb = c->buffer)) {
-		internal_error("read socket has no buffer");
+		internal("read socket has no buffer");
 		setcstate(c, S_INTERNAL);
 		abort_connection(c);
 		return;
 	}
-	set_handlers(rb->sock, NULL, NULL, NULL);
+	set_handlers(rb->sock, NULL, NULL, NULL, NULL);
 
 read_more:
 	if ((unsigned)rb->len > MAXINT - sizeof(struct read_buffer) - READ_SIZE) overalloc();
@@ -1043,16 +722,11 @@ read_more:
 
 #ifdef HAVE_SSL
 	if (c->ssl) {
-		if ((rd = SSL_read(c->ssl->ssl, (void *)(rb->data + rb->len), READ_SIZE)) <= 0) {
+		if ((rd = SSL_read(c->ssl, rb->data + rb->len, READ_SIZE)) <= 0) {
 			int err;
 			if (total_read) goto success;
-			err = SSL_get_error(c->ssl->ssl, rd);
-			if (err == SSL_ERROR_WANT_READ) {
-				set_handlers(rb->sock, read_select, NULL, c);
-				return;
-			}
-			if (err == SSL_ERROR_WANT_WRITE) {
-				set_handlers(rb->sock, NULL, read_select, c);
+			if ((err = SSL_get_error(c->ssl, rd)) == SSL_ERROR_WANT_READ) {
+				read_from_socket(c, rb->sock, rb, rb->done);
 				return;
 			}
 			if (rb->close && !rd) {
@@ -1061,12 +735,13 @@ read_more:
 				return;
 			}
 			setcstate(c, rd ? (err == SSL_ERROR_SYSCALL ? get_error_from_errno(errno) : S_SSL_ERROR) : S_CANT_READ);
-			log_ssl_error(c->url, __LINE__, rd, err);
+			if (c->state == S_SSL_ERROR)
+				log_ssl_error(__LINE__, rd, err);
+			/*mem_free(rb);*/
 			if (!rd || err == SSL_ERROR_SYSCALL) retry_connection(c);
 			else abort_connection(c);
 			return;
 		}
-		c->ssl->bytes_read += rd;
 	} else
 #endif
 	{
@@ -1078,7 +753,24 @@ read_more:
 				rb->done(c, rb);
 				return;
 			}
+			if (!rd) {
+/* Many servers supporting compression have a bug
+   --- they send the size of uncompressed data.
+   Turn off compression support once before the final retry.
+*/
+				unsigned char *prot, *h;
+				if (is_last_try(c) && (prot = get_protocol_name(c->url))) {
+					if (!strcasecmp(cast_const_char prot, "http")) {
+						if ((h = get_host_name(c->url))) {
+							add_blacklist_entry(h, BL_NO_COMPRESSION);
+							mem_free(h);
+						}
+					}
+					mem_free(prot);
+				}
+			}
 			setcstate(c, rd ? get_error_from_errno(errno) : S_CANT_READ);
+			/*mem_free(rb);*/
 			retry_connection(c);
 			return;
 		}
@@ -1096,7 +788,6 @@ read_more:
 			goto read_more;
 	}
 success:
-	retrieve_ssl_session(c);
 	rb->done(c, rb);
 }
 
@@ -1114,13 +805,13 @@ void read_from_socket(struct connection *c, int s, struct read_buffer *buf, void
 	buf->sock = s;
 	if (c->buffer && buf != c->buffer) mem_free(c->buffer);
 	c->buffer = buf;
-	set_handlers(s, read_select, NULL, c);
+	set_handlers(s, (void (*)(void *))read_select, NULL, (void (*)(void *))exception, c);
 }
 
 void kill_buffer_data(struct read_buffer *rb, int n)
 {
 	if (n > rb->len || n < 0) {
-		internal_error("called kill_buffer_data with bad value");
+		internal("called kill_buffer_data with bad value");
 		rb->len = 0;
 		return;
 	}
