@@ -20,11 +20,11 @@
 
 #ifdef GRDRV_DIRECTFB
 
-#include "links.h"
+#include <netinet/in.h>  /* for htons */
 
-#undef debug
 #include <directfb.h>
 
+#include "links.h"
 #include "dfb_cur.h"
 
 
@@ -52,17 +52,18 @@ static IDirectFBDisplayLayer *layer       = NULL;
 static IDirectFBSurface      *arrow       = NULL;
 static IDirectFBEventBuffer  *events      = NULL;
 static DFBSurfacePixelFormat  pixelformat = DSPF_UNKNOWN;
-static struct timer *event_timer = NULL;
+static int                    event_timer = -1;
 
 
+static inline void directfb_set_color  (IDirectFBSurface *surface, long color);
 static void directfb_register_flip     (DFBDeviceData *data,
                                         int x, int y, int w, int h);
 static void directfb_flip_surface      (void *pointer);
 static void directfb_check_events      (void *pointer);
 static void directfb_translate_key     (DFBWindowEvent *event,
                                         int *key, int *flag);
-static void directfb_add_to_table      (struct graphics_device *dev);
-static void directfb_remove_from_table (struct graphics_device *dev);
+static void directfb_add_to_table      (struct graphics_device *gd);
+static void directfb_remove_from_table (struct graphics_device *gd);
 static struct graphics_device * directfb_lookup_in_table (DFBWindowID  id);
 
 
@@ -99,7 +100,7 @@ directfb_init_driver (unsigned char *param, unsigned char *display)
 	directfb_driver.depth = 196;
 
   /* endian test */
-  if (big_endian) {
+  if (htons (0x1234) == 0x1234) {
      if ((directfb_driver.depth & 0x7) == 2)
 	directfb_driver.depth |= 0x100;
      if ((directfb_driver.depth & 0x7) == 4)
@@ -142,7 +143,7 @@ ret:
 static struct graphics_device *
 directfb_init_device (void)
 {
-  struct graphics_device *dev;
+  struct graphics_device *gd;
   DFBDeviceData          *data;
   IDirectFBWindow        *window;
   DFBWindowDescription    desc;
@@ -160,13 +161,13 @@ directfb_init_device (void)
     return NULL;
   }
 
-  dev = mem_alloc (sizeof (struct graphics_device));
+  gd = mem_alloc (sizeof (struct graphics_device));
 
-  dev->size.x1 = 0;
-  dev->size.y1 = 0;
-  window->GetSize (window, &dev->size.x2, &dev->size.y2);
+  gd->size.x1 = 0;
+  gd->size.y1 = 0;
+  window->GetSize (window, &gd->size.x2, &gd->size.y2);
 
-  dev->clip = dev->size;
+  gd->clip = gd->size;
 
   data = mem_alloc (sizeof (DFBDeviceData));
 
@@ -179,37 +180,37 @@ directfb_init_device (void)
   window->GetSurface (window, &data->surface);
   window->GetID (window, &data->id);
 
-  dev->driver_data = data;
-  dev->user_data   = NULL;
+  gd->driver_data = data;
+  gd->user_data   = NULL;
 
-  directfb_add_to_table (dev);
+  directfb_add_to_table (gd);
 
   window->AttachEventBuffer (window, events);
 
   window->SetOpacity (window, FOCUSED_OPACITY);
 
-  return dev;
+  return gd;
 }
 
 static void
-directfb_shutdown_device (struct graphics_device *dev)
+directfb_shutdown_device (struct graphics_device *gd)
 {
   DFBDeviceData *data;
 
-  if (!dev)
+  if (!gd)
     return;
 
-  data = dev->driver_data;
+  data = gd->driver_data;
 
   unregister_bottom_half (directfb_flip_surface, data);
-  directfb_remove_from_table (dev);
+  directfb_remove_from_table (gd);
 
   data->surface->Release (data->surface);
   data->window->Destroy (data->window);
   data->window->Release (data->window);
 
   mem_free (data);
-  mem_free (dev);
+  mem_free (gd);
 }
 
 static void
@@ -234,12 +235,17 @@ directfb_shutdown_driver (void)
   dfb = NULL;
 }
 
+static unsigned char *
+directfb_get_driver_param (void)
+{
+  return NULL;
+}
+
 static int
 directfb_get_empty_bitmap (struct bitmap *bmp)
 {
   IDirectFBSurface      *surface;
   DFBSurfaceDescription  desc;
-  int skip;
 
   bmp->data = bmp->flags = NULL;
 
@@ -250,13 +256,12 @@ directfb_get_empty_bitmap (struct bitmap *bmp)
   retry:
   if (dfb->CreateSurface (dfb, &desc, &surface) != DFB_OK) {
     if (out_of_memory(MF_GPI, NULL, 0))
-	goto retry;
+    	goto retry;
     return -1;
   }
 
-  surface->Lock (surface, (DFBSurfaceLockFlags)(DSLF_READ | DSLF_WRITE), &bmp->data, &skip);
+  surface->Lock (surface, (DFBSurfaceLockFlags)(DSLF_READ | DSLF_WRITE), &bmp->data, &bmp->skip);
 
-  bmp->skip = skip;
   bmp->flags = surface;
 
   return 0;
@@ -276,10 +281,9 @@ static void *
 directfb_prepare_strip (struct bitmap *bmp, int top, int lines)
 {
   IDirectFBSurface *surface = bmp->flags;
-  int skip;
   if (!surface) return NULL;
 
-  surface->Lock (surface, (DFBSurfaceLockFlags)(DSLF_READ | DSLF_WRITE), &bmp->data, &skip);
+  surface->Lock (surface, (DFBSurfaceLockFlags)(DSLF_READ | DSLF_WRITE), &bmp->data, &bmp->skip);
 
   return ((unsigned char *) bmp->data + top * bmp->skip);
 }
@@ -304,14 +308,15 @@ directfb_unregister_bitmap (struct bitmap *bmp)
 }
 
 static void
-directfb_draw_bitmap (struct graphics_device *dev, struct bitmap *bmp,
+directfb_draw_bitmap (struct graphics_device *gd, struct bitmap *bmp,
                       int x, int y)
 {
-  DFBDeviceData    *data = dev->driver_data;
+  DFBDeviceData    *data = gd->driver_data;
   IDirectFBSurface *src  = bmp->flags;
   if (!src) return;
 
-  CLIP_DRAW_BITMAP
+  if (gd->clip.x1 >= gd->clip.x2 ||
+      gd->clip.y1 >= gd->clip.y2) return;
 
   data->surface->Blit (data->surface, src, NULL, x, y);
 
@@ -325,26 +330,13 @@ directfb_get_color (int rgb)
 }
 
 
-static inline void directfb_set_color (IDirectFBSurface *surface, long color)
-{
-  surface->SetColor (surface,
-                     (color & 0xFF0000) >> 16,
-                     (color & 0xFF00)   >> 8,
-                     (color & 0xFF),
-                     0xFF);
-}
-
 static void
-directfb_fill_area (struct graphics_device *dev,
+directfb_fill_area (struct graphics_device *gd,
                     int x1, int y1, int x2, int y2, long color)
 {
-  DFBDeviceData *data = dev->driver_data;
-  int w, h;
-
-  CLIP_FILL_AREA
-
-  w = x2 - x1;
-  h = y2 - y1;
+  DFBDeviceData *data = gd->driver_data;
+  int w = x2 - x1;
+  int h = y2 - y1;
 
   directfb_set_color (data->surface, color);
   data->surface->FillRectangle (data->surface, x1, y1, w, h);
@@ -353,70 +345,144 @@ directfb_fill_area (struct graphics_device *dev,
 }
 
 static void
-directfb_draw_hline (struct graphics_device *dev,
-                     int x1, int y, int x2, long color)
+directfb_draw_hline (struct graphics_device *gd,
+                     int left, int y, int right, long color)
 {
-  DFBDeviceData *data = dev->driver_data;
+  DFBDeviceData *data = gd->driver_data;
 
-  CLIP_DRAW_HLINE
+  if (right <= left) return;
 
-  x2--;
+  right--;
 
   directfb_set_color (data->surface, color);
-  data->surface->DrawLine (data->surface, x1, y, x2, y);
+  data->surface->DrawLine (data->surface, left, y, right, y);
 
-  directfb_register_flip (data, x1, y, x2 - x1, 1);
+  directfb_register_flip (data, left, y, right - left, 1);
 }
 
 static void
-directfb_draw_vline (struct graphics_device *dev,
-                     int x, int y1, int y2, long color)
+directfb_draw_vline (struct graphics_device *gd,
+                     int x, int top, int bottom, long color)
 {
-  DFBDeviceData *data = dev->driver_data;
+  DFBDeviceData *data = gd->driver_data;
 
-  CLIP_DRAW_VLINE
+  if (bottom <= top) return;
 
-  y2--;
+  bottom--;
 
   directfb_set_color (data->surface, color);
-  data->surface->DrawLine (data->surface, x, y1, x, y2);
+  data->surface->DrawLine (data->surface, x, top, x, bottom);
 
-  directfb_register_flip (data, x, y1, 1, y2 - y1);
+  directfb_register_flip (data, x, top, 1, bottom - top);
 }
 
 static void
-directfb_set_clip_area (struct graphics_device *dev)
+directfb_set_clip_area (struct graphics_device *gd, struct rect *r)
 {
-  DFBDeviceData *data   = dev->driver_data;
+  DFBDeviceData *data   = gd->driver_data;
   DFBRegion      region;
-
-  region.x1 = dev->clip.x1;
-  region.y1 = dev->clip.y1;
-  region.x2 = dev->clip.x2 - 1;
-  region.y2 = dev->clip.y2 - 1;
+  generic_set_clip_area(gd, r);
+  region.x1 = gd->clip.x1;
+  region.y1 = gd->clip.y1;
+  region.x2 = gd->clip.x2 - 1;
+  region.y2 = gd->clip.y2 - 1;
 
   data->surface->SetClip (data->surface, &region);
 }
 
 static int
-directfb_scroll (struct graphics_device *dev, struct rect_set **set, int scx, int scy)
+directfb_hscroll (struct graphics_device *gd, struct rect_set **set, int sc)
 {
-  DFBDeviceData *data = dev->driver_data;
+  DFBDeviceData *data = gd->driver_data;
   DFBRectangle   rect;
 
-  rect.x = dev->clip.x1;
-  rect.y = dev->clip.y1;
-  rect.w = dev->clip.x2 - rect.x;
-  rect.h = dev->clip.y2 - rect.y;
+  *set = NULL;
+  if (!sc)
+    return 0;
+
+  rect.x = gd->clip.x1;
+  rect.y = gd->clip.y1;
+  rect.w = gd->clip.x2 - rect.x;
+  rect.h = gd->clip.y2 - rect.y;
 
   data->surface->Blit (data->surface,
-                       data->surface, &rect, rect.x + scx, rect.y + scy);
+                       data->surface, &rect, rect.x + sc, rect.y);
 
   directfb_register_flip (data, rect.x, rect.y, rect.w, rect.h);
 
   return 1;
 }
 
+static int
+directfb_vscroll (struct graphics_device *gd, struct rect_set **set, int sc)
+{
+  DFBDeviceData *data = gd->driver_data;
+  DFBRectangle   rect;
+
+  *set = NULL;
+  if (!sc)
+    return 0;
+
+  rect.x = gd->clip.x1;
+  rect.y = gd->clip.y1;
+  rect.w = gd->clip.x2 - rect.x;
+  rect.h = gd->clip.y2 - rect.y;
+
+  data->surface->Blit (data->surface,
+                       data->surface, &rect, rect.x, rect.y + sc);
+
+  directfb_register_flip (data, rect.x, rect.y, rect.w, rect.h);
+
+  return 1;
+}
+
+struct graphics_driver directfb_driver =
+{
+  cast_uchar "directfb",
+  directfb_init_driver,
+  directfb_init_device,
+  directfb_shutdown_device,
+  directfb_shutdown_driver,
+  dummy_emergency_shutdown,
+  directfb_get_driver_param,
+  NULL,
+  NULL,
+  directfb_get_empty_bitmap,
+  directfb_register_bitmap,
+  directfb_prepare_strip,
+  directfb_commit_strip,
+  directfb_unregister_bitmap,
+  directfb_draw_bitmap,
+  directfb_get_color,
+  directfb_fill_area,
+  directfb_draw_hline,
+  directfb_draw_vline,
+  directfb_hscroll,
+  directfb_vscroll,
+  directfb_set_clip_area,
+  dummy_block,
+  dummy_unblock,
+  NULL,	 /*  set_title  */
+  NULL,	 /*  exec	*/
+  NULL,	 /*  set_clipboard_text */
+  NULL,	 /*  get_clipboard_text */
+  0,	 /*  depth      */
+  0, 0,	 /*  size       */
+  GD_NO_OS_SHELL, /*  flags      */
+  0,     /*  codepage   */
+  NULL,	 /*  shell	*/
+};
+
+
+static inline void directfb_set_color (IDirectFBSurface *surface, long color)
+{
+  surface->SetColor (surface,
+                     (color & 0xFF0000) >> 16,
+                     (color & 0xFF00)   >> 8,
+                     (color & 0xFF),
+                     0xFF);
+}
+
 static void directfb_register_flip (DFBDeviceData *data,
                                     int x, int y, int w, int h)
 {
@@ -459,18 +525,10 @@ directfb_flip_surface (void *pointer)
   data->flip_pending = 0;
 }
 
-static void
-directfb_flush(struct graphics_device *dev)
-{
-  DFBDeviceData *data = dev->driver_data;
-  unregister_bottom_half (directfb_flip_surface, data);
-  directfb_flip_surface(data);
-}
-
 static void
 directfb_check_events (void *pointer)
 {
-  struct graphics_device *dev   = NULL;
+  struct graphics_device *gd   = NULL;
   DFBDeviceData          *data = NULL;
   DFBWindowEvent          event;
   DFBWindowEvent          next;
@@ -493,14 +551,14 @@ directfb_check_events (void *pointer)
           continue;
         }
 
-      if (!dev || data->id != event.window_id)
+      if (!data || data->id != event.window_id)
         {
-          dev = directfb_lookup_in_table (event.window_id);
-          if (!dev)
+          gd = directfb_lookup_in_table (event.window_id);
+          if (!gd)
             continue;
         }
 
-      data = dev->driver_data;
+      data = gd->driver_data;
 
       switch (event.type)
         {
@@ -521,9 +579,9 @@ directfb_check_events (void *pointer)
                  (next.window_id == data->id))
             events->GetEvent (events, DFB_EVENT (&event));
 
-          dev->size.x2 = event.w;
-          dev->size.y2 = event.h;
-          dev->resize_handler (dev);
+          gd->size.x2 = event.w;
+          gd->size.y2 = event.h;
+          gd->resize_handler (gd);
           break;
 
         case DWET_KEYDOWN:
@@ -532,7 +590,7 @@ directfb_check_events (void *pointer)
 
             directfb_translate_key (&event, &key, &flag);
             if (key)
-              dev->keyboard_handler (dev, key, flag);
+              gd->keyboard_handler (gd, key, flag);
           }
           break;
 
@@ -575,12 +633,12 @@ directfb_check_events (void *pointer)
                 continue;
               }
 
-            dev->mouse_handler (dev, event.x, event.y, flags);
+            gd->mouse_handler (gd, event.x, event.y, flags);
           }
           break;
 
         case DWET_WHEEL:
-          dev->mouse_handler (dev, event.x, event.y,
+          gd->mouse_handler (gd, event.x, event.y,
                              B_MOVE |
                              (event.step > 0 ? B_WHEELUP : B_WHEELDOWN));
 	  break;
@@ -610,12 +668,12 @@ directfb_check_events (void *pointer)
                 break;
               }
 
-            dev->mouse_handler (dev, event.x, event.y, flags);
+            gd->mouse_handler (gd, event.x, event.y, flags);
           }
           break;
 
         case DWET_CLOSE:
-          dev->keyboard_handler (dev, KBD_CLOSE, 0);
+          gd->keyboard_handler (gd, KBD_CLOSE, 0);
           break;
 
         default:
@@ -685,9 +743,9 @@ directfb_translate_key (DFBWindowEvent *event, int *key, int *flag)
 }
 
 static void
-directfb_add_to_table (struct graphics_device *dev)
+directfb_add_to_table (struct graphics_device *gd)
 {
-  DFBDeviceData           *data = dev->driver_data;
+  DFBDeviceData           *data = gd->driver_data;
   struct graphics_device **devices;
   int i;
 
@@ -704,13 +762,13 @@ directfb_add_to_table (struct graphics_device *dev)
 
       if ((unsigned)c > MAXINT / sizeof(void *) - 1) overalloc();
       devices = mem_realloc (devices, (c + 1) * sizeof (void *));
-      devices[c-1] = dev;
+      devices[c-1] = gd;
       devices[c]   = NULL;
     }
   else
     {
       devices = mem_alloc (2 * sizeof (void *));
-      devices[0] = dev;
+      devices[0] = gd;
       devices[1] = NULL;
     }
 
@@ -718,9 +776,9 @@ directfb_add_to_table (struct graphics_device *dev)
 }
 
 static void
-directfb_remove_from_table (struct graphics_device *dev)
+directfb_remove_from_table (struct graphics_device *gd)
 {
-  DFBDeviceData           *data = dev->driver_data;
+  DFBDeviceData           *data = gd->driver_data;
   struct graphics_device **devices;
   int i, j, c;
 
@@ -731,7 +789,7 @@ directfb_remove_from_table (struct graphics_device *dev)
     return;
 
   for (j = 0, c = -1; devices[j]; j++)
-    if (devices[j] == dev)
+    if (devices[j] == gd)
       c = j;
 
   if (c < 0)
@@ -768,44 +826,4 @@ directfb_lookup_in_table (DFBWindowID id)
   return NULL;
 }
 
-struct graphics_driver directfb_driver =
-{
-  cast_uchar "directfb",
-  directfb_init_driver,
-  directfb_init_device,
-  directfb_shutdown_device,
-  directfb_shutdown_driver,
-  NULL,
-  NULL,
-  NULL,
-  NULL,
-  NULL,
-  NULL,
-  directfb_get_empty_bitmap,
-  directfb_register_bitmap,
-  directfb_prepare_strip,
-  directfb_commit_strip,
-  directfb_unregister_bitmap,
-  directfb_draw_bitmap,
-  directfb_get_color,
-  directfb_fill_area,
-  directfb_draw_hline,
-  directfb_draw_vline,
-  directfb_scroll,
-  directfb_set_clip_area,
-  directfb_flush,
-  NULL,	 /*  block */
-  NULL,	 /*  unblock */
-  NULL,	 /*  set_palette */
-  NULL,	 /*  get_real_colors */
-  NULL,	 /*  set_title  */
-  NULL,	 /*  exec	*/
-  NULL,	 /*  set_clipboard_text */
-  NULL,	 /*  get_clipboard_text */
-  0,	 /*  depth      */
-  0, 0,	 /*  size       */
-  GD_UNICODE_KEYS | GD_NO_OS_SHELL | GD_NOAUTO, /*  flags      */
-  NULL,	 /*  param	*/
-};
-
 #endif /* GRDRV_DIRECTFB */
