@@ -6,7 +6,6 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <fcntl.h>
-#include <string.h>
 #include <ctype.h>
 #include <png.h>
 
@@ -43,13 +42,14 @@ int file_select(const struct dirent * entry)
 {
 
 	const char *s=entry->d_name;
-	char *ext;
-	unsigned long int codepoint;
 	FILE *f;
 
-	codepoint= strtoul(s, &ext, 16);
-	if (codepoint > 0x10FFFFUL || strcasecmp(ext, ".png"))
-		return 0;
+	if (strlen(s)!=8) return 0;
+	if (!((s[0]>='0'&&s[0]<='9')||(tolower(s[0])>='a'&&tolower(s[0])<='f'))) return 0;
+	if (!((s[1]>='0'&&s[1]<='9')||(tolower(s[1])>='a'&&tolower(s[1])<='f'))) return 0;
+	if (!((s[2]>='0'&&s[2]<='9')||(tolower(s[2])>='a'&&tolower(s[2])<='f'))) return 0;
+	if (!((s[3]>='0'&&s[3]<='9')||(tolower(s[3])>='a'&&tolower(s[3])<='f'))) return 0;
+	if (s[4]!='.'||tolower(s[5])!='p'||tolower(s[6])!='n'||tolower(s[7])!='g') return 0;
 	f=fopen(s,"r");
 	if (!f) return 0; /* If it can't be open as a file then it's probably
 		           * a directory
@@ -58,24 +58,6 @@ int file_select(const struct dirent * entry)
 	return 1;
 }
 
-int hexstrsort(const struct dirent **a, const struct dirent **b)
-{
-	const char *aname, *bname;
-	int diff;
-
-	for( aname= (**a).d_name; *aname == '0'; ++aname );
-	for( bname= (**b).d_name; *bname == '0'; ++bname );
-	diff= (int)strlen(aname) - (int)strlen(bname);
-	if( diff ) return diff;
-	while( *aname ) {
-		diff= (int)tolower(*aname) - (int)tolower(*bname);
-		if( diff ) return diff;
-		++aname;
-		++bname;
-	}
-	return 0;
-}
-
 /* If the directory name has bad structure (must have 4 dashes in the name),
  * then nothing is done
  * Fills weight, slant, adstyl, spacing with appropriate strings
@@ -265,8 +247,8 @@ two:
 }
 
 
-/* name:		path to file to process
- * output:		where to put the C code
+/* name:        path to file to process
+ * output:      where to put the C code
  * char_number: number of the char. -1 means do not write anything into the tables.
 */
 void process_file(unsigned char *name, FILE *output, int char_number)
@@ -275,6 +257,7 @@ void process_file(unsigned char *name, FILE *output, int char_number)
 	int c;
 	int letter_code=letter_code;
 	int count=0;
+	unsigned char btr[5];
 	int forbidden_0_to_7;
 
 	f=fopen((const char *)name,"r");
@@ -285,18 +268,20 @@ void process_file(unsigned char *name, FILE *output, int char_number)
 		exit(1);
 	}
 	if (char_number>=0){
-		letter_code=strtoul((const char *)name,NULL,16);
+		memcpy(btr,name,4);
+		btr[4]=0;
+		letter_code=strtoul((const char *)btr,NULL,16);
 		fonts[n_fonts-1].letters[char_number].begin=stamp;
 	}
 	char_pos = 38;
-	fprintf(output,"static_const unsigned char letter_%d[] = \"", stamp);
+	fprintf(output,"static const unsigned char letter_%d[] = \"", stamp);
 	forbidden_0_to_7=0;
 
 	while(EOF!=(c=fgetc(f))){
 		forbidden_0_to_7=print_char(output,c,forbidden_0_to_7);
 		count++;
 	}
-	fprintf(output,"\";\n\n");
+	fprintf(output,"\";\n");
 	stamp++;
 	rewind(f);
 	fonts[n_fonts-1].letters[char_number].length=count;
@@ -312,7 +297,7 @@ void process_letters(FILE *output)
 	struct dirent **ptr;
 	int nr,a;
 
-	nr=scandir(".",&namelist,file_select,hexstrsort);
+	nr=scandir(".",&namelist,file_select,alphasort);
 	if (nr<0){
 		perror(PROGNAME);
 		exit(1);
@@ -349,12 +334,12 @@ int dir_select(const struct dirent *directory_entry)
 /* build_font_table:
  * scans the font directory
  * for each directory:
- *	* renews font description structures
- *	* scans the images
- *	* builds a binary search array
- *	* writes the font structure into the ../font_inc.c
- *	* releases font description structures
- *	Also prints the font_data byte table.
+ *  * renews font description structures
+ *  * scans the images
+ *  * builds a binary search array
+ *  * writes the font structure into the ../font_inc.c
+ *  * releases font description structures
+ *  Also prints the font_data byte table.
  */
 void build_font_table(FILE *output)
 {
@@ -431,6 +416,7 @@ void build_font_table(FILE *output)
 		}
 	}
 	free(namelist);
+	fprintf(output,"\n");
 }
 
 void print_string(unsigned char * ptr, FILE * output)
@@ -446,7 +432,7 @@ void print_string(unsigned char * ptr, FILE * output)
 
 void print_letter(struct letter *p, FILE * output)
 {
-	fprintf(output,"\t{ letter_%d, 0x%08x, 0x%08x,% 4d,% 4d, NULL }",p->begin,p->length,
+	fprintf(output,"{letter_%d,0x%08x,0x%08x,% 5d,% 5d, 0, 0}",p->begin,p->length,
 		p->code, p->xsize, p->ysize);
 }
 
@@ -455,25 +441,28 @@ void print_font(int a, FILE * output)
 {
 	struct font *f=fonts+a;
 
-	fprintf(output,"\t{\n\t\t(unsigned char *)");
+	fprintf(output,"{\n(unsigned char *)");
 	print_string(f->family,output);
-	fprintf(output,",\n\t\t(unsigned char *)");
+	fprintf(output,",\n(unsigned char *)");
 	print_string(f->weight,output);
-	fprintf(output,",\n\t\t(unsigned char *)");
+	fprintf(output,",\n(unsigned char *)");
 	print_string(f->slant,output);
-	fprintf(output,",\n\t\t(unsigned char *)");
+	fprintf(output,",\n(unsigned char *)");
 	print_string(f->adstyl,output);
-	fprintf(output,",\n\t\t(unsigned char *)");
+	fprintf(output,",\n(unsigned char *)");
 	print_string(f->spacing,output);
-	fprintf(output,",\n\t\t%d,\n\t\t%d,\n\t},\n",n_letters, f->n_letters);
+	fprintf(output,",\n");
+	fprintf(output,"%d,\n",n_letters);
+	fprintf(output,"%d,\n",f->n_letters);
 	n_letters+=f->n_letters;
+	fprintf(output,"},\n");
 }
 
 void print_font_table(FILE *output)
 {
 	int a;
 
-	fprintf(output,"struct letter letter_data[%d] = {\n",n_letters);
+	fprintf(output,"struct letter letter_data[%d]={\n",n_letters);
 	for (a=0;a<n_fonts;a++){
 		struct letter *ptr=fonts[a].letters;
 		int count=fonts[a].n_letters;
@@ -485,13 +474,13 @@ void print_font_table(FILE *output)
 		}
 	}
 	fprintf(output,"};\n\n");
-	fprintf(output,"struct font font_table[%d] = {\n",n_fonts);
+	fprintf(output,"struct font font_table[%d]={\n",n_fonts);
 	n_letters=0;
 	for (a=0;a<n_fonts;a++){
 		print_font(a,output);
 	}
 	fprintf(output,"};\n");
-	fprintf(output,"\nint n_fonts = %d;\n",n_fonts);
+	fprintf(output,"\nint n_fonts=%d;\n",n_fonts);
 }
 
 int main(int argc, char **argv)
