@@ -10,9 +10,6 @@
 
 #define NUMBER_OF_DEVICES	12
 
-static void grx_mouse_init(void);
-static void grx_mouse_terminate(void);
-
 void _GrCloseVideoDriver(void);
 
 extern struct graphics_driver grx_driver;
@@ -28,43 +25,23 @@ static unsigned char in_graphics_mode = 0;
 
 static void grx_set_palette(void)
 {
-	unsigned i, colors;
-	if (grx_driver.depth == 33)
-		colors = 16;
-	else if (grx_driver.depth == (65 | 0x300))
-		colors = 216;
-	else if (grx_driver.depth == 65)
-		colors = 256;
-	else
-		return;
-	GrResetColors();
-	for (i = 0; i < colors; i++) {
-		unsigned rgb[3];
-		q_palette(colors, i, 255, rgb);
-		GrSetColor(i, rgb[0], rgb[1], rgb[2]);
+	int i;
+	if (grx_driver.depth == 33) {
+		for (i = 0; i < 16; i++) {
+			GrSetColor(i,
+				i & 8 ? 255 : 0,
+				((i >> 1) & 3) * 85,
+				i & 1 ? 255 : 0);
+		}
+	}
+	if (grx_driver.depth == 65) {
+		for (i = 0; i < 256; i++) {
+			GrSetColor(i,
+				((i >> 5) & 7) * (255./7.),
+				((i >> 2) & 7) * (255./7.),
+				(i & 3) * 85);
+		}
 	}
-	return;
-}
-
-static void grx_update_palette(void)
-{
-	int new_depth;
-	if (!grx_driver.param->palette_mode)
-		new_depth = grx_driver.depth | 0x300;
-	else
-		new_depth = grx_driver.depth & ~0x300;
-
-	if (new_depth == grx_driver.depth)
-		return;
-
-	grx_driver.depth = new_depth;
-
-	grx_driver.get_color = get_color_fn(new_depth);
-	if (!grx_driver.get_color) internal_error("Unknown bit depth %x", new_depth);
-
-	grx_mouse_terminate();
-	grx_set_palette();
-	grx_mouse_init();
 }
 
 static void grx_set_text(void)
@@ -73,7 +50,9 @@ static void grx_set_text(void)
 		GrSetMode(GR_default_text);
 		in_graphics_mode = 0;
 		_GrCloseVideoDriver();
-		restore_terminal();
+#ifdef DOS
+		dos_restore_screen();
+#endif
 	}
 }
 
@@ -93,33 +72,31 @@ static void grx_set_clip(void)
 		);
 }
 
-static void grx_key_in(struct itrm *p, unsigned char *ev_, int size)
+static void grx_key_in(void *p, unsigned char *ev_, int size)
 {
 	int vd;
-	struct links_event *ev = (struct links_event *)(void *)ev_;
-	if (size != sizeof(struct links_event)) return;
+	struct event *ev = (struct event *)(void *)ev_;
+	if (size != sizeof(struct event)) return;
 	if (ev->ev == EV_ABORT) terminate_loop = 1;
 	if (ev->ev != EV_KBD) return;
-	if (ev->y & KBD_PASTING) goto skip;
-	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
+	if ((ev->y & KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
 		vd = (ev->x - '1' + 10) % 10;
 		goto switch_vd;
 	}
-	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x == '-') {
+	if ((ev->y & KBD_ALT) && ev->x == '-') {
 		vd = 10;
 		goto switch_vd;
 	}
-	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x == '=') {
+	if ((ev->y & KBD_ALT) && ev->x == '=') {
 		vd = 11;
 		goto switch_vd;
 	}
-	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x <= KBD_F1 && ev->x >= KBD_F12) {
+	if ((ev->y & KBD_ALT) && ev->x <= KBD_F1 && ev->x >= KBD_F12) {
 		vd = KBD_F1 - ev->x;
 		goto switch_vd;
 	}
-skip:
-	if (g_kbd_codepage(&grx_driver) != utf8_table && ev->x >= 128 && ev->x <= 255)
-		if ((ev->x = cp2u(ev->x, g_kbd_codepage(&grx_driver))) == -1) return;
+	if (grx_driver.codepage!=utf8_table&&(ev->x)>=128&&(ev->x)<=255)
+		if ((ev->x=cp2u(ev->x,grx_driver.codepage)) == -1) return;
 	if (current_virtual_device && current_virtual_device->keyboard_handler) current_virtual_device->keyboard_handler(current_virtual_device, ev->x, ev->y);
 	return;
 
@@ -155,7 +132,7 @@ static void grx_mouse_init(void)
 
 	grx_mouse_cursor = GrBuildCursor(cast_char cursor, arrow_width, arrow_width, arrow_height, 0, 0, table);
 
-	if (!grx_mouse_cursor) GrMouseSetColors(grx_driver.get_color(0xffffff), grx_driver.get_color(0x000000));
+	if (!grx_mouse_cursor) GrMouseSetColors(GrWhite(), GrBlack());
 	else GrMouseSetCursor(grx_mouse_cursor);
 
 	GrMouseDisplayCursor();
@@ -164,7 +141,6 @@ static void grx_mouse_init(void)
 static void grx_mouse_terminate(void)
 {
 	if (!grx_mouse_initialized) return;
-	GrMouseEraseCursor();
 	GrMouseUnInit();
 	if (grx_mouse_cursor) GrDestroyCursor(grx_mouse_cursor);
 	grx_mouse_initialized = 0;
@@ -211,32 +187,31 @@ get_another:
 static unsigned char *set_mode_according_to_param(unsigned char *param)
 {
 	unsigned char e[MAX_STR_LEN];
-	save_terminal();
+#ifdef DOS
+	dos_save_screen();
+#endif
 	if (!param || !*param) {
 		return stracpy(cast_uchar "There is no default videomode.\nUse the switch \"-mode WIDTHxHEIGHTxCOLORS\" to select video mode.\nThe COLORS argument may be 16, 256, 32k, 64k, 16M.\n");
-#if 0
 		if (!GrSetMode(GR_default_graphics))
 			return stracpy(cast_uchar "GrSetMode failed.\n");
 		in_graphics_mode = 1;
-#endif
 	} else {
 		unsigned long x, y;
 		int pl;
-		unsigned char *ptr = param;
-		char *end;
-		x = strtoul(cast_const_char ptr, &end, 10);
-		if (cast_uchar end == ptr || *end != 'x' || x >= MAXINT)
+		unsigned char *ptr = param, *end;
+		x = strtoul(cast_const_char ptr, (char **)(void *)&end, 10);
+		if (end == ptr || *end != 'x' || x >= MAXINT)
 			goto bad_param;
-		ptr = cast_uchar end + 1;
-		y = strtoul(cast_const_char ptr, &end, 10);
-		if (cast_uchar end == ptr || *end != 'x' || y >= MAXINT)
+		ptr = end + 1;
+		y = strtoul(cast_const_char ptr, (char **)(void *)&end, 10);
+		if (end == ptr || *end != 'x' || y >= MAXINT)
 			goto bad_param;
-		ptr = cast_uchar end + 1;
+		ptr = end + 1;
 		if (!strcmp(cast_const_char ptr, "16")) pl = 4;
 		else if (!strcmp(cast_const_char ptr, "256")) pl = 8;
-		else if (!casestrcmp(ptr, cast_uchar "32k") || !strcmp(cast_const_char ptr, "32768")) pl = 15;
-		else if (!casestrcmp(ptr, cast_uchar "64k") || !strcmp(cast_const_char ptr, "65536")) pl = 16;
-		else if (!casestrcmp(ptr, cast_uchar "16M") || !casestrcmp(ptr, cast_uchar "16M32") || !strcmp(cast_const_char ptr, "16777216")) pl = 24;
+		else if (!strcasecmp(cast_const_char ptr, "32k") || !strcmp(cast_const_char ptr, "32768")) pl = 15;
+		else if (!strcasecmp(cast_const_char ptr, "64k") || !strcmp(cast_const_char ptr, "65536")) pl = 16;
+		else if (!strcasecmp(cast_const_char ptr, "16M") || !strcasecmp(cast_const_char ptr, "16M32") || !strcmp(cast_const_char ptr, "16777216")) pl = 24;
 		else goto bad_param;
 		if (x == 800 && y == 600 && pl == 4) y = 599; /* a bug in GRX */
 		if (!GrSetMode(GR_width_height_bpp_graphics, (int)x, (int)y, pl)) {
@@ -284,16 +259,12 @@ unsupported_videomode:
 		grx_restore();
 		return stracpy(e);
 	}
-	grx_driver.flags &= ~GD_SELECT_PALETTE;
 	switch (mode->bpp) {
 		case 4:
 			grx_driver.depth = 33;
 			break;
 		case 8:
 			grx_driver.depth = 65;
-			if (!grx_driver.param->palette_mode)
-				grx_driver.depth |= 0x300;
-			grx_driver.flags |= GD_SELECT_PALETTE;
 			break;
 		case 15:
 			grx_driver.depth = 122;
@@ -314,7 +285,6 @@ unsupported_videomode:
 			goto unsupported_videomode;
 	}
 	grx_driver.get_color = get_color_fn(grx_driver.depth);
-	if (!grx_driver.get_color) internal_error("Unknown bit depth %x", grx_driver.depth);
 	grx_driver.x = mode->width;
 	grx_driver.y = mode->height;
 	if (mode->bpp >= 8) grx_driver.flags |= GD_DONT_USE_SCROLL;
@@ -325,7 +295,10 @@ unsupported_videomode:
 
 	grx_driver_param = stracpy(param ? param : cast_uchar "");
 
-	init_virtual_devices(&grx_driver, NUMBER_OF_DEVICES);
+	if (init_virtual_devices(&grx_driver, NUMBER_OF_DEVICES)) {
+		grx_restore();
+		return stracpy(cast_uchar "Allocation of virtual devices failed.\n");
+	}
 
 	svgalib_kbd = handle_svgalib_keyboard(grx_key_in);
 	grx_mouse_init();
@@ -359,7 +332,7 @@ static int grx_get_empty_bitmap(struct bitmap *bmp)
 	struct grx_bmp *b;
 
 	if (!in_graphics_mode)
-		internal_error("grx_get_empty_bitmap: not in graphics mode");
+		internal("grx_get_empty_bitmap: not in graphics mode");
 
 	b = mem_alloc_mayfail(sizeof(struct grx_bmp));
 	if (!b) goto fail0;
@@ -457,7 +430,6 @@ static void grx_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int
 	struct grx_bmp *b = bmp->flags;
 	if (!b) return;
 	TEST_INACTIVITY
-	CLIP_DRAW_BITMAP
 	/* note: GrImageDisplay has buggy clipping.
 	   It doesn't draw if the target x or y coordinate is one pixel
 	   before the end of clip area */
@@ -467,7 +439,6 @@ static void grx_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int
 static void grx_fill_area(struct graphics_device *dev, int x1, int y1, int x2, int y2, long color)
 {
 	TEST_INACTIVITY
-	CLIP_FILL_AREA
 	if (x1 >= x2 || y1 >= y2) return;
 	GrFilledBox(x1, y1, x2 - 1, y2 - 1, color);
 }
@@ -475,34 +446,56 @@ static void grx_fill_area(struct graphics_device *dev, int x1, int y1, int x2, i
 static void grx_draw_hline(struct graphics_device *dev, int x1, int y, int x2, long color)
 {
 	TEST_INACTIVITY
-	CLIP_DRAW_HLINE
+	if (x1 >= x2) return;
 	GrHLine(x1, x2 - 1, y, color);
 }
 
 static void grx_draw_vline(struct graphics_device *dev, int x, int y1, int y2, long color)
 {
 	TEST_INACTIVITY
-	CLIP_DRAW_VLINE
+	if (y1 >= y2) return;
 	GrVLine(x, y1, y2 - 1, color);
 }
 
-static int grx_scroll(struct graphics_device *dev, struct rect_set **ignore, int scx, int scy)
+static int grx_hscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
 {
 	TEST_INACTIVITY_0
-	GrBitBlt(NULL,
-		dev->clip.x1 + (scx >= 0 ? scx : 0),
-		dev->clip.y1 + (scy >= 0 ? scy : 0),
-		NULL,
-		dev->clip.x1 - (scx < 0 ? scx : 0),
-		dev->clip.y1 - (scy < 0 ? scy : 0),
-		dev->clip.x2 - (scx >= 0 ? scx : 0) - 1,
-		dev->clip.y2 - (scy >= 0 ? scy : 0) - 1,
-		GrWRITE);
+	if (!sc) return 0;
+	if (sc < 0) {
+		if (-sc > dev->clip.x2 - dev->clip.x1) return 1;
+		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1,
+			NULL, dev->clip.x1 - sc, dev->clip.y1,
+			dev->clip.x2 - 1, dev->clip.y2 - 1, GrWRITE);
+	} else {
+		if (sc > dev->clip.x2 - dev->clip.x1) return 1;
+		GrBitBlt(NULL, dev->clip.x1 + sc, dev->clip.y1,
+			NULL, dev->clip.x1, dev->clip.y1,
+			dev->clip.x2 - sc - 1, dev->clip.y2 - 1, GrWRITE);
+	}
 	return 1;
 }
 
-static void grx_set_clip_area(struct graphics_device *dev)
+static int grx_vscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
 {
+	TEST_INACTIVITY_0
+	if (!sc) return 0;
+	if (sc < 0) {
+		if (-sc > dev->clip.y2 - dev->clip.y1) return 1;
+		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1,
+			NULL, dev->clip.x1, dev->clip.y1 - sc,
+			dev->clip.x2 - 1, dev->clip.y2 - 1, GrWRITE);
+	} else {
+		if (sc > dev->clip.y2 - dev->clip.y1) return 1;
+		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1 + sc,
+			NULL, dev->clip.x1, dev->clip.y1,
+			dev->clip.x2 - 1, dev->clip.y2 - sc - 1, GrWRITE);
+	}
+	return 1;
+}
+
+static void grx_set_clip_area(struct graphics_device *dev, struct rect *r)
+{
+	generic_set_clip_area(dev, r);
 	if (dev != current_virtual_device) return;
 	grx_set_clip();
 }
@@ -521,7 +514,7 @@ static int grx_block(struct graphics_device *dev)
 static int grx_unblock(struct graphics_device *dev)
 {
 	if (current_virtual_device) {
-		internal_error("grx_unblock called without grx_block");
+		internal("grx_unblock called without grx_block");
 		return 0;
 	}
 	grx_set_graphics();
@@ -542,28 +535,26 @@ struct graphics_driver grx_driver = {
 	shutdown_virtual_device,
 	grx_shutdown_driver,
 	grx_emergency_shutdown,
-	NULL,
 	grx_get_driver_param,
 	NULL,
 	NULL,
-	NULL,
 	grx_get_empty_bitmap,
+	/*grx_get_filled_bitmap,*/
 	grx_register_bitmap,
 	grx_prepare_strip,
 	grx_commit_strip,
 	grx_unregister_bitmap,
 	grx_draw_bitmap,
+	/*grx_draw_bitmaps,*/
 	NULL,
 	grx_fill_area,
 	grx_draw_hline,
 	grx_draw_vline,
-	grx_scroll,
+	grx_hscroll,
+	grx_vscroll,
 	grx_set_clip_area,
-	NULL,
 	grx_block,
 	grx_unblock,
-	grx_update_palette, /* set_palette */
-	NULL, /* get_real_colors */
 	NULL, /* set_title */
 	NULL, /* exec */
 	NULL, /* set_clipboard_text */
@@ -571,7 +562,8 @@ struct graphics_driver grx_driver = {
 	0,				/* depth */
 	0, 0,				/* size */
 	GD_NEED_CODEPAGE,		/* flags */
-	NULL,				/* param */
+	0,				/* codepage */
+	NULL,				/* shell */
 };
 
 #endif
