@@ -45,11 +45,11 @@ static void get_align(unsigned char *attr, int *a)
 {
 	unsigned char *al;
 	if ((al = get_attr_val(attr, cast_uchar "align"))) {
-		if (!(casestrcmp(al, cast_uchar "left"))) *a = AL_LEFT;
-		if (!(casestrcmp(al, cast_uchar "right"))) *a = AL_RIGHT;
-		if (!(casestrcmp(al, cast_uchar "center"))) *a = AL_CENTER;
-		if (!(casestrcmp(al, cast_uchar "justify"))) *a = AL_BLOCK;
-		if (!(casestrcmp(al, cast_uchar "char"))) *a = AL_RIGHT; /* NOT IMPLEMENTED */
+		if (!(strcasecmp(cast_const_char al, "left"))) *a = AL_LEFT;
+		if (!(strcasecmp(cast_const_char al, "right"))) *a = AL_RIGHT;
+		if (!(strcasecmp(cast_const_char al, "center"))) *a = AL_CENTER;
+		if (!(strcasecmp(cast_const_char al, "justify"))) *a = AL_BLOCK;
+		if (!(strcasecmp(cast_const_char al, "char"))) *a = AL_RIGHT; /* NOT IMPLEMENTED */
 		mem_free(al);
 	}
 }
@@ -58,10 +58,10 @@ static void get_valign(unsigned char *attr, int *a)
 {
 	unsigned char *al;
 	if ((al = get_attr_val(attr, cast_uchar "valign"))) {
-		if (!(casestrcmp(al, cast_uchar "top"))) *a = VAL_TOP;
-		if (!(casestrcmp(al, cast_uchar "middle"))) *a = VAL_MIDDLE;
-		if (!(casestrcmp(al, cast_uchar "bottom"))) *a = VAL_BOTTOM;
-		if (!(casestrcmp(al, cast_uchar "baseline"))) *a = VAL_TOP; /* NOT IMPLEMENTED */
+		if (!(strcasecmp(cast_const_char al, "top"))) *a = VAL_TOP;
+		if (!(strcasecmp(cast_const_char al, "middle"))) *a = VAL_MIDDLE;
+		if (!(strcasecmp(cast_const_char al, "bottom"))) *a = VAL_BOTTOM;
+		if (!(strcasecmp(cast_const_char al, "baseline"))) *a = VAL_TOP; /* NOT IMPLEMENTED */
 		mem_free(al);
 	}
 }
@@ -71,11 +71,11 @@ static void get_c_width(unsigned char *attr, int *w, int sh)
 	unsigned char *al;
 	if ((al = get_attr_val(attr, cast_uchar "width"))) {
 		if (*al && al[strlen(cast_const_char al) - 1] == '*') {
-			char *end;
+			unsigned char *en;
 			unsigned long n;
 			al[strlen(cast_const_char al) - 1] = 0;
-			n = strtoul(cast_const_char al, &end, 10);
-			if (n < 10000 && !*end) *w = W_REL - (int)n;
+			n = strtoul(cast_const_char al, (char **)(void *)&en, 10);
+			if (n < 10000 && !*en) *w = W_REL - (int)n;
 		} else {
 			int p = get_width(attr, cast_uchar "width", sh);
 			if (p >= 0) *w = p;
@@ -164,7 +164,7 @@ struct table {
 
 
 #ifdef DEBUG
-#define CELL(t, x, y) (((x) < 0 || (x) >= (t)->rx || (y) < 0 || (y) >= (t)->ry) ? (internal_error("accessing cell out of table (%d,%d) - limit (%d,%d)", (x), (y), (t)->rx, (t)->ry), (t)->cells) : &(t)->cells[(y) * (t)->rx + (x)])
+#define CELL(t, x, y) (((x) < 0 || (x) >= (t)->rx || (y) < 0 || (y) >= (t)->ry) ? (internal("accessing cell out of table (%d,%d) - limit (%d,%d)", (x), (y), (t)->rx, (t)->ry), (t)->cells) : &(t)->cells[(y) * (t)->rx + (x)])
 #else
 #define CELL(t, x, y) (&(t)->cells[(y) * (t)->rx + (x)])
 #endif
@@ -221,8 +221,8 @@ static void free_table(struct table *t)
 		struct table_cell *c = CELL(t, i, j);
 		if (c->tag) mem_free(c->tag);
 #ifdef G
-		if (c->root) c->root->go.destruct(&c->root->go);
-		if (c->tag_object) c->tag_object->go.destruct(&c->tag_object->go);
+		if (c->root) c->root->destruct(c->root);
+		if (c->tag_object) c->tag_object->destruct((struct g_object *)c->tag_object);
 		if (c->brd) mem_free(c->brd);
 #endif
 	}
@@ -377,7 +377,7 @@ unsigned char *skip_element(unsigned char *html, unsigned char *eof, unsigned ch
 	int namelen;
 	r:
 	while (html < eof && (*html != '<')) rr:html++;
-	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
+	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto r;
 	}
@@ -439,7 +439,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		if (lbhp) (*bad_html)[*bhp-1].e = html;
 		goto scan_done;
 	}
-	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
+	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto se;
 	}
@@ -452,10 +452,8 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		goto see;
 	}
 	if (t_namelen == 6 && !casecmp(t_name, cast_uchar "SCRIPT", 5)) {
-		if (should_skip_script(t_attr)) {
-			en = skip_element(en, eof, cast_uchar "SCRIPT", 0);
-			goto see;
-		}
+		en = skip_element(en, eof, cast_uchar "SCRIPT", 0);
+		goto see;
 	}
 	if (t_namelen == 6 && !casecmp(t_name, cast_uchar "/TABLE", 6)) {
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
@@ -564,11 +562,8 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	if ((a = get_attr_val(t_attr, cast_uchar "id")))
 		cell->tag = a;
 	if ((a = get_attr_val(t_attr, cast_uchar "class"))) {
-		if (strstr(cast_const_char a, "blob-code-inner") ||	/* github hack */
-		    !strcmp(cast_const_char a, "changelog") ||
-		    !strcmp(cast_const_char a, "vc_file_line_text") || /* https://gcc.gnu.org/viewcvs/gcc/trunk/gcc/opts.c?view=markup */
-		    !strcmp(cast_const_char a, "FileContents-lineContents") || /* https://android.googlesource.com/ */
-		    0) {
+		if (!strncmp(cast_const_char a, "blob-code", 9) ||	/* github hack */
+		    !strcmp(cast_const_char a, "changelog")) {
 			cell->align = !par_format.implicit_pre_wrap ? AL_NO : AL_NO_BREAKABLE;
 		}
 		mem_free(a);
@@ -612,7 +607,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 			if (!(sc = new_cell(t, x + k, y + j)) || sc->used) {
 				int l, m;
 				if (sc->mx == x && sc->my == y) continue;
-				/*internal_error("boo");*/
+				/*internal("boo");*/
 				for (l = 0; l < k; l++) memset(CELL(t, x + l, y + j), 0, sizeof(struct table_cell));
 				rsp = j;
 				for (l = 0; l < i; l++) for (m = 0; m < j; m++) CELL(t, x + l, y + m)->rowspan = j;
@@ -671,7 +666,7 @@ static void get_cell_width(struct table *t, struct table_cell *c, int w, int a,
 #endif
 	}
 	/*debug("get_cell_width: %d < %d", *min, *max);*/
-	/*if (min && max && *min > *max) internal_error("get_cell_width: %d > %d", *min, *max);*/
+	/*if (min && max && *min > *max) internal("get_cell_width: %d > %d", *min, *max);*/
 }
 
 #define g_c_w(cc)							\
@@ -708,7 +703,7 @@ static void dst_width(int *p, int n, int w, int *lim)
 		if (lim && p[i] > lim[i]) w = safe_add(w, p[i] - lim[i]), p[i] = lim[i];
 	}
 	if (w) {
-		/*if (!lim) internal_error("bug in dst_width");*/
+		/*if (!lim) internal("bug in dst_width");*/
 		lim = NULL;
 		s = 0;
 		goto again;
@@ -824,7 +819,7 @@ static int get_column_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (c->spanned || !c->used) continue;
 			if (c->colspan + i > t->x) {
-				/*internal_error("colspan out of table");
+				/*internal("colspan out of table");
 				return -1;*/
 				continue;
 			}
@@ -882,7 +877,7 @@ static void get_table_width(struct table *t)
 	}
 	t->min_t = min;
 	t->max_t = max;
-	/*if (min > max) internal_error("min(%d) > max(%d)", min, max);*/
+	/*if (min > max) internal("min(%d) > max(%d)", min, max);*/
 }
 
 static void distribute_widths(struct table *t, int width)
@@ -896,7 +891,7 @@ static void distribute_widths(struct table *t, int width)
 	t->rw = 0;
 	if (!t->x) return;
 	if (d < 0) {
-		/*internal_error("too small width %d, required %d", width, t->min_t);*/
+		/*internal("too small width %d, required %d", width, t->min_t);*/
 		return;
 	}
 	for (i = 0; i < t->x; i++) if (t->max_c[i] > mmax_c) mmax_c = t->max_c[i];
@@ -949,7 +944,7 @@ static void distribute_widths(struct table *t, int width)
 					w[i] = 1, mx[i] = MAXINT;
 					break;
 				default:
-					/*internal_error("could not expand table");*/
+					/*internal("could not expand table");*/
 					goto end2;
 			}
 			p = safe_add(p, w[i]);
@@ -977,7 +972,7 @@ static void distribute_widths(struct table *t, int width)
 			d -= t->w_c[mii] - q;
 			while (d < 0) t->w_c[mii]--, d++;
 			if (t->w_c[mii] < q) {
-				/*internal_error("shrinking cell");*/
+				/*internal("shrinking cell");*/
 				t->w_c[mii] = q;
 			}
 			wq = 1;
@@ -1009,7 +1004,7 @@ static void check_table_widths(struct table *t)
 		if (c->x_width > p) {
 			/*int min, max;
 			get_cell_width(t, c, 0, 0, &min, &max, NULL);
-			internal_error("cell is now wider (%d > %d) min = %d, max = %d, now_min = %d, now_max = %d", c->x_width, p, t->min_c[i], t->max_c[i], min, max);*/
+			internal("cell is now wider (%d > %d) min = %d, max = %d, now_min = %d, now_max = %d", c->x_width, p, t->min_c[i], t->max_c[i], min, max);*/
 			/* sbohem, internale. chytl jsi mi spoustu chyb v tabulkovaci, ale ted je proste cas jit ... ;-( */
 			c->x_width = p;
 		}
@@ -1021,7 +1016,7 @@ static void check_table_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (!c->start) continue;
 			if (c->colspan + i > t->x) {
-				/*internal_error("colspan out of table");*/
+				/*internal("colspan out of table");*/
 				mem_free(w);
 				return;
 			}
@@ -1036,7 +1031,7 @@ static void check_table_widths(struct table *t)
 					c = 0;
 					for (l = i; l < i + s; l++) if (w[l] < t->w_c[k]) w[l] = safe_add(w[l], 1), w[k]--, c = 1;
 					if (w[k] > t->w_c[k]) {
-						if (!c) internal_error("can't shrink cell");
+						if (!c) internal("can't shrink cell");
 						else goto ag;
 					}
 				}*/
@@ -1054,7 +1049,7 @@ static void check_table_widths(struct table *t)
 		}*/
 	}
 	if (ns > s) {
-		/*internal_error("new width(%d) is larger than previous(%d)", ns, s);*/
+		/*internal("new width(%d) is larger than previous(%d)", ns, s);*/
 		mem_free(w);
 		return;
 	}
@@ -1114,7 +1109,7 @@ static void get_table_heights(struct table *t)
 #ifdef G
 			} else {
 				if (!(gp = g_format_html_part(cell->start, cell->end, cell->align, 0, xw, NULL, cell->link_num, NULL /* FIX: background image */, cell->bgcolor_str, t->gp->data))) return;
-				/*if (gp->root->xw > xw) internal_error("vono to neumi formatovat... buuuu (%d > %d)", gp->root->xw, xw);*/
+				/*if (gp->root->xw > xw) internal("vono to neumi formatovat... buuuu (%d > %d)", gp->root->xw, xw);*/
 				cell->root = gp->root;
 				gp->root = NULL;
 				cell->height = gp->y;
@@ -1404,11 +1399,7 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	if (!F && !border) cellsp = 0;
 	else if (!F && !cellsp) cellsp = 1;
 	/* Sparc gcc-2.7.2.1 miscompiles this */
-#ifdef __GNUC__
-#if __GNUC__ == 2
 	do_not_optimize_here(&cellsp);
-#endif
-#endif
 	if (!F && border > 2) border = 2;
 	if (!F && cellsp > 2) cellsp = 2;
 #ifdef G
@@ -1431,31 +1422,31 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		}
 	}
 	if ((al = get_attr_val(attr, cast_uchar "align"))) {
-		if (!casestrcmp(al, cast_uchar "left")) align = AL_LEFT;
-		if (!casestrcmp(al, cast_uchar "center")) align = AL_CENTER;
-		if (!casestrcmp(al, cast_uchar "right")) align = AL_RIGHT;
+		if (!strcasecmp(cast_const_char al, "left")) align = AL_LEFT;
+		if (!strcasecmp(cast_const_char al, "center")) align = AL_CENTER;
+		if (!strcasecmp(cast_const_char al, "right")) align = AL_RIGHT;
 		mem_free(al);
 	}
 	frame = F_BOX;
 	if ((al = get_attr_val(attr, cast_uchar "frame"))) {
-		if (!casestrcmp(al, cast_uchar "void")) frame = F_VOID;
-		if (!casestrcmp(al, cast_uchar "above")) frame = F_ABOVE;
-		if (!casestrcmp(al, cast_uchar "below")) frame = F_BELOW;
-		if (!casestrcmp(al, cast_uchar "hsides")) frame = F_HSIDES;
-		if (!casestrcmp(al, cast_uchar "vsides")) frame = F_VSIDES;
-		if (!casestrcmp(al, cast_uchar "lhs")) frame = F_LHS;
-		if (!casestrcmp(al, cast_uchar "rhs")) frame = F_RHS;
-		if (!casestrcmp(al, cast_uchar "box")) frame = F_BOX;
-		if (!casestrcmp(al, cast_uchar "border")) frame = F_BOX;
+		if (!strcasecmp(cast_const_char al, "void")) frame = F_VOID;
+		if (!strcasecmp(cast_const_char al, "above")) frame = F_ABOVE;
+		if (!strcasecmp(cast_const_char al, "below")) frame = F_BELOW;
+		if (!strcasecmp(cast_const_char al, "hsides")) frame = F_HSIDES;
+		if (!strcasecmp(cast_const_char al, "vsides")) frame = F_VSIDES;
+		if (!strcasecmp(cast_const_char al, "lhs")) frame = F_LHS;
+		if (!strcasecmp(cast_const_char al, "rhs")) frame = F_RHS;
+		if (!strcasecmp(cast_const_char al, "box")) frame = F_BOX;
+		if (!strcasecmp(cast_const_char al, "border")) frame = F_BOX;
 		mem_free(al);
 	}
 	rules = border ? R_ALL : R_NONE;
 	if ((al = get_attr_val(attr, cast_uchar "rules"))) {
-		if (!casestrcmp(al, cast_uchar "none")) rules = R_NONE;
-		if (!casestrcmp(al, cast_uchar "groups")) rules = R_GROUPS;
-		if (!casestrcmp(al, cast_uchar "rows")) rules = R_ROWS;
-		if (!casestrcmp(al, cast_uchar "cols")) rules = R_COLS;
-		if (!casestrcmp(al, cast_uchar "all")) rules = R_ALL;
+		if (!strcasecmp(cast_const_char al, "none")) rules = R_NONE;
+		if (!strcasecmp(cast_const_char al, "groups")) rules = R_GROUPS;
+		if (!strcasecmp(cast_const_char al, "rows")) rules = R_ROWS;
+		if (!strcasecmp(cast_const_char al, "cols")) rules = R_COLS;
+		if (!strcasecmp(cast_const_char al, "all")) rules = R_ALL;
 		mem_free(al);
 	}
 	if (!border) frame = F_VOID;
@@ -1552,7 +1543,7 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		goto ret2;
 	}
 
-	n = list_struct(p->data->nodes.next, struct node);
+	n = p->data->nodes.next;
 	n->yw = p->yp - n->y + p->cy;
 	display_complicated_table(t, x, p->cy, &cye);
 	display_table_frames(t, x, p->cy);
@@ -1561,7 +1552,7 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	nn->y = safe_add(p->yp, cye);
 	nn->xw = n->xw;
 	add_to_list(p->data->nodes, nn);
-	/*if (p->cy + t->rh != cye) internal_error("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);*/
+	/*if (p->cy + t->rh != cye) internal("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);*/
 	p->cy = cye;
 
 	ret2:
@@ -1601,15 +1592,15 @@ static void add_to_cell_sets(struct table_cell ****s, int **nn, int *n, struct r
 {
 	int i, j;
 	if (r->y1 < 0 || r->y2 < 0)
-		internal_error("add_to_cell_sets: integer overflow: %d, %d", r->y1, r->y2);
+		internal("add_to_cell_sets: integer overflow: %d, %d", r->y1, r->y2);
 	for (i = r->y1 >> RECT_BOUND_BITS; i <= (r->y2 - 1) >> RECT_BOUND_BITS; i++) {
 		if (i >= *n) {
 			struct table_cell ***ns;
 			int *nnn;
 			if ((unsigned)i > MAXINT / sizeof(struct table_cell ***) - 1) overalloc();
 			if ((unsigned)i > MAXINT / sizeof(int *) - 1) overalloc();
-			ns = mem_realloc(*s, (i + 1) * sizeof(struct table_cell **));
-			nnn = mem_realloc(*nn, (i + 1) * sizeof(int));
+			ns = mem_realloc(*s, (i + 1) * sizeof(struct table_cell ***));
+			nnn = mem_realloc(*nn, (i + 1) * sizeof(int *));
 			for (j = *n; j < i + 1; j++) ns[j] = DUMMY, nnn[j] = 0;
 			*s = ns;
 			*nn = nnn;
@@ -1626,14 +1617,13 @@ static void add_to_cell_sets(struct table_cell ****s, int **nn, int *n, struct r
 	}
 }
 
-static void table_mouse_event(struct f_data_c *fd, struct g_object *o_, int x, int y, int b)
+static void table_mouse_event(struct f_data_c *fd, struct g_object_table *o, int x, int y, int b)
 {
-	struct g_object_table *o = get_struct(o_, struct g_object_table, go);
 	struct table *t = o->t;
 	int i, j;
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
 		struct table_cell *c = CELL(t, i, j);
-		if (c->root) if (!g_forward_mouse(fd, &c->root->go, x, y, b)) return;
+		if (c->root) if (!g_forward_mouse(fd, (struct g_object *)c->root, x, y, b)) return;
 	}
 }
 
@@ -1655,9 +1645,8 @@ static void draw_rect_sets(struct graphics_device *dev, struct background *bg, s
 	}
 }
 
-static void table_draw(struct f_data_c *fd, struct g_object *o_, int x, int y)
+static void table_draw(struct f_data_c *fd, struct g_object_table *o, int x, int y)
 {
-	struct g_object_table *o = get_struct(o_, struct g_object_table, go);
 	static tcount dgen = 0;
 	tcount my_dgen = ++dgen;
 	int i, j;
@@ -1683,33 +1672,31 @@ static void table_draw(struct f_data_c *fd, struct g_object *o_, int x, int y)
 			clip.y2 = clip.y2 + y;
 			if (!do_rects_intersect(&clip, &dev->clip)) continue;
 			draw_rect_set(dev, c->root->bg, c->brd, x, y);
-			restrict_clip_area(dev, &clip, x + c->root->go.x, y + c->root->go.y, x + c->root->go.x + c->root->go.xw /*c->g_width*/, y + c->root->go.y + c->root->go.yw);
-			c->root->go.draw(fd, &c->root->go, x + c->root->go.x, y + c->root->go.y);
-			set_clip_area(dev, &clip);
+			restrict_clip_area(dev, &clip, x + c->root->x, y + c->root->y, x + c->root->x + c->root->xw /*c->g_width*/, y + c->root->y + c->root->yw);
+			c->root->draw(fd, c->root, x + c->root->x, y + c->root->y);
+			drv->set_clip_area(dev, &clip);
 		}
 	}
 	draw_rect_sets(dev, t->bg, t->r_bg, t->nr_bg, x, y);
 	draw_rect_sets(dev, t->frame_bg, t->r_frame, t->nr_frame, x, y);
 }
 
-static void table_destruct(struct g_object *o_)
+static void table_destruct(struct g_object_table *o)
 {
-	struct g_object_table *o = get_struct(o_, struct g_object_table, go);
 	free_table(o->t);
 	mem_free(o);
 }
 
-static void table_get_list(struct g_object *o_, void (*fn)(struct g_object *parent, struct g_object *child))
+static void table_get_list(struct g_object_table *o, void (*fn)(struct g_object *parent, struct g_object *child))
 {
-	struct g_object_table *o = get_struct(o_, struct g_object_table, go);
 	struct table *t = o->t;
 	int i, j;
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
 		struct table_cell *c = CELL(t, i, j);
 		if (c->tag_object)
-			fn(&o->go, &c->tag_object->go);
+			fn((struct g_object *)o, (struct g_object *)c->tag_object);
 		if (c->root)
-			fn(&o->go, &c->root->go);
+			fn((struct g_object *)o, (struct g_object *)c->root);
 	}
 }
 
@@ -1754,25 +1741,25 @@ static void process_g_table(struct g_part *gp, struct table *t)
 						yw = safe_add(yw, er);
 					}
 				}
-				c->root->go.x = x, c->root->go.y = y;
-				c->root->go.y = safe_add(c->root->go.y, c->valign != VAL_MIDDLE && c->valign != VAL_BOTTOM ? 0 : (yw - c->root->go.yw) / (c->valign == VAL_MIDDLE ? 2 : 1));
+				c->root->x = x, c->root->y = y;
+				c->root->y = safe_add(c->root->y, c->valign != VAL_MIDDLE && c->valign != VAL_BOTTOM ? 0 : (yw - c->root->yw) / (c->valign == VAL_MIDDLE ? 2 : 1));
 			}
 			x = safe_add(x, t->w_c[i]);
 		}
 		y = safe_add(y, t->r_heights[j]);
 	}
 
-	if (html_top.list_entry.next != &html_stack) ta = &list_struct(html_top.list_entry.next, struct html_element)->attr;
+	if (html_top.next != (struct html_element *)(void *)&html_stack) ta = &html_top.next->attr;
 	else ta = &format_;
 
 	if (t->bordercolor && !decode_color(t->bordercolor, &dummy)) {
-		if (!(t->frame_bg = g_get_background(NULL, t->bordercolor))) {
+		if (!(t->frame_bg = get_background(NULL, t->bordercolor))) {
 			free_table(t);
 			return;
 		}
 	} else {
 		table_bg(ta, bgstr);
-		if (!(t->frame_bg = g_get_background(NULL, bgstr))) {
+		if (!(t->frame_bg = get_background(NULL, bgstr))) {
 			free_table(t);
 			return;
 		}
@@ -1857,12 +1844,14 @@ static void process_g_table(struct g_part *gp, struct table *t)
 			add_to_cell_sets(&t->r_cells, &t->w_cells, &t->nr_cells, &r, c);
 			memcpy(&c->rect, &r, sizeof(struct rect));
 			c->brd = init_rect_set();
+			/*debug("%d,%d %d,%d", r.x1, r.y1, r.x2, r.y2);*/
 			add_to_rect_set(&c->brd, &r);
-			r.x1 = c->root->go.x;
-			r.y1 = c->root->go.y;
-			r.x2 = safe_add(c->root->go.x, c->root->go.xw);
-			r.y2 = safe_add(c->root->go.y, c->root->go.yw);
+			r.x1 = c->root->x;
+			r.y1 = c->root->y;
+			r.x2 = safe_add(c->root->x, c->root->xw);
+			r.y2 = safe_add(c->root->y, c->root->yw);
 			exclude_rect_from_set(&c->brd, &r);
+			/*debug("%d,%d %d,%d", r.x1, r.y1, r.x2, r.y2);*/
 		}
 	}
 
@@ -1870,19 +1859,19 @@ static void process_g_table(struct g_part *gp, struct table *t)
 	mem_free(fv);
 
 	o = mem_calloc(sizeof(struct g_object_table));
-	o->go.mouse_event = table_mouse_event;
-	o->go.draw = table_draw;
-	o->go.destruct = table_destruct;
-	o->go.get_list = table_get_list;
-	o->go.xw = t->rw;
-	o->go.yw = t->rh;
+	o->mouse_event = table_mouse_event;
+	o->draw = table_draw;
+	o->destruct = table_destruct;
+	o->get_list = table_get_list;
+	o->xw = t->rw;
+	o->yw = t->rh;
 	o->t = t;
 	t->bg = gp->root->bg;
 	flush_pending_text_to_line(gp);
 	flush_pending_line_to_obj(gp, 0);
 	gp->cx = -1;
 	gp->cx_w = 0;
-	add_object_to_line(gp, &gp->line, &o->go);
+	add_object_to_line(gp, &gp->line, (struct g_object *)o);
 	flush_pending_text_to_line(gp);
 	par_format.align = t->align;
 	flush_pending_line_to_obj(gp, 0);
@@ -1894,17 +1883,13 @@ static void process_g_table(struct g_part *gp, struct table *t)
 			struct table_cell *c;
 			c = CELL(t, i, j);
 			if (c->root && c->tag) {
-				struct g_object_tag *tag;
-				size_t sl;
-				sl = strlen(cast_const_char c->tag);
-				if (sl > MAXINT - sizeof(struct g_object_tag)) overalloc();
-				tag = mem_calloc(sizeof(struct g_object_tag) + sl);
-				tag->go.mouse_event = g_dummy_mouse;
-				tag->go.draw = g_dummy_draw;
-				tag->go.destruct = g_tag_destruct;
+				struct g_object_tag *tag = mem_calloc(sizeof(struct g_object_tag) + strlen(cast_const_char c->tag));
+				tag->mouse_event = g_dummy_mouse;
+				tag->draw = g_dummy_draw;
+				tag->destruct = g_tag_destruct;
 				strcpy(cast_char tag->name, cast_const_char c->tag);
-				tag->go.x = c->root->go.x;
-				tag->go.y = c->root->go.y + o->go.yw;
+				tag->x = c->root->x;
+				tag->y = c->root->y + o->yw;
 				c->tag_object = tag;
 			}
 		}
@@ -1915,7 +1900,8 @@ static void process_g_table(struct g_part *gp, struct table *t)
 #endif
 
 struct table_cache_entry {
-	list_entry_1st
+	struct table_cache_entry *next;
+	struct table_cache_entry *prev;
 	struct table_cache_entry *hash_next;
 	unsigned char *start;
 	unsigned char *end;
@@ -1924,7 +1910,6 @@ struct table_cache_entry {
 	int width;
 	int xs;
 	int link_num;
-	list_entry_last
 	union {
 		struct part p;
 #ifdef G
@@ -1990,10 +1975,9 @@ void add_table_cache_entry(unsigned char *start, unsigned char *end, int align,
 static void free_table_cache(void)
 {
 	struct table_cache_entry *tce;
-	struct list_head *ltce;
-	foreach(struct table_cache_entry, tce, ltce, table_cache) {
+	foreach(tce, table_cache) {
 		int hash = make_hash(tce->start, tce->xs);
 		table_cache_hash[hash] = NULL;
 	}
-	free_list(struct table_cache_entry, table_cache);
+	free_list(table_cache);
 }
