@@ -18,17 +18,15 @@ struct http_connection_info {
 };
 
 /* prototypes */
-static void http_send_header(struct connection *c);
-static void http_get_header(struct connection *c);
-static void test_restart(struct connection *c);
-static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url);
+static void http_send_header(struct connection *);
+static void http_get_header(struct connection *);
+static void add_user_agent(unsigned char **hdr, int *l);
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url);
-static void add_accept(unsigned char **hdr, int *l, struct connection *c);
+static void add_accept(unsigned char **hdr, int *l);
 static void add_accept_language(unsigned char **hdr, int *l, struct http_connection_info *info);
 static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url, struct connection *c);
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info);
 static void add_connection(unsigned char **hdr, int *l, int http10, int proxy, int post);
-static void add_upgrade(unsigned char **hdr, int *l);
 static void add_dnt(unsigned char **hdr, int *l);
 static void add_if_modified(unsigned char **hdr, int *l, struct connection *c);
 static void add_range(unsigned char **hdr, int *l, unsigned char *url, struct connection *c);
@@ -57,13 +55,15 @@ unsigned char *parse_http_header(unsigned char *head, unsigned char *item, unsig
 			for (g = ++f; *g >= ' '; g++)
 				;
 			while (g > f && g[-1] == ' ') g--;
-			h = memacpy(f, g - f);
+			h = mem_alloc(g - f + 1);
+			memcpy(h, f, g - f);
+			h[g - f] = 0;
 			if (ptr) {
 				*ptr = f;
 			}
 			return h;
 		}
-		cont:
+		cont:;
 		f--;
 	}
 	return NULL;
@@ -116,164 +116,6 @@ int get_http_code(unsigned char *head, int *code, int *version)
 	return 0;
 }
 
-static const unsigned char *months[12] = {
-	cast_uchar "Jan",
-	cast_uchar "Feb",
-	cast_uchar "Mar",
-	cast_uchar "Apr",
-	cast_uchar "May",
-	cast_uchar "Jun",
-	cast_uchar "Jul",
-	cast_uchar "Aug",
-	cast_uchar "Sep",
-	cast_uchar "Oct",
-	cast_uchar "Nov",
-	cast_uchar "Dec"
-};
-
-#ifdef HAVE_GMTIME
-static const unsigned char *days[7] = {
-	cast_uchar "Sun",
-	cast_uchar "Mon",
-	cast_uchar "Tue",
-	cast_uchar "Wed",
-	cast_uchar "Thu",
-	cast_uchar "Fri",
-	cast_uchar "Sat",
-};
-#endif
-
-time_t parse_http_date(unsigned char *date)	/* this functions is bad !!! */
-{
-#ifdef HAVE_MKTIME
-	time_t t = 0;
-	time_t offset;
-	/* Mon, 03 Jan 2000 21:29:33 GMT */
-	int y;
-	struct tm tm;
-	memset(&tm, 0, sizeof(struct tm));
-
-	date = cast_uchar strchr(cast_const_char date, ' ');
-	if (!date) return (time_t)-1;
-	date++;
-	if (*date >= '0' && *date <= '9') {
-			/* Sun, 06 Nov 1994 08:49:37 GMT */
-			/* Sunday, 06-Nov-94 08:49:37 GMT */
-		y = 0;
-		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-		tm.tm_mday = (date[0] - '0') * 10 + date[1] - '0';
-		date += 2;
-		if (*date != ' ' && *date != '-') return (time_t)-1;
-		date += 1;
-		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
-			if (!casecmp(date, months[tm.tm_mon], 3)) goto f1;
-		return (time_t)-1;
-		f1:
-		date += 3;
-		if (*date != '-' && *date != ' ') return (time_t)-1;
-		date++;
-		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-		if (date[2] == ' ') {
-				/* Sunday, 06-Nov-94 08:49:37 GMT */
-			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-			tm.tm_year = (date[0] >= '7' ? 1900 : 2000) + (date[0] - '0') * 10 + date[1] - '0' - 1900;
-			date += 2;
-		} else if (date[2] >= '0' && date[2] <= '9' && date[3] >= '0' && date[3] <= '9') {
-				/* Sun, 06 Nov 1994 08:49:37 GMT */
-			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-			if (date[2] < '0' || date[2] > '9') return (time_t)-1;
-			if (date[3] < '0' || date[3] > '9') return (time_t)-1;
-			tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
-			date += 4;
-		} else return (time_t)-1;
-		if (*date != ' ') return (time_t)-1;
-		date++;
-	} else {
-			/* Sun Nov  6 08:49:37 1994 */
-		y = 1;
-		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
-			if (!casecmp(date, months[tm.tm_mon], 3)) goto f2;
-		return (time_t)-1;
-		f2:
-		date += 3;
-		while (*date == ' ') date++;
-		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-		tm.tm_mday = date[0] - '0';
-		date++;
-		if (*date != ' ') {
-			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-			tm.tm_mday = tm.tm_mday * 10 + date[0] - '0';
-			date++;
-		}
-		if (*date != ' ') return (time_t)-1;
-		date++;
-	}
-
-	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-	tm.tm_hour = (date[0] - '0') * 10 + date[1] - '0';
-	date += 2;
-	if (*date != ':') return (time_t)-1;
-	date++;
-	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-	tm.tm_min = (date[0] - '0') * 10 + date[1] - '0';
-	date += 2;
-	if (*date != ':') return (time_t)-1;
-	date++;
-	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-	tm.tm_sec = (date[0] - '0') * 10 + date[1] - '0';
-	date += 2;
-	if (y) {
-		if (*date != ' ') return (time_t)-1;
-		date++;
-		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
-		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
-		if (date[2] < '0' || date[2] > '9') return (time_t)-1;
-		if (date[3] < '0' || date[3] > '9') return (time_t)-1;
-		tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
-		date += 4;
-	}
-	if (*date != ' ' && *date) return (time_t)-1;
-
-	/*debug("%02d:%02d:%02d", tm.tm_hour, tm.tm_min, tm.tm_sec);*/
-#if defined(HAVE_TIMEGM) && !defined(HAIKU)
-	t = timegm(&tm);
-	if (t == (time_t)-1)
-		return (time_t)-1;
-	offset = 0;
-#else
-	t = mktime(&tm);
-	if (t == (time_t)-1)
-		return (time_t)-1;
-	memset(&tm, 0, sizeof(struct tm));
-	tm.tm_year = 80;
-	tm.tm_mday = 1;
-	offset = 315532800 - mktime(&tm);
-#endif
-	return (time_t)((uttime)t + (uttime)offset);
-#else
-	return (time_t)-1;
-#endif
-}
-
-unsigned char *print_http_date(time_t date)
-{
-	static unsigned char str[64];
-#ifdef HAVE_GMTIME
-	struct tm *tm = gmtime(&date);
-	snprintf(cast_char str, sizeof str, "%s %s %2d %02d:%02d:%02d %04d", days[tm->tm_wday], months[tm->tm_mon], tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, tm->tm_year + 1900);
-#else
-	snprintf(cast_char str, sizeof str, "%.0f", (double)date);
-#endif
-	return str;
-}
-
 static struct {
 	char *name;
 	int bugs;
@@ -307,7 +149,6 @@ static int check_http_server_bugs(unsigned char *url, struct http_connection_inf
 
 static void http_end_request(struct connection *c, int notrunc, int nokeepalive, int state)
 {
-	struct http_connection_info *info = c->info;
 	if (state == S__OK) {
 		if (c->cache) {
 			if (!notrunc) truncate_entry(c->cache, c->from, 1);
@@ -316,8 +157,8 @@ static void http_end_request(struct connection *c, int notrunc, int nokeepalive,
 	}
 	setcstate(c, state);
 	if (c->info &&
-	    !info->close &&
-	    !info->send_close &&
+	    !((struct http_connection_info *)c->info)->close &&
+	    !((struct http_connection_info *)c->info)->send_close &&
 	    !nokeepalive) {
 		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT, 0);
 	} else {
@@ -351,7 +192,7 @@ static void add_url_to_str(unsigned char **str, int *l, unsigned char *url)
 {
 	unsigned char *sp;
 	for (sp = url; *sp && *sp != POST_CHAR; sp++) {
-		if (*sp <= ' ' || *sp >= 127) {
+		if (*sp <= ' ') {
 			unsigned char esc[4];
 			sprintf(cast_char esc, "%%%02X", (int)*sp);
 			add_to_str(str, l, esc);
@@ -364,7 +205,7 @@ static void add_url_to_str(unsigned char **str, int *l, unsigned char *url)
 static void http_send_header(struct connection *c)
 {
 	struct http_connection_info *info;
-	int http10 = http_options.http10;
+	int http10 = http_options.http10 && !SCRUB_HEADERS;
 	int proxy;
 	unsigned char *hdr;
 	unsigned char *h, *u;
@@ -378,15 +219,20 @@ static void http_send_header(struct connection *c)
 			c->cache->refcount--;
 	}
 
-	proxy = is_proxy_url(c->url);
-	host = remove_proxy_prefix(c->url);
-	set_connection_timeout_keepal(c);
+	proxy = upcase(c->url[0]) == 'P';
+	host = !proxy ? c->url : get_url_data(c->url);
+	set_connection_timeout(c);
 	info = mem_calloc(sizeof(struct http_connection_info));
 	c->info = info;
 #ifdef HAVE_SSL
 	info->https_forward = !c->ssl && proxy && host && !casecmp(host, cast_uchar "https://", 8);
 	if (c->ssl) proxy = 0;
 #endif
+	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive)
+#ifdef HAVE_SSL
+		|| c->ssl
+#endif
+		;
 	hdr = init_str();
 	if (!host) {
 		http_bad_url:
@@ -404,7 +250,6 @@ static void http_send_header(struct connection *c)
 		post = cast_uchar strchr(cast_const_char host, POST_CHAR);
 		if (post) post++;
 	}
-	info->send_close = info->https_forward || http10;
 #ifdef HAVE_SSL
 	if (info->https_forward) {
 		add_to_str(&hdr, &l, cast_uchar "CONNECT ");
@@ -420,14 +265,14 @@ static void http_send_header(struct connection *c)
 		goto added_connect;
 	} else
 #endif
-	if (!post) {
+	  if (!post) {
 		add_to_str(&hdr, &l, cast_uchar "GET ");
 	} else {
 		add_to_str(&hdr, &l, cast_uchar "POST ");
 		c->unrestartable = 2;
 	}
 	if (!proxy) {
-		add_chr_to_str(&hdr, &l, '/');
+		add_to_str(&hdr, &l, cast_uchar "/");
 		u = get_url_data(host);
 	} else {
 		u = host;
@@ -455,9 +300,6 @@ static void http_send_header(struct connection *c)
 	else add_to_str(&hdr, &l, cast_uchar " HTTP/1.0\r\n");
 	if (!info->https_forward && (h = get_host_name(host))) {
 		add_to_str(&hdr, &l, cast_uchar "Host: ");
-		if (*h && h[strlen(cast_const_char h) - 1] == '.') {
-			h[strlen(cast_const_char h) - 1] = 0;
-		}
 		if (h[0] == '[' && h[strlen(cast_const_char h) - 1] == ']') {
 			unsigned char *pc = cast_uchar strchr(cast_const_char h, '%');
 			if (pc) {
@@ -468,44 +310,28 @@ static void http_send_header(struct connection *c)
 		add_to_str(&hdr, &l, h);
 		mem_free(h);
 		if ((h = get_port_str(host))) {
-			unsigned char *default_port = cast_uchar "80";
-#ifdef HAVE_SSL
-			if (c->ssl) default_port = cast_uchar "443";
-#endif
-			if (strcmp(cast_const_char h, cast_const_char default_port)) {
-				add_chr_to_str(&hdr, &l, ':');
-				add_to_str(&hdr, &l, h);
-			}
+			add_chr_to_str(&hdr, &l, ':');
+			add_to_str(&hdr, &l, h);
 			mem_free(h);
 		}
 		add_to_str(&hdr, &l, cast_uchar "\r\n");
 	}
-	add_user_agent(&hdr, &l, info->https_forward ? NULL : host);
+	add_user_agent(&hdr, &l);
 	if (proxy) add_proxy_auth_string(&hdr, &l, c->url);
 	if (!info->https_forward) {
-		test_restart(c);
-		if (!c->doh) {
-			add_referer(&hdr, &l, host, c->prev_url);
-		}
-		add_accept(&hdr, &l, c);
-		if (!c->doh) {
-			add_accept_language(&hdr, &l, info);
-			add_accept_encoding(&hdr, &l, host, c);
-			add_accept_charset(&hdr, &l, info);
-		}
+		add_referer(&hdr, &l, host, c->prev_url);
+		add_accept(&hdr, &l);
+		add_accept_language(&hdr, &l, info);
+		add_accept_encoding(&hdr, &l, host, c);
+		add_accept_charset(&hdr, &l, info);
 		add_dnt(&hdr, &l);
 		add_connection(&hdr, &l, http10, proxy, !info->send_close);
-		if (!c->doh) {
-			add_upgrade(&hdr, &l);
-			add_if_modified(&hdr, &l, c);
-			add_range(&hdr, &l, host, c);
-			add_pragma_no_cache(&hdr, &l, c->no_cache);
-			add_auth_string(&hdr, &l, host);
-		}
+		add_if_modified(&hdr, &l, c);
+		add_range(&hdr, &l, host, c);
+		add_pragma_no_cache(&hdr, &l, c->no_cache);
+		add_auth_string(&hdr, &l, host);
 		add_post_header(&hdr, &l, &post);
-		if (!c->doh) {
-			add_cookies(&hdr, &l, host);
-		}
+		add_cookies(&hdr, &l, host);
 		add_extra_options(&hdr, &l);
 	}
 	add_to_str(&hdr, &l, cast_uchar "\r\n");
@@ -525,52 +351,12 @@ static void http_send_header(struct connection *c)
 	setcstate(c, S_SENT);
 }
 
-static void test_restart(struct connection *c)
-{
-/* If the cached entity is compressed, request the whole file and turn off compression */
-	if (c->cache && c->from) {
-		unsigned char *d;
-		if ((d = parse_http_header(c->cache->head, cast_uchar "Content-Encoding", NULL))) {
-			mem_free(d);
-			c->from = 0;
-			c->no_compress = 1;
-#ifdef HAVE_ANY_COMPRESSION
-			if (c->tries >= 1) {
-				unsigned char *h;
-				if ((h = get_host_name(c->url))) {
-					add_blacklist_entry(h, BL_NO_COMPRESSION);
-					mem_free(h);
-				}
-			}
-#endif
-		}
-	}
-}
-
-static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url)
+static void add_user_agent(unsigned char **hdr, int *l)
 {
 	add_to_str(hdr, l, cast_uchar "User-Agent: ");
 	if (SCRUB_HEADERS) {
-		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 10.0; rv:102.0) Gecko/20100101 Firefox/102.0\r\n");
+		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 6.1; rv:31.0) Gecko/20100101 Firefox/31.0\r\n");
 	} else if (!(*http_options.header.fake_useragent)) {
-
-		/*
-		 * Google started to return css-styled page for searches.
-		 * It returns non-css page if the user agent begins with Lynx.
-		 */
-#if 0
-		if (url &&
-			(casestrstr(url, cast_uchar "/www.google.") ||
-			 casestrstr(url, cast_uchar "/google.")) &&
-			strstr(cast_const_char url, "/search?") &&
-			(strstr(cast_const_char url, "?q=") ||
-			 strstr(cast_const_char url, "&q=")) &&
-			!strstr(cast_const_char url, "?tbm=isch") &&
-			!strstr(cast_const_char url, "&tbm=isch")
-			)
-				add_to_str(hdr, l, cast_uchar("Lynx/"));
-#endif
-
 		add_to_str(hdr, l, cast_uchar("Links (" VERSION_STRING "; "));
 		add_to_str(hdr, l, system_name);
 		add_to_str(hdr, l, cast_uchar "; ");
@@ -578,9 +364,8 @@ static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url)
 		add_to_str(hdr, l, cast_uchar "; ");
 		if (!F && !list_empty(terminals)) {
 			struct terminal *term;
-			struct list_head *lterm;
 			unsigned char *t = cast_uchar "text";
-			foreach(struct terminal, term, lterm, terminals) if (term->spec->braille) t = cast_uchar "braille";
+			foreach(term, terminals) if (term->spec->braille) t = cast_uchar "braille";
 			add_to_str(hdr, l, t);
 		}
 #ifdef G
@@ -600,44 +385,38 @@ static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url)
 
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url)
 {
-	switch (http_options.header.referer) {
-		case REFERER_FAKE: {
-			add_to_str(hdr, l, cast_uchar "Referer: ");
-			add_to_str(hdr, l, http_options.header.fake_referer);
-			add_to_str(hdr, l, cast_uchar "\r\n");
-			break;
-		}
+	if (SCRUB_HEADERS)
+		return;
+	switch (http_options.header.referer)
+	{
+		case REFERER_FAKE:
+		add_to_str(hdr, l, cast_uchar "Referer: ");
+		add_to_str(hdr, l, http_options.header.fake_referer);
+		add_to_str(hdr, l, cast_uchar "\r\n");
+		break;
 
-		case REFERER_SAME_URL: {
-			add_to_str(hdr, l, cast_uchar "Referer: ");
-			add_url_to_str(hdr, l, url);
-			add_to_str(hdr, l, cast_uchar "\r\n");
-			break;
-		}
+		case REFERER_SAME_URL:
+		add_to_str(hdr, l, cast_uchar "Referer: ");
+		add_url_to_str(hdr, l, url);
+		add_to_str(hdr, l, cast_uchar "\r\n");
+		break;
 
-		case REFERER_REAL_SAME_SERVER: {
+		case REFERER_REAL_SAME_SERVER:
+		{
 			unsigned char *h, *j;
 			int brk = 1;
 			if ((h = get_host_name(url))) {
 				if ((j = get_host_name(prev_url))) {
-					if (!casestrcmp(h, j)) brk = 0;
-					else if (!casestrcmp(h, cast_uchar "imageproxy.jxs.cz")) {
-						int l = (int)strlen(cast_const_char j);
-						int q = (int)strlen(".blog.cz");
-						if (l > q && !casestrcmp((j + l - q), cast_uchar ".blog.cz")) brk = 0;
-						else if (!casestrcmp(j, cast_uchar "blog.cz")) brk = 0;
-					} else if (!casestrcmp(h, cast_uchar "www.google.com")) {
-						unsigned char *c = get_url_data(url);
-						if (c && !strncmp(cast_const_char c, "recaptcha/api/", 14)) brk = 0;
-					}
+					if (!strcasecmp(cast_const_char h, cast_const_char j)) brk = 0;
 					mem_free(j);
 				}
 				mem_free(h);
 			}
 			if (brk) break;
+			/* fall through */
 		}
-			/*-fallthrough*/
-		case REFERER_REAL: {
+		case REFERER_REAL:
+		{
 			unsigned char *ref;
 			unsigned char *user, *ins;
 			int ulen;
@@ -651,17 +430,15 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 			add_url_to_str(hdr, l, ref);
 			add_to_str(hdr, l, cast_uchar "\r\n");
 			mem_free(ref);
-			break;
 		}
+		break;
 	}
 }
 
-static void add_accept(unsigned char **hdr, int *l, struct connection *c)
+static void add_accept(unsigned char **hdr, int *l)
 {
-	if (c->doh) {
-		add_to_str(hdr, l, cast_uchar "Accept: application/dns-message\r\n");
-	} else if (SCRUB_HEADERS) {
-		add_to_str(hdr, l, cast_uchar "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8\r\n");
+	if (SCRUB_HEADERS) {
+		add_to_str(hdr, l, cast_uchar "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n");
 	} else {
 		add_to_str(hdr, l, cast_uchar "Accept: */*\r\n");
 	}
@@ -672,12 +449,12 @@ static void add_accept_language(unsigned char **hdr, int *l, struct http_connect
 	if (!(info->bl_flags & BL_NO_ACCEPT_LANGUAGE)) {
 		add_to_str(hdr, l, cast_uchar "Accept-Language: ");
 		if (SCRUB_HEADERS) {
-			add_to_str(hdr, l, cast_uchar "en-US,en;q=0.5\r\n");
+			add_to_str(hdr, l, cast_uchar "en-us,en;q=0.5\r\n");
 		} else {
 			int la;
 			la = *l;
-			add_to_str(hdr, l, get_text_translation(TEXT_(T__ACCEPT_LANGUAGE), NULL));
-			add_chr_to_str(hdr, l, ',');
+			add_to_str(hdr, l, _(TEXT_(T__ACCEPT_LANGUAGE), NULL));
+			add_to_str(hdr, l, cast_uchar ",");
 			if (!strstr(cast_const_char(*hdr + la), "en,") &&
 			    !strstr(cast_const_char(*hdr + la), "en;")) add_to_str(hdr, l, cast_uchar "en;q=0.2,");
 			add_to_str(hdr, l, cast_uchar "*;q=0.1\r\n");
@@ -707,7 +484,7 @@ static int advertise_compression(unsigned char *url, struct connection *c)
 
 static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url, struct connection *c)
 {
-#if defined(HAVE_ANY_COMPRESSION)
+#if defined(HAVE_ZLIB) || defined(HAVE_BZIP2) || defined(HAVE_LZMA)
 #define info	((struct http_connection_info *)c->info)
 	if (advertise_compression(url, c)) {
 		int orig_l = *l;
@@ -718,22 +495,6 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 		if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
 		add_to_str(hdr, l, cast_uchar "gzip, deflate");
 #endif
-#if defined(HAVE_BROTLI)
-		if ((!SCRUB_HEADERS
-#ifdef HAVE_SSL
-		    || c->ssl
-#endif
-		    ) && !(info->bl_flags & BL_NO_BZIP2)) {
-			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
-			add_to_str(hdr, l, cast_uchar "br");
-		}
-#endif
-#if defined(HAVE_ZSTD)
-		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
-			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
-			add_to_str(hdr, l, cast_uchar "zstd");
-		}
-#endif
 #if defined(HAVE_BZIP2)
 		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
 			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
@@ -744,13 +505,7 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 #if defined(HAVE_LZMA) && !defined(DOS)
 		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
 			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
-			add_to_str(hdr, l, cast_uchar "lzma, lzma2");
-		}
-#endif
-#if defined(HAVE_LZIP)
-		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
-			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
-			add_to_str(hdr, l, cast_uchar "lzip");
+			add_to_str(hdr, l, cast_uchar "lzma,lzma2");
 		}
 #endif
 		if (*l != l1) add_to_str(hdr, l, cast_uchar "\r\n");
@@ -763,31 +518,30 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info)
 {
 	static unsigned char *accept_charset = NULL;
-
-	if (SCRUB_HEADERS ||
-	    info->bl_flags & BL_NO_CHARSET ||
-	    http_options.no_accept_charset)
+	if (SCRUB_HEADERS)
 		return;
-
 	if (!accept_charset) {
 		int i;
 		unsigned char *cs, *ac;
 		int aclen = 0;
 		ac = init_str();
 		for (i = 0; (cs = get_cp_mime_name(i)); i++) {
-			if (aclen) add_chr_to_str(&ac, &aclen, ',');
+			if (aclen) add_to_str(&ac, &aclen, cast_uchar ",");
 			else add_to_str(&ac, &aclen, cast_uchar "Accept-Charset: ");
 			add_to_str(&ac, &aclen, cs);
 		}
 		if (aclen) add_to_str(&ac, &aclen, cast_uchar "\r\n");
-		if (!(accept_charset = cast_uchar strdup(cast_const_char ac))) {
-			add_to_str(hdr, l, ac);
+		retry:
+		if (!(accept_charset = malloc(strlen(cast_const_char ac) + 1))) {
+			if (out_of_memory(0, NULL, 0))
+				goto retry;
 			mem_free(ac);
 			return;
 		}
+		strcpy(cast_char accept_charset, cast_const_char ac);
 		mem_free(ac);
 	}
-	add_to_str(hdr, l, accept_charset);
+	if (!(info->bl_flags & BL_NO_CHARSET) && !http_options.no_accept_charset) add_to_str(hdr, l, accept_charset);
 }
 
 static void add_dnt(unsigned char **hdr, int *l)
@@ -806,13 +560,6 @@ static void add_connection(unsigned char **hdr, int *l, int http10, int proxy, i
 	}
 }
 
-static void add_upgrade(unsigned char **hdr, int *l)
-{
-	if (proxies.only_proxies) {
-		add_to_str(hdr, l, cast_uchar "Upgrade-Insecure-Requests: 1\r\n");
-	}
-}
-
 static void add_if_modified(unsigned char **hdr, int *l, struct connection *c)
 {
 	struct cache_entry *e;
@@ -846,6 +593,17 @@ static void add_range(unsigned char **hdr, int *l, unsigned char *url, struct co
 			return;
 	}
 	if (c->from /*&& (c->est_length == -1 || c->from < c->est_length)*/ && c->no_cache < NC_IF_MOD && !(info->bl_flags & BL_NO_RANGE)) {
+/* If the cached entity is compressed and we turned off compression,
+   request the whole file */
+#ifdef HAVE_ANY_COMPRESSION
+		if (!advertise_compression(url, c) && e) {
+			unsigned char *d;
+			if ((d = parse_http_header(e->head, cast_uchar "Transfer-Encoding", NULL))) {
+				mem_free(d);
+				return;
+			}
+		}
+#endif
 		add_to_str(hdr, l, cast_uchar "Range: bytes=");
 		add_num_to_str(hdr, l, c->from);
 		add_to_str(hdr, l, cast_uchar "-\r\n");
@@ -972,7 +730,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 		if (info->length >= 0) info->length -= l;
 		c->from += l;
 		kill_buffer_data(rb, l);
-		if (!info->length) {
+		if (!info->length && !rb->close) {
 			http_end_request(c, 0, 0, S__OK);
 			return;
 		}
@@ -996,10 +754,10 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 		} else if (info->chunk_remaining == -1) {
 			int l;
 			if ((l = is_line_in_buffer(rb))) {
-				char *end;
+				unsigned char *de;
 				long n = 0;	/* warning, go away */
-				if (l != -1) n = strtol(cast_const_char rb->data, &end, 16);
-				if (l == -1 || n < 0 || n >= MAXINT || cast_uchar end == rb->data) {
+				if (l != -1) n = strtol(cast_const_char rb->data, (char **)(void *)&de, 16);
+				if (l == -1 || n < 0 || n >= MAXINT || de == rb->data) {
 					setcstate(c, S_HTTP_ERROR);
 					abort_connection(c);
 					return;
@@ -1030,7 +788,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 			if (!info->chunk_remaining && rb->len >= 1) {
 				if (rb->data[0] == 10) kill_buffer_data(rb, 1);
 				else {
-					if (rb->data[0] != 13 || (rb->len >= 2 && ((unsigned char *)rb->data)[1] != 10)) {
+					if (rb->data[0] != 13 || (rb->len >= 2 && rb->data[1] != 10)) {
 						setcstate(c, S_HTTP_ERROR);
 						abort_connection(c);
 						return;
@@ -1054,11 +812,11 @@ static int get_header(struct read_buffer *rb)
 	if (rb->len <= 0) return 0;
 	if (rb->data[0] != 'H') return -2;
 	if (rb->len <= 1) return 0;
-	if (((unsigned char *)rb->data)[1] != 'T') return -2;
+	if (rb->data[1] != 'T') return -2;
 	if (rb->len <= 2) return 0;
-	if (((unsigned char *)rb->data)[2] != 'T') return -2;
+	if (rb->data[2] != 'T') return -2;
 	if (rb->len <= 3) return 0;
-	if (((unsigned char *)rb->data)[3] != 'P') return -2;
+	if (rb->data[3] != 'P') return -2;
 	for (i = 0; i < rb->len; i++) {
 		unsigned char a = rb->data[i];
 		if (/*a < ' ' && a != 10 && a != 13*/ !a) return -1;
@@ -1079,24 +837,25 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	off_t cf;
 	int state = c->state != S_PROC ? S_GETH : S_PROC;
 	unsigned char *head;
+	unsigned char *cookie, *ch;
 	int a, h = 0, version = 0;	/* against warning */
 	unsigned char *d;
 	struct cache_entry *e;
 	int previous_http_code;
 	struct http_connection_info *info;
-	unsigned char *host = remove_proxy_prefix(c->url);
+	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
 	set_connection_timeout(c);
 	info = c->info;
 	if (rb->close == 2) {
-		unsigned char *hs;
-		if (!c->tries && (hs = get_host_name(host))) {
+		unsigned char *h;
+		if (!c->tries && (h = get_host_name(host))) {
 			if (info->bl_flags & BL_NO_CHARSET) {
-				del_blacklist_entry(hs, BL_NO_CHARSET);
+				del_blacklist_entry(h, BL_NO_CHARSET);
 			} else {
-				add_blacklist_entry(hs, BL_NO_CHARSET);
+				add_blacklist_entry(h, BL_NO_CHARSET);
 				c->tries = -1;
 			}
-			mem_free(hs);
+			mem_free(h);
 		}
 		setcstate(c, S_CANT_READ);
 		retry_connection(c);
@@ -1115,7 +874,8 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	if (a != -2) {
-		head = memacpy(rb->data, a);
+		head = mem_alloc(a + 1);
+		memcpy(head, rb->data, a); head[a] = 0;
 		kill_buffer_data(rb, a);
 	} else {
 		head = stracpy(cast_uchar "HTTP/0.9 200 OK\r\nContent-Type: text/html\r\n\r\n");
@@ -1132,6 +892,12 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		retry_connection(c);
 		return;
 	}
+	ch = head;
+	while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
+		unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
+		set_cookie(NULL, host, cookie);
+		mem_free(cookie);
+	}
 	if (h == 100) {
 		mem_free(head);
 		state = S_PROC;
@@ -1159,14 +925,6 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 #endif
-	if (h != 401 && h != 407 && !c->doh) {
-		unsigned char *cookie;
-		unsigned char *ch = head;
-		while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-			set_cookie(host, cookie);
-			mem_free(cookie);
-		}
-	}
 	if (h == 204) {
 		mem_free(head);
 		http_end_request(c, 0, 0, S_HTTP_204);
@@ -1182,18 +940,6 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		http_end_request(c, 0, 1, S__OK);
 		return;
 	}
-	if (h == 431) {
-		unsigned char *hs;
-		if (!(info->bl_flags & BL_NO_CHARSET) && (hs = get_host_name(host))) {
-			mem_free(head);
-			add_blacklist_entry(hs, BL_NO_CHARSET);
-			mem_free(hs);
-			c->tries = -1;
-			setcstate(c, S_RESTART);
-			retry_connection(c);
-			return;
-		}
-	}
 	if ((h == 500 || h == 502 || h == 503 || h == 504) && http_options.retry_internal_errors && is_connection_restartable(c)) {
 			/* !!! FIXME: wait some time ... */
 		if (is_last_try(c)) {
@@ -1209,7 +955,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	if (!c->cache) {
-		if (get_connection_cache_entry(c)) {
+		if (get_cache_entry(c->url, &c->cache)) {
 			mem_free(head);
 			setcstate(c, S_OUT_OF_MEM);
 			abort_connection(c);
@@ -1222,11 +968,9 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	e->http_code = h;
 	if (e->head) mem_free(e->head);
 	e->head = head;
-	if (c->doh)
-		e->expire_time = 1;
 	if ((d = parse_http_header(head, cast_uchar "Expires", NULL))) {
 		time_t t = parse_http_date(d);
-		if (t != (time_t)-1 && e->expire_time != 1) e->expire_time = t;
+		if (t && e->expire_time != 1) e->expire_time = t;
 		mem_free(d);
 	}
 	if ((d = parse_http_header(head, cast_uchar "Pragma", NULL))) {
@@ -1243,8 +987,9 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (!casecmp(f, cast_uchar "max-age=", 8)) {
 				if (e->expire_time != 1) {
-					e->expire_time = get_absolute_seconds();
-					e->expire_time = (time_t)((uttime)e->expire_time + (uttime)atoi(cast_const_char(f + 8)));
+					errno = 0;
+					EINTRLOOPX(e->expire_time, time(NULL), (time_t)-1);
+					e->expire_time += atoi(cast_const_char(f + 8));
 				}
 			}
 			while (*f && *f != ',') f++;
@@ -1253,15 +998,19 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 #ifdef HAVE_SSL
 	if (c->ssl) {
+		int l = 0;
 		if (e->ssl_info) mem_free(e->ssl_info);
-		e->ssl_info = get_cipher_string(c->ssl);
-		if (e->ssl_authority) mem_free(e->ssl_authority);
-		e->ssl_authority = stracpy(c->ssl->ca);
+		e->ssl_info = init_str();
+		add_num_to_str(&e->ssl_info, &l, SSL_get_cipher_bits(c->ssl, NULL));
+		add_to_str(&e->ssl_info, &l, cast_uchar "-bit ");
+		add_to_str(&e->ssl_info, &l, cast_uchar SSL_get_cipher_version(c->ssl));
+		add_to_str(&e->ssl_info, &l, cast_uchar " ");
+		add_to_str(&e->ssl_info, &l, cast_uchar SSL_get_cipher_name(c->ssl));
 	}
 #endif
 	if (e->redirect) mem_free(e->redirect), e->redirect = NULL;
-	if ((h == 302 || h == 303 || h == 307 || h == 511) && !e->expire_time) e->expire_time = 1;
-	if (h == 301 || h == 302 || h == 303 || h == 307 || h == 308) {
+	if (h == 301 || h == 302 || h == 303 || h == 307) {
+		if ((h == 302 || h == 303 || h == 307) && !e->expire_time) e->expire_time = 1;
 		if ((d = parse_http_header(e->head, cast_uchar "Location", NULL))) {
 			unsigned char *user, *ins;
 			unsigned char *newuser, *newpassword;
@@ -1283,11 +1032,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (e->redirect) mem_free(e->redirect);
 			e->redirect = d;
-			if (h == 307 || h == 308) {
-				unsigned char *p;
-				if ((p = cast_uchar strchr(cast_const_char host, POST_CHAR)))
-					add_to_strn(&e->redirect, p);
-			}
+			e->redirect_get = h == 303;
 		}
 	}
 	if (!e->expire_time && strchr(cast_const_char c->url, POST_CHAR)) e->expire_time = 1;
@@ -1295,7 +1040,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	info->length = -1;
 	info->version = version;
 	if ((d = parse_http_header(e->head, cast_uchar "Connection", NULL)) || (d = parse_http_header(e->head, cast_uchar "Proxy-Connection", NULL))) {
-		if (!casestrcmp(d, cast_uchar "close")) info->close = 1;
+		if (!strcasecmp(cast_const_char d, "close")) info->close = 1;
 		mem_free(d);
 	} else if (version < 11) info->close = 1;
 	cf = c->from;
@@ -1303,7 +1048,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	if ((d = parse_http_header(e->head, cast_uchar "Content-Range", NULL))) {
 		if (strlen(cast_const_char d) > 6) {
 			d[5] = 0;
-			if (!(casestrcmp(d, cast_uchar "bytes")) && d[6] >= '0' && d[6] <= '9') {
+			if (!(strcasecmp(cast_const_char d, "bytes")) && d[6] >= '0' && d[6] <= '9') {
 				my_strtoll_t f = my_strtoll(d + 6, NULL);
 				if (f >= 0 && (off_t)f >= 0 && (off_t)f == f) c->from = f;
 			}
@@ -1323,20 +1068,20 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		unsigned char *ep;
 		my_strtoll_t l = my_strtoll(d, &ep);
 		if (!*ep && l >= 0 && (off_t)l >= 0 && (off_t)l == l) {
-			if (!info->close || version >= 11 || h / 100 == 3) info->length = l;
+			if (!info->close || version >= 11) info->length = l;
 			if (c->from + l >= 0) c->est_length = c->from + l;
 		}
 		mem_free(d);
 	}
 	if ((d = parse_http_header(e->head, cast_uchar "Accept-Ranges", NULL))) {
-		if (!casestrcmp(d, cast_uchar "none") && !c->unrestartable) c->unrestartable = 1;
+		if (!strcasecmp(cast_const_char d, "none") && !c->unrestartable) c->unrestartable = 1;
 		mem_free(d);
 	} else {
 		if (!c->unrestartable && !c->from) c->unrestartable = 1;
 	}
 	if (info->bl_flags & BL_NO_RANGE && !c->unrestartable) c->unrestartable = 1;
 	if ((d = parse_http_header(e->head, cast_uchar "Transfer-Encoding", NULL))) {
-		if (!casestrcmp(d, cast_uchar "chunked")) {
+		if (!strcasecmp(cast_const_char d, "chunked")) {
 			info->length = -2;
 			info->chunk_remaining = -1;
 		}
@@ -1344,7 +1089,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if (!info->close && info->length == -1) info->close = 1;
 	if ((d = parse_http_header(e->head, cast_uchar "Last-Modified", NULL))) {
-		if (e->last_modified && casestrcmp(e->last_modified, d)) {
+		if (e->last_modified && strcasecmp(cast_const_char e->last_modified, cast_const_char d)) {
 			delete_entry_content(e);
 			if (c->from) {
 				c->from = 0;
@@ -1364,33 +1109,25 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 
 	/*
 	 * Truncate entry if:
-	 *	- we are using DNS-over-HTTPS
 	 *	- it is compressed (the mix of an old and new document
 	 *	  would likely produce decompression error).
 	 *	- it was http authentication (the user doesn't need to see the
 	 *	  authentication message).
 	 */
-	if (c->doh) {
-		truncate_entry(e, c->from, 0);
-	} else if ((d = parse_http_header(e->head, cast_uchar "Content-Encoding", NULL))) {
+	if ((d = parse_http_header(e->head, cast_uchar "Content-Encoding", NULL))) {
 		mem_free(d);
 		truncate_entry(e, c->from, 0);
 	} else if (previous_http_code == 401 || previous_http_code == 407) {
 		truncate_entry(e, c->from, 0);
 	}
 
-	if (info->https_forward && h == 407) {
-		http_end_request(c, 0, 1, S__OK);
-		return;
-	}
-
 	read_http_data(c, rb);
 }
 
 static void http_get_header(struct connection *c)
 {
 	struct read_buffer *rb;
-	set_connection_timeout_keepal(c);
+	set_connection_timeout(c);
 	if (!(rb = alloc_read_buffer(c))) return;
 	rb->close = 1;
 	read_from_socket(c, c->sock1, rb, http_got_header);
