@@ -2,7 +2,7 @@
  * Generic image decoding and PNG and JPG decoders.
  * (c) 2002 Karel 'Clock' Kulhavy
  * This is a part of the Links program, released under GPL.
- 
+
  * Used in graphics mode of Links only
  TODO: odstranit zbytecne ditherovani z strip_optimized header_dimensions_known,
        protoze pozadi obrazku musi byt stejne jako pozadi stranky, a to se nikdy
@@ -39,9 +39,6 @@ struct decoded_image {
 static struct g_object_image *global_goi;
 struct cached_image *global_cimg;
 int end_callback_hit;
-#endif /* #ifdef G */
-int dither_images=1;
-#ifdef G
 
 static int is_image_size_sane(int x, int y)
 {
@@ -100,11 +97,8 @@ static void destroy_decoder (struct cached_image *cimg)
 #ifdef HAVE_TIFF
 		case IM_TIFF:
 			td=(struct tiff_decoder *)cimg->decoder;
-			if (td->tiff_open)
-			{
-				if (td->tiff_data)mem_free(td->tiff_data);
-				td->tiff_open=0;
-			}
+			if (td->tiff_data)mem_free(td->tiff_data);
+			td->tiff_open=0;
 			break;
 #endif
 		}
@@ -112,10 +106,16 @@ static void destroy_decoder (struct cached_image *cimg)
 	}
 }
 
+static void mem_free_buffer(struct cached_image *cimg)
+{
+	mem_free(cimg->buffer);
+	mem_freed_large(cimg->width * cimg->height * cimg->buffer_bytes_per_pixel);
+}
+
 static void img_destruct_image(struct g_object *object)
 {
 	struct g_object_image *goi=(struct g_object_image *)object;
-	
+
 	if (goi->orig_src)mem_free(goi->orig_src);
 	if (goi->alt)mem_free(goi->alt);
 	if (goi->name)mem_free(goi->name);
@@ -140,17 +140,17 @@ void img_destruct_cached_image(struct cached_image *cimg)
 		case 9:
 		case 11:
 		break;
-		
+
 		case 12:
 		case 14:
 		if (cimg->gamma_table) mem_free(cimg->gamma_table);
-		if (cimg->bmp.user){
+		if (cimg->bmp_used){
 			drv->unregister_bitmap(&(cimg->bmp));
 		}
 		if (cimg->strip_optimized){
 			if (cimg->dregs) mem_free(cimg->dregs);
 		}else{
-			mem_free(cimg->buffer);
+			mem_free_buffer(cimg);
 		}
 		case 8:
 		case 10:
@@ -161,7 +161,7 @@ void img_destruct_cached_image(struct cached_image *cimg)
 		case 15:
 		drv->unregister_bitmap(&(cimg->bmp));
 		break;
-		
+
 #ifdef DEBUG
 		default:
 		fprintf(stderr,"img_destruct_cached_image: state=%d\n",cimg->state);
@@ -189,7 +189,7 @@ static int img_scale_h(unsigned scale, int in){
 
 	if (in<=0) return in;
 	pre=((unsigned long)(aspect<65536UL?65536UL:aspect)*scale+128)>>8;
-	out=((unsigned long)in*pre+12800UL)/25600UL;
+	out=(int)(((unsigned long)in*pre+12800UL)/25600UL);
 	if (out<1) out=1;
 	return out;
 }
@@ -200,30 +200,30 @@ static int img_scale_v(unsigned scale, int in){
 
 	if (in<=0) return in;
 	divisor=(100*(aspect>=65536UL?65536UL:aspect)+128)>>8;
-	out=((unsigned long)in*(scale*256)+(divisor>>1))/divisor;
+	out=(int)(((unsigned long)in*(scale*256)+(divisor>>1))/divisor);
 	if (out<1) out=1;
 	return out;
 }
 
 /* Returns height (pixels) for prescribed width (pixels). Honours aspect. */
-static int width2height(float width_px, float width_mm, float height_mm)
+static int width2height(double width_px, double width_mm, double height_mm)
 {
 	int height_px;
 
-	if (width_px<=0) return width_px;
-	height_px=(height_mm*width_px*65536)/(aspect*width_mm);
+	if (width_px<=0) return 0;
+	height_px=(int)((height_mm*width_px*65536)/(aspect*width_mm));
 	if (height_px<1) height_px=1;
 	return height_px;
 
 }
 
 /* Returns width (pixels) for prescribed height (pixels). Honours aspect. */
-static int height2width(float height_px, float width_mm, float height_mm)
+static int height2width(double height_px, double width_mm, double height_mm)
 {
 	int width_px;
 
-	if (height_px<=0) return height_px;
-	width_px=(width_mm*height_px*aspect)/(65536*height_mm);
+	if (height_px<=0) return 0;
+	width_px=(int)((width_mm*height_px*aspect)/(65536*height_mm));
 	if (width_px<1) width_px=1;
 	return width_px;
 
@@ -237,22 +237,22 @@ void compute_background_8(unsigned char *rgb, struct cached_image *cimg)
 
 	round_color_sRGB_to_48(&red, &green, &blue
 		, cimg->background_color);
-	rgb[0]=apply_gamma_single_16_to_8(red
-		,cimg->red_gamma/user_gamma);
-	rgb[1]=apply_gamma_single_16_to_8(green
-		,cimg->green_gamma/user_gamma);
-	rgb[2]=apply_gamma_single_16_to_8(blue
-		,cimg->blue_gamma/user_gamma);
+	rgb[0]=ags_16_to_8(red
+		,(float)(cimg->red_gamma/user_gamma));
+	rgb[1]=ags_16_to_8(green
+		,(float)(cimg->green_gamma/user_gamma));
+	rgb[2]=ags_16_to_8(blue
+		,(float)(cimg->blue_gamma/user_gamma));
 }
 
 /* updates cimg state when header dimensions are know. Only allowed to be called
  * in state 8 and 10.
  * Allocates right amount of memory into buffer, formats it (with background or
  * zeroes, depens on buffer_bytes_per_pixel). Updates dimensions (xww and yww)
- * according to newly known header dimensions. Fills in gamma_stamp, bmp.user
- * (NULL because we not bother with generating bitmap here)
+ * according to newly known header dimensions. Fills in gamma_stamp, bmp_used
+ * (zero because we not bother with generating bitmap here)
  * and rows_added.
- * Resets strip_optimized if image will be scaled or 
+ * Resets strip_optimized if image will be scaled or
  * Allocates dregs if on exit strip_optimized is nonzero.
  * Allocates and computes gamma_table, otherwise
  * 	sets gamma_table to NULL. Also doesn't make gamma table if image contains less
@@ -296,7 +296,7 @@ int header_dimensions_known(struct cached_image *cimg)
 			cimg->xww=height2width(cimg->yww,
 					cimg->width, cimg->height);
 			if (cimg->xww<=0) cimg->xww=1;
-			
+
 		}
 	}else{
 		/* Specified width */
@@ -318,7 +318,7 @@ int header_dimensions_known(struct cached_image *cimg)
 				cimg->yww=width2height(cimg->xww,
 						cimg->width, cimg->height);
 			}
-			
+
 			/* Some sanity checks */
 			if (cimg->xww<=0) cimg->xww=1;
 			if (cimg->yww<=0) cimg->yww=1;
@@ -376,7 +376,7 @@ int header_dimensions_known(struct cached_image *cimg)
 		skip_img:
 		drv->register_bitmap(&(cimg->bmp));
 		if(cimg->dregs) memset(cimg->dregs,0,cimg->width*sizeof(*cimg->dregs)*3);
-		cimg->bmp.user=(void *)&end_callback_hit; /* Nonzero value */
+		cimg->bmp_used=1; /* Nonzero value */
 		/* This ensures the dregs are none and because strip
 		 * optimization is unusable in interlaced pictures,
 		 * this saves the zeroing out at the beginning of the
@@ -384,15 +384,17 @@ int header_dimensions_known(struct cached_image *cimg)
 		 */
 	}else {
 		cimg->rows_added=1;
-		cimg->bmp.user=NULL;
+		cimg->bmp_used=0;
 		if (cimg->width && (unsigned)cimg->width * (unsigned)cimg->height / (unsigned)cimg->width != (unsigned)cimg->height) overalloc();
 		if ((unsigned)cimg->width * (unsigned)cimg->height > (unsigned)MAXINT / cimg->buffer_bytes_per_pixel) overalloc();
-		cimg->buffer=mem_alloc(cimg->width*cimg->height
+		cimg->buffer=mem_alloc_mayfail(cimg->width*cimg->height
 			*cimg->buffer_bytes_per_pixel);
+		if (!cimg->buffer)
+			return 1;
 		if (cimg->buffer_bytes_per_pixel==4
 			 	||cimg->buffer_bytes_per_pixel==4
 				*sizeof(unsigned short))
-			{	
+			{
 			/* Make the buffer contain full transparency */
 			memset(cimg->buffer,0,cimg->width*cimg->height
 				*cimg->buffer_bytes_per_pixel);
@@ -405,18 +407,18 @@ int header_dimensions_known(struct cached_image *cimg)
 				round_color_sRGB_to_48(&red, &green, &blue
 					, cimg->background_color);
 
-				red=apply_gamma_single_16_to_16(red
-					,cimg->red_gamma/user_gamma);
-				green=apply_gamma_single_16_to_16(green
-					,cimg->green_gamma/user_gamma);
-				blue=apply_gamma_single_16_to_16(blue
-					,cimg->blue_gamma / user_gamma);
+				red=ags_16_to_16(red
+					,(float)(cimg->red_gamma/user_gamma));
+				green=ags_16_to_16(green
+					,(float)(cimg->green_gamma/user_gamma));
+				blue=ags_16_to_16(blue
+					,(float)(cimg->blue_gamma/user_gamma));
 				mix_one_color_48((unsigned short *)cimg->buffer
 					,cimg->width*cimg->height,red
 					,green, blue);
 			}else{
 				unsigned char rgb[3];
-				
+
 				/* 8-bit */
 				compute_background_8(rgb,cimg);
 				mix_one_color_24(cimg->buffer
@@ -451,50 +453,49 @@ static unsigned short *buffer_to_16(unsigned short *tmp, struct cached_image *ci
 	switch (cimg->buffer_bytes_per_pixel){
 		case 3:
 			if (cimg->gamma_table){
-				apply_gamma_exponent_24_to_48_table(tmp, buffer,
+				agx_24_to_48_table(tmp, buffer,
 					cimg->width*height
 					,cimg->gamma_table);
 			}
 			else{
-				apply_gamma_exponent_24_to_48(tmp,buffer,cimg->width
+				agx_24_to_48(tmp,buffer,cimg->width
 					*height
-					,user_gamma/cimg->red_gamma
-					,user_gamma/cimg->green_gamma
-					,user_gamma/cimg->blue_gamma);
+					,(float)(user_gamma/cimg->red_gamma)
+					,(float)(user_gamma/cimg->green_gamma)
+					,(float)(user_gamma/cimg->blue_gamma));
 			}
 		break;
 
 		case 3*sizeof(unsigned short):
 			if (cimg->gamma_table){
-				apply_gamma_exponent_48_to_48_table(tmp
+				agx_48_to_48_table(tmp
 					,(unsigned short *)buffer
 					,cimg->width*height, cimg->gamma_table);
 			}else{
-				apply_gamma_exponent_48_to_48(tmp,(unsigned short *)buffer
-					,cimg->width*height,
-					user_gamma/cimg->red_gamma,
-					user_gamma/cimg->green_gamma,
-					user_gamma/cimg->blue_gamma);
+				agx_48_to_48(tmp,(unsigned short *)buffer
+					,cimg->width*height
+					,(float)(user_gamma/cimg->red_gamma)
+					,(float)(user_gamma/cimg->green_gamma)
+					,(float)(user_gamma/cimg->blue_gamma));
 			}
 		break;
 
 		/* Alpha's: */
-		case 4: 
+		case 4:
 		{
 
 			round_color_sRGB_to_48(&red,&green,&blue,cimg->background_color);
 			if (cimg->gamma_table){
-				apply_gamma_exponent_and_undercolor_32_to_48_table(
+				agx_and_uc_32_to_48_table(
 						tmp, buffer, cimg->width *height,
 						cimg->gamma_table, red, green, blue);
 			}else{
-				
-				apply_gamma_exponent_and_undercolor_32_to_48(tmp,buffer
-					,cimg->width*height,
-					user_gamma/cimg->red_gamma,
-					user_gamma/cimg->green_gamma,
-					user_gamma/cimg->blue_gamma,
-					red, green, blue);
+				agx_and_uc_32_to_48(tmp,buffer
+					,cimg->width*height
+					,(float)(user_gamma/cimg->red_gamma)
+					,(float)(user_gamma/cimg->green_gamma)
+					,(float)(user_gamma/cimg->blue_gamma)
+					,red, green, blue);
 			}
 		}
 		break;
@@ -504,20 +505,20 @@ static unsigned short *buffer_to_16(unsigned short *tmp, struct cached_image *ci
 			round_color_sRGB_to_48(&red, &green, &blue,
 				cimg->background_color);
 			if (cimg->gamma_table){
-				apply_gamma_exponent_and_undercolor_64_to_48_table
+				agx_and_uc_64_to_48_table
 					(tmp, (unsigned short *)buffer, cimg->width*height
 					,cimg->gamma_table, red, green, blue);
 			}else{
-				apply_gamma_exponent_and_undercolor_64_to_48(tmp
-					,(unsigned short*)buffer,cimg->width*height,
-					user_gamma/cimg->red_gamma,
-					user_gamma/cimg->green_gamma,
-					user_gamma/cimg->blue_gamma,
-					red,green,blue);
+				agx_and_uc_64_to_48(tmp
+					,(unsigned short*)buffer,cimg->width*height
+					,(float)(user_gamma/cimg->red_gamma)
+					,(float)(user_gamma/cimg->green_gamma)
+					,(float)(user_gamma/cimg->blue_gamma)
+					,red,green,blue);
 			}
 		}
 		break;
-		
+
 #ifdef DEBUG
 		default:
 		internal("buffer_to_16: unknown mem organization");
@@ -536,7 +537,7 @@ static unsigned short *buffer_to_16(unsigned short *tmp, struct cached_image *ci
  * dregs must be externally allocated and contain required value or must be
  * NULL.
  * if !dregs then rounding is performed instead of dithering.
- * dregs are not freed. 
+ * dregs are not freed.
  * bottom dregs are placed back into dregs.
  * Before return the bitmap will be in registered state and changes will be
  * commited.
@@ -588,7 +589,6 @@ not_enough:
 	       	dither_restart(tmp, &tmpbmp, dregs);
 	}
 	else {
-		
 		(*round_fn)(tmp, &tmpbmp);
 	}
 	if (use_strip) {
@@ -599,7 +599,7 @@ not_enough:
 	if (!height) goto end;
 	buffer+=add2;
 	yoff+=tmpbmp.y;
-	tmpbmp.data=(unsigned char *)tmpbmp.data+add1; 
+	tmpbmp.data=(unsigned char *)tmpbmp.data+add1;
 	/* This has no effect if use_strip but it's faster
 	 * to add to bogus value than to play with
 	 * conditional jumps.
@@ -607,12 +607,13 @@ not_enough:
 	goto not_enough;
 end:
 	mem_free(tmp);
+	mem_freed_large(cimg->width*(height<max_height?height:max_height)*3*sizeof(*tmp));
 	if (!use_strip) drv->register_bitmap(&(cimg->bmp));
 }
 
 /* Takes the buffer and resamples the data into the bitmap. Automatically
  * destroys the previous bitmap. Must be called only when cimg->buffer is valid.
- * Sets bmp->user to non-NULL
+ * Sets bmp_used to non-zero.
  * If gamma_table is used, it must be still allocated here (take care if you
  * want to destroy gamma table and call buffer_to_bitmap, first call buffer_to_bitmap
  * and then destroy gamma_table).
@@ -651,31 +652,34 @@ buffer_to_bitmap");
 		gonna_be_smart=0;
 		if (ix && (unsigned)ix * (unsigned)iy / (unsigned)ix != (unsigned)iy) overalloc();
 		if ((unsigned)ix * (unsigned)iy > MAXINT / sizeof(*tmp) / 3) overalloc();
-		tmp=mem_alloc(ix*iy*3*sizeof(*tmp));
-		buffer_to_16(tmp,cimg,cimg->buffer,iy);
+		tmp=mem_alloc_mayfail(ix*iy*3*sizeof(*tmp));
+		if (tmp) buffer_to_16(tmp,cimg,cimg->buffer,iy);
 		if (!cimg->decoder){
-			mem_free(cimg->buffer);
+			mem_free_buffer(cimg);
 			cimg->buffer=NULL;
 		}
-	
+
 		/* Scale the image to said size */
 #ifdef DEBUG
 		if (ox<=0||oy<=0){
 			internal("ox or oy <=0 before resampling image");
 		}
 #endif /* #ifdef DEBUG */
-		if (ix!=ox||iy!=oy){
+		if (tmp && (ix!=ox || iy!=oy)) {
 			/* We must really scale */
 			tmp1=tmp;
 			scale_color(tmp1,ix,iy,&tmp,ox,oy);
 		}
 	}
-	if (cimg->bmp.user) drv->unregister_bitmap(&cimg->bmp);
+	if (cimg->bmp_used) drv->unregister_bitmap(&cimg->bmp);
 	cimg->bmp.x=ox;
 	cimg->bmp.y=oy;
 	if (drv->get_empty_bitmap(&(cimg->bmp))) {
 		if (!gonna_be_smart) {
-			mem_free(tmp);
+			if (tmp) {
+				mem_free(tmp);
+				mem_freed_large(ix*iy*3*sizeof(*tmp));
+			}
 		}
 		goto bitmap_failed;
 	}
@@ -690,15 +694,25 @@ buffer_to_bitmap");
 			0, dregs, 0);
 		if (dregs) mem_free(dregs);
 	}else{
-		if (dither_images)
-			dither(tmp,&(cimg->bmp));
-		else
-			(*round_fn)(tmp,&(cimg->bmp));
-		mem_free(tmp);
+		if (tmp) {
+			if (dither_images)
+				dither(tmp,&(cimg->bmp));
+			else
+				(*round_fn)(tmp,&(cimg->bmp));
+			mem_free(tmp);
+			mem_freed_large(ix*iy*3*sizeof(*tmp));
+		} else {
+			int i;
+			unsigned char *ptr = cimg->bmp.data;
+			for (i = 0; i < cimg->bmp.y; i++) {
+				memset(ptr, 0, cimg->bmp.x * (drv->depth & 7));
+				ptr += cimg->bmp.skip;
+			}
+		}
 		bitmap_failed:
 		drv->register_bitmap(&(cimg->bmp));
 	}
-	cimg->bmp.user=(void *)&end_callback_hit;
+	cimg->bmp_used=1;
 	/* Indicate that the bitmap is valid. The value is just any
 	   nonzero value */
 	cimg->rows_added=0;
@@ -717,7 +731,7 @@ void img_end(struct cached_image *cimg)
 		}
 		else{
 			buffer_to_bitmap(cimg);
-			mem_free(cimg->buffer);
+			mem_free_buffer(cimg);
 		}
 		if (cimg->gamma_table) mem_free(cimg->gamma_table);
 		case 8:
@@ -725,13 +739,13 @@ void img_end(struct cached_image *cimg)
 		destroy_decoder(cimg);
 		case 0:
 		case 1:
-		case 2:	
+		case 2:
 		case 3:
 		case 9:
 		case 11:
 		case 13:
 		case 15:
-		break;	
+		break;
 #ifdef DEBUG
 		default:
 		fprintf(stderr,"state=%d\n",cimg->state);
@@ -765,14 +779,14 @@ static void r3l0ad(struct cached_image *cimg, struct g_object_image *goi)
 		if (cimg->strip_optimized){
 			if (cimg->dregs) mem_free(cimg->dregs);
 		}else{
-			mem_free(cimg->buffer);
+			mem_free_buffer(cimg);
 		}
-		if (cimg->bmp.user){
+		if (cimg->bmp_used){
 			case 13:
 			drv->unregister_bitmap(&cimg->bmp);
 		}
-                cimg->xww=img_scale_h(cimg->scale, cimg->wanted_xw<0?32:cimg->wanted_xw);
-                cimg->yww=img_scale_v(cimg->scale, cimg->wanted_yw<0?32:cimg->wanted_yw);
+		cimg->xww=img_scale_h(cimg->scale, cimg->wanted_xw<0?32:cimg->wanted_xw);
+		cimg->yww=img_scale_v(cimg->scale, cimg->wanted_yw<0?32:cimg->wanted_yw);
 		break;
 
 		case 14:
@@ -781,9 +795,9 @@ static void r3l0ad(struct cached_image *cimg, struct g_object_image *goi)
 		if (cimg->strip_optimized){
 			if (cimg->dregs) mem_free(cimg->dregs);
 		}else{
-			mem_free(cimg->buffer);
+			mem_free_buffer(cimg);
 		}
-		if (cimg->bmp.user){
+		if (cimg->bmp_used){
 			case 15:
 			drv->unregister_bitmap(&cimg->bmp);
 		}
@@ -802,9 +816,9 @@ static void r3l0ad(struct cached_image *cimg, struct g_object_image *goi)
  * If doesn't return 1 then returns 0
  * dtest - Destructive TEST
  */
-static inline int dtest(unsigned char *template, unsigned char *test)
+static inline int dtest(unsigned char *templat, unsigned char *test)
 {
-	if (strcasecmp(template,test)) return 0;
+	if (strcasecmp(cast_const_char templat, cast_const_char test)) return 0;
 	else{
 		mem_free(test);
 		return 1;
@@ -822,43 +836,43 @@ static void type(struct cached_image *cimg, unsigned char *content_type)
 	}
 #endif /* #ifdef DEBUG */
 #ifdef HAVE_JPEG
-	if (dtest("image/jpeg",content_type)){
+	if (dtest(cast_uchar "image/jpeg",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else if (dtest("image/jpg",content_type)){
+	}else if (dtest(cast_uchar "image/jpg",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else if (dtest("image/jpe",content_type)){
+	}else if (dtest(cast_uchar "image/jpe",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else if (dtest("image/pjpe",content_type)){
+	}else if (dtest(cast_uchar "image/pjpe",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else if (dtest("image/pjpeg",content_type)){
+	}else if (dtest(cast_uchar "image/pjpeg",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else if (dtest("image/pjpg",content_type)){
+	}else if (dtest(cast_uchar "image/pjpg",content_type)){
 		cimg->image_type=IM_JPG;
 		jpeg_start(cimg);
-	}else 
+	}else
 #endif /* #ifdef HAVE_JPEG */
-		if (dtest("image/png",content_type)){
+		if (dtest(cast_uchar "image/png",content_type)){
 		cimg->image_type=IM_PNG;
 		png_start(cimg);
-	}else if (dtest("image/x-png",content_type)){
+	}else if (dtest(cast_uchar "image/x-png",content_type)){
 		cimg->image_type=IM_PNG;
 		png_start(cimg);
-	}else if (dtest("image/gif",content_type)){
+	}else if (dtest(cast_uchar "image/gif",content_type)){
 		cimg->image_type=IM_GIF;
 		gif_start(cimg);
-	}else if (dtest("image/x-xbitmap",content_type)){
+	}else if (dtest(cast_uchar "image/x-xbitmap",content_type)){
 		cimg->image_type=IM_XBM;
 		xbm_start(cimg);
 #ifdef HAVE_TIFF
-	}else if (dtest("image/tiff",content_type)){
+	}else if (dtest(cast_uchar "image/tiff",content_type)){
 		cimg->image_type=IM_TIFF;
 		tiff_start(cimg);
-	}else if (dtest("image/tif",content_type)){
+	}else if (dtest(cast_uchar "image/tif",content_type)){
 		cimg->image_type=IM_TIFF;
 		tiff_start(cimg);
 #endif /* #ifdef HAVE_TIFF */
@@ -890,7 +904,7 @@ static int img_process_download(struct g_object_image *goi, struct f_data_c *fda
 
 #ifdef DEBUG
 	if (!goi->af) internal("NULL goi->af in process_download\n");
-	if (cimg->state>=16){ /* Negative don't occur becaus it's unsigned char */
+	if (cimg->state>=16){ /* Negative don't occur because it's unsigned char */
 		fprintf(stderr,"cimg->state=%d\n",cimg->state);
 		internal("Invalid cimg->state in img_process_download\n");
 	}
@@ -910,17 +924,17 @@ static int img_process_download(struct g_object_image *goi, struct f_data_c *fda
 		/* Type still unknown */
 		ctype=get_content_type(goi->af->rq->ce->head,
 			goi->af->rq->url);
-		if (!ctype) ctype = stracpy("application/octet-stream");
+		if (!ctype) ctype = stracpy(cast_uchar "application/octet-stream");
 		type(cimg,ctype);
 	}
-		
+
 	/* Now, if we are in state where decoder is running (8, 10, 12, 14), we may feed
 	 * some data into it.
 	 */
 
 	if (!((cimg->state^8)&9)){
 		if (get_file(goi->af->rq, &data, &dataend)) goto end;
-		length = dataend - data;
+		length = (int)(dataend - data);
 		if (length<=cimg->last_length) goto end; /* No new data */
 
 		data+=cimg->last_length;
@@ -963,7 +977,7 @@ img_process_download.\n");
 		cimg->last_length+=length;
 	}
 	end:
-	
+
 	/* Test end */
 	if (!is_entry_used(goi->af->rq->ce) && (goi->af->rq->state < 0
 		||(goi->af->rq->ce&&goi->af->rq->stat.state<0))){
@@ -1006,9 +1020,7 @@ int get_foreground(int rgb)
 	return (r<<16)|(g<<8)|b;
 }
 
-static void draw_frame_mark (struct graphics_driver *drv, struct 
-	graphics_device *dev, int x, int y, int xw, int yw
-	, long bg, long fg, int broken)
+static void draw_frame_mark(struct graphics_device *dev, int x, int y, int xw, int yw, long bg, long fg, int broken)
 {
 #ifdef DEBUG
 	if (xw<1||yw<1) internal("zero dimension in draw_frame_mark");
@@ -1018,7 +1030,7 @@ static void draw_frame_mark (struct graphics_driver *drv, struct
 		 * side (0-1)
 		 */
 		 int xl, xh, yl, yh;
-		 
+
 		 xh=xw-(xl=xw>>2);
 		 yh=yw-(yl=yw>>2);
 		/* Draw full sides and the box inside */
@@ -1064,7 +1076,7 @@ static void draw_frame_mark (struct graphics_driver *drv, struct
 			drv->draw_hline(dev,x,y+yw-1,x+xw,fg);
 		}
 		if (broken == 2 && xw > 2 && yw > 2) {
-			draw_frame_mark(drv, dev, x + 1, y + 1, xw - 2, yw - 2, bg, fg, 3);
+			draw_frame_mark(dev, x + 1, y + 1, xw - 2, yw - 2, bg, fg, 3);
 		}
 	}
 }
@@ -1075,7 +1087,7 @@ static void draw_frame_mark (struct graphics_driver *drv, struct
  * the buffer.
  */
 static void draw_picture(struct f_data_c *fdatac, struct g_object_image *goi,
-		int x, int y, int bg)
+		int x, int y, long bg)
 {
 	struct graphics_device *dev=fdatac->ses->term->dev;
 	struct cached_image *cimg=goi->cimg;
@@ -1088,11 +1100,11 @@ static void draw_picture(struct f_data_c *fdatac, struct g_object_image *goi,
 	}
 #endif /* #ifdef DEBUG */
 	if (!(cimg->state&1)){
-		if (!cimg->bmp.user)
+		if (!cimg->bmp_used)
 			buffer_to_bitmap(cimg);
 	}
 #ifdef DEBUG
-	else if (!cimg->bmp.user){
+	else if (!cimg->bmp_used){
 		fprintf(stderr,"cimg->state=%d\n",cimg->state);
 		internal("Nonexistent bitmap in said cimg->state in draw_picture");
 	}
@@ -1131,7 +1143,7 @@ static void img_draw_image (struct f_data_c *fdatac, struct g_object_image *goi,
 	 * navod, jak se vola to refresh_image.  Nicmene ja jsem milostive
 	 * usoudil, ze zadnejch 1000, ale 0.
 	 */
-	
+
 	if (cimg) {
 		color_bg=dip_get_color_sRGB(cimg->background_color);
 		color_fg=dip_get_color_sRGB(get_foreground(cimg->background_color));
@@ -1145,7 +1157,7 @@ static void img_draw_image (struct f_data_c *fdatac, struct g_object_image *goi,
 
 	memcpy(&r, &fdatac->ses->term->dev->clip, sizeof(struct rect));
 	if (fdatac->vs->g_display_link && fdatac->active && fdatac->vs->current_link != -1 && fdatac->vs->current_link == goi->link_num) {
-		draw_frame_mark(drv, fdatac->ses->term->dev,x,y,goi->xw, 
+		draw_frame_mark(fdatac->ses->term->dev,x,y,goi->xw,
 			goi->yw,color_bg,color_fg,2);
 		restrict_clip_area(fdatac->ses->term->dev, &r, x + 2, y + 2, x + goi->xw - 2, y + goi->yw - 2);
 	}
@@ -1155,9 +1167,9 @@ static void img_draw_image (struct f_data_c *fdatac, struct g_object_image *goi,
 	if (img_process_download(goi, fdatac)) goto ret; /* Choked with data, will not
 							* draw. */
 	/* Now we will only draw... */
-	if (cimg->state<12){
-		draw_frame_mark(drv, fdatac->ses->term->dev,x,y,goi->xw, 
-			goi->yw,color_bg,color_fg,cimg->state&1);
+	if (!cimg || cimg->state<12){
+		draw_frame_mark(fdatac->ses->term->dev,x,y,goi->xw,
+			goi->yw,color_bg,color_fg,!cimg || cimg->state&1);
 	}else
 #ifdef DEBUG
 	if (cimg->state<16){
@@ -1175,6 +1187,10 @@ static void img_draw_image (struct f_data_c *fdatac, struct g_object_image *goi,
 #endif /* #ifdef DEBUG */
 	ret:;
 	drv->set_clip_area(fdatac->ses->term->dev, &r);
+#ifdef LINKS_TESTMODE_IMAGE_AUTO_EXIT
+	if (cimg->state & 1)
+		terminate_loop = 1;
+#endif
 }
 
 /* Prior to calling this function you have to fill out
@@ -1182,7 +1198,7 @@ static void img_draw_image (struct f_data_c *fdatac, struct g_object_image *goi,
  * image -> yw (<0 means not known)
  * image -> xyw meaning (MEANING_AUTOSCALE or MEANING_DIMS)
  * image -> background
- * 
+ *
  * The URL will not be freed.
  */
 static void find_or_make_cached_image(struct g_object_image *image, unsigned char *url,
@@ -1294,7 +1310,7 @@ struct g_object_image *insert_image(struct g_part *p, struct image_description *
 	*/
 	image->af=request_additional_file(current_f_data,im->url);
 	if (image->xw < 0 || image->yw < 0) image->af->unknown_image_size = 1;
-	image->background=p->root->bg->u.sRGB;
+	image->background=hack_rgb(p->root->bg->u.sRGB);
 
 	/* This supplies the result into image->cimg and global_cimg */
 	find_or_make_cached_image(image, im->url, d_opt->image_scale);
@@ -1341,21 +1357,21 @@ void change_image (struct g_object_image *goi, unsigned char *url, unsigned char
 int known_image_type(unsigned char *type)
 {
 #ifdef G
-	if (!strcasecmp(type, "image/png")) return 1;
-	if (!strcasecmp(type, "image/x-png")) return 1;
-	if (!strcasecmp(type, "image/gif")) return 1;
-	if (!strcasecmp(type, "image/x-xbitmap")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/png")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/x-png")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/gif")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/x-xbitmap")) return 1;
 #ifdef HAVE_JPEG
-	if (!strcasecmp(type, "image/jpeg")) return 1;
-	if (!strcasecmp(type, "image/jpg")) return 1;
-	if (!strcasecmp(type, "image/jpe")) return 1;
-	if (!strcasecmp(type, "image/pjpe")) return 1;
-	if (!strcasecmp(type, "image/pjpeg")) return 1;
-	if (!strcasecmp(type, "image/pjpg")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/jpeg")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/jpg")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/jpe")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/pjpe")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/pjpeg")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/pjpg")) return 1;
 #endif
 #ifdef HAVE_TIFF
-	if (!strcasecmp(type, "image/tiff")) return 1;
-	if (!strcasecmp(type, "image/tif")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/tiff")) return 1;
+	if (!strcasecmp(cast_const_char type, "image/tif")) return 1;
 #endif
 #endif
 	return 0;
