@@ -25,12 +25,15 @@ struct jerr_struct{
 
 static struct jerr_struct *global_jerr;
 static struct jpeg_decompress_struct *global_cinfo;
-static int mesg_unsup_emitted; /* Defaults to zero at program startup and once
+static int mesg_unsup_emitted = 0; /* Defaults to zero at program startup and once
 				* set is never reset back to zero */
 
 
+/*#include <jerror.h>*/
+
 METHODDEF(void) my_error_exit(j_common_ptr cinfo)
 {
+	/*fprintf(stderr, "jpeg error %d (%d)\n", cinfo->err->msg_code, JERR_OUT_OF_MEMORY);*/
 	longjmp(global_jerr->setjmp_buffer,2);
 }
 
@@ -56,7 +59,7 @@ METHODDEF(void) my_skip_input_data(j_decompress_ptr cinfo,long num_bytes)
 	 	/* We have to enter skipping state */
 	 	cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
 		((struct jpg_decoder *)(global_cimg->decoder))->skip_bytes
-  			=num_bytes-cinfo->src->bytes_in_buffer;
+  			=(int)(num_bytes-cinfo->src->bytes_in_buffer);
   		cinfo->src->bytes_in_buffer=0;
 	}
 	else
@@ -87,7 +90,6 @@ g19_2000:
 	if (setjmp(global_jerr->setjmp_buffer)){
 		jpeg_destroy_decompress(global_cinfo);
 		goto g19_2000;
-		return;
 	}
 	jpeg_stdio_src(global_cinfo,stdin);
 	global_cinfo->src->init_source=&nop;
@@ -146,6 +148,9 @@ static void gray_to_rgb(unsigned char *data, int pixels)
 /* Fixes returned data in case they are CMYK or grayscale. */
 static inline void fix_data( struct jpg_decoder *deco, int lines_read)
 {
+#ifdef __ICC
+	volatile	/* ICC bug */
+#endif
 	int a;
 
 	switch (global_cinfo->output_components){
@@ -221,7 +226,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 
 	/* Update the next input byte pointer for the decoder to continue at
 	 * the right position */
-	global_cinfo->src->next_input_byte=deco->jdata;
+	global_cinfo->src->next_input_byte=(void *)deco->jdata;
 
 	/* ...:::...:..:.:::.:.::::.::.:.:.:.::..::::.::::.:...: */
 	/* Update the length of data in the decoder buffer */
@@ -275,7 +280,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 			/* Let's make a decompression fatal error here */
 
 			if (!mesg_unsup_emitted){
-				fprintf(stderr,
+				error(
 			"Unsupported JPEG output components number: %d.\n",
 			cimg->buffer_bytes_per_pixel);
 				mesg_unsup_emitted=1;
@@ -286,17 +291,17 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 			 * understand and maintain.  If possible an alternative
 			 * should be used. Hahaha :) ;-)
 			 */
-			/* Free will makes people hard to understand 
-			 * and maintain. If possible an alternative should be 
+			/* Free will makes people hard to understand
+			 * and maintain. If possible an alternative should be
 			 * used.
 			 */
 			/* With our new LongJump(TM) your jumps will be longer
 			 * than with ordinary commercially available jumps.
 			 */
 		}
-		cimg->red_gamma=sRGB_gamma;
-		cimg->green_gamma=sRGB_gamma;
-		cimg->blue_gamma=sRGB_gamma;
+		cimg->red_gamma=(float)sRGB_gamma;
+		cimg->green_gamma=(float)sRGB_gamma;
+		cimg->blue_gamma=(float)sRGB_gamma;
 		/* This is defined in the JPEG standard somehow that sRGB
 		 * color space is used. */
 
@@ -325,7 +330,7 @@ susp0:
 		case 3:
 		/* jpeg_read_scanlines */
 			/* color */
-        	while (global_cinfo->output_scanline
+		while (global_cinfo->output_scanline
 			<global_cinfo->output_height){
 			int a, lines;
 
@@ -336,19 +341,19 @@ susp0:
 				*global_cinfo->output_width*cimg->
 					buffer_bytes_per_pixel;
 			}
-		
-         		if ((lines=
+
+	 		if ((lines=
 				jpeg_read_scanlines(
 				global_cinfo,deco->scanlines,1))){
 				/* Some lines were written into cimg buffer */
 				cimg->rows_added=1;
-				
+
 				fix_data(deco, lines);
 			}else{
 				/* No lines have been written into cimg
 				 * buffer */
-                	 	/* We are suspended and we want more data */
-         	 		goto susp0; /* Break the outer 
+			 	/* We are suspended and we want more data */
+	 	 		goto susp0; /* Break the outer
 					     * switch statement */
 			}
 		}
@@ -376,6 +381,22 @@ decoder_ended:
 		img_end(cimg);
 	}
 }
+
+void add_jpeg_version(unsigned char **s, int *l)
+{
+	add_to_str(s, l, cast_uchar "JPEG (");
+#if defined(JPEG_LIB_VERSION_MAJOR) && defined(JPEG_LIB_VERSION_MINOR)
+	add_num_to_str(s, l, JPEG_LIB_VERSION_MAJOR);
+	add_to_str(s, l, cast_uchar ".");
+	add_num_to_str(s, l, JPEG_LIB_VERSION_MINOR);
+#else
+	add_num_to_str(s, l, JPEG_LIB_VERSION / 10);
+	add_to_str(s, l, cast_uchar ".");
+	add_num_to_str(s, l, JPEG_LIB_VERSION % 10);
+#endif
+	add_to_str(s, l, cast_uchar ")");
+}
+
 #endif /* #ifdef HAVE_JPEG */
 
 #endif /* #ifdef G */
