@@ -48,8 +48,8 @@ void sig_tstp(struct terminal *t)
 {
 #if defined(SIGSTOP) && !defined(NO_CTRL_Z)
 #if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
-	pid_t pid = getpid();
-	pid_t newpid;
+	pid_t pid, newpid;
+	EINTRLOOP(pid, getpid());
 #endif
 	if (!F) {
 		block_itrm(1);
@@ -60,16 +60,24 @@ void sig_tstp(struct terminal *t)
 	}
 #endif
 #if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
-	if (!(newpid = fork())) {
+	EINTRLOOP(newpid, fork());
+	if (!newpid) {
 		while (1) {
+			int rr;
 			sleep(1);
-			kill(pid, SIGCONT);
+			EINTRLOOP(rr, kill(pid, SIGCONT));
 		}
 	}
 #endif
-	raise(SIGSTOP);
+	{
+		int rr;
+		EINTRLOOP(rr, raise(SIGSTOP));
+	}
 #if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
-	if (newpid != -1) kill(newpid, SIGKILL);
+	if (newpid != -1) {
+		int rr;
+		EINTRLOOP(rr, kill(newpid, SIGKILL));
+	}
 #endif
 #endif
 	if (fg_poll_timer != -1) kill_timer(fg_poll_timer);
@@ -93,7 +101,7 @@ static void poll_fg(void *t)
 	if (r == -2) {
 		/* This will unblock externally spawned viewer, if it exists */
 #ifdef SIGCONT
-		kill(0, SIGCONT);
+		EINTRLOOP(r, kill(0, SIGCONT));
 #endif
 	}
 }
@@ -151,7 +159,7 @@ void unhandle_terminal_signals(struct terminal *term)
 	if (fg_poll_timer != -1) kill_timer(fg_poll_timer), fg_poll_timer = -1;
 }
 
-void unhandle_basic_signals(struct terminal *term)
+static void unhandle_basic_signals(struct terminal *term)
 {
 	install_signal_handler(SIGHUP, NULL, NULL, 0);
 	if (!F) install_signal_handler(SIGINT, NULL, NULL, 0);
@@ -176,16 +184,17 @@ static int terminal_pipe[2];
 int attach_terminal(int in, int out, int ctl, void *info, int len)
 {
 	struct terminal *term;
-	fcntl(terminal_pipe[0], F_SETFL, O_NONBLOCK);
-	fcntl(terminal_pipe[1], F_SETFL, O_NONBLOCK);
+	int rs;
+	set_nonblock(terminal_pipe[0]);
+	set_nonblock(terminal_pipe[1]);
 	handle_trm(in, out, out, terminal_pipe[1], ctl, info, len);
 	mem_free(info);
 	if ((term = init_term(terminal_pipe[0], out, win_func))) {
 		handle_basic_signals(term);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
 		return terminal_pipe[1];
 	}
-	close(terminal_pipe[0]);
-	close(terminal_pipe[1]);
+	EINTRLOOP(rs, close(terminal_pipe[0]));
+	EINTRLOOP(rs, close(terminal_pipe[1]));
 	return -1;
 }
 
@@ -216,8 +225,11 @@ static void end_dump(struct object_request *r, void *p)
 			struct fragment *frag;
 			nextfrag:
 			foreach(frag, ce->frag) if (frag->offset <= dump_pos && frag->offset + frag->length > dump_pos) {
-				int l = frag->length - (dump_pos - frag->offset);
-				int w = hard_write(oh, frag->data + dump_pos - frag->offset, l);
+				off_t l;
+				int w;
+				l = frag->length - (dump_pos - frag->offset);
+				if (l >= MAXINT) l = MAXINT;
+				w = hard_write(oh, frag->data + dump_pos - frag->offset, (int)l);
 				if (w != l) {
 					detach_object_connection(r, dump_pos);
 					if (w < 0) fprintf(stderr, "Error writing to stdout: %s.\n", strerror(errno));
@@ -242,15 +254,11 @@ static void end_dump(struct object_request *r, void *p)
 		o.yw = 25;
 		o.col = 0;
 		o.cp = dump_codepage == -1 ? 0 : dump_codepage;
-		ds2do(&dds, &o);
+		ds2do(&dds, &o, 0);
 		o.plain = 0;
 		o.frames = 0;
 		o.js_enable = 0;
-		memcpy(&o.default_fg, &default_fg, sizeof(struct rgb));
-		memcpy(&o.default_bg, &default_bg, sizeof(struct rgb));
-		memcpy(&o.default_link, &default_link, sizeof(struct rgb));
-		memcpy(&o.default_vlink, &default_vlink, sizeof(struct rgb));
-		o.framename = "";
+		o.framename = cast_uchar "";
 		if (!(fd->f_data = cached_format_html(fd, r, r->url, &o, NULL))) goto term_1;
 		dump_to_file(fd->f_data, oh);
 		term_1:
@@ -277,36 +285,39 @@ static unsigned char init_b = 0;
 static void initialize_all_subsystems(void);
 static void initialize_all_subsystems_2(void);
 
+static void fixup_g(void)
+{
+	if (ggr_drv[0] || ggr_mode[0] || force_g) ggr = 1;
+	if (dmp) ggr = 0;
+}
+
 static void init(void)
 {
 	int uh;
 	void *info;
 	int len;
 	unsigned char *u;
-	initialize_all_subsystems();
+	int rs;
 
-	utf8_table=get_cp_index("UTF-8");
+	initialize_all_subsystems();
 
 /* OS/2 has some stupid bug and the pipe must be created before socket :-/ */
 	if (c_pipe(terminal_pipe)) {
-		error("ERROR: can't create pipe for internal communication");
-		retval = RET_FATAL;
-		goto ttt;
+		fatal_exit("ERROR: can't create pipe for internal communication");
 	}
 	if (!(u = parse_options(g_argc - 1, g_argv + 1))) {
 		retval = RET_SYNTAX;
 		goto ttt;
 	}
-	if (ggr_drv[0] || ggr_mode[0]) ggr = 1;
-	if (dmp) ggr = 0;
+	fixup_g();
 	if (!dmp && !ggr) {
 		init_os_terminal();
 	}
 	if (!ggr && !no_connect && (uh = bind_to_af_unix()) != -1) {
-		close(terminal_pipe[0]);
-		close(terminal_pipe[1]);
+		EINTRLOOP(rs, close(terminal_pipe[0]));
+		EINTRLOOP(rs, close(terminal_pipe[1]));
 		if (!(info = create_session_info(base_session, u, default_target, &len))) {
-			close(uh);
+			EINTRLOOP(rs, close(uh));
 			retval = RET_FATAL;
 			goto ttt;
 		}
@@ -314,9 +325,12 @@ static void init(void)
 		handle_trm(get_input_handle(), get_output_handle(), uh, uh, get_ctl_handle(), info, len);
 		handle_basic_signals(NULL);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
 		mem_free(info);
+#if defined(HAVE_MALLOC_TRIM)
+		malloc_trim(0);
+#endif
 		return;
 	}
-	if ((dds.assume_cp = get_cp_index("ISO-8859-1")) == -1) dds.assume_cp = 0;
+	if ((dds.assume_cp = get_cp_index(cast_uchar "ISO-8859-1")) == -1) dds.assume_cp = 0;
 	load_config();
 	init_b = 1;
 	init_bookmarks();
@@ -324,6 +338,7 @@ static void init(void)
 	load_url_history();
 	init_cookies();
 	u = parse_options(g_argc - 1, g_argv + 1);
+	fixup_g();
 	if (!u) {
 		ttt:
 		initialize_all_subsystems_2();
@@ -352,21 +367,20 @@ static void init(void)
 		}
 		initialize_all_subsystems_2();
 		if (!((info = create_session_info(base_session, u, default_target, &len)) && gf_val(attach_terminal(get_input_handle(), get_output_handle(), get_ctl_handle(), info, len), attach_g_terminal(info, len)) != -1)) {
-			retval = RET_FATAL;
-			terminate_loop = 1;
+			fatal_exit("Could not open initial session");
 		}
 	} else {
 		unsigned char *uu, *wd;
 		initialize_all_subsystems_2();
-		close(terminal_pipe[0]);
-		close(terminal_pipe[1]);
+		EINTRLOOP(rs, close(terminal_pipe[0]));
+		EINTRLOOP(rs, close(terminal_pipe[1]));
 		if (!*u) {
-			fprintf(stderr, "URL expected after %s\n.", dmp == D_DUMP ? "-dump" : "-source");
+			fprintf(stderr, "URL expected after %s\n", dmp == D_DUMP ? "-dump" : "-source");
 			retval = RET_SYNTAX;
 			goto tttt;
 		}
 		if (!(uu = translate_url(u, wd = get_cwd()))) uu = stracpy(u);
-		request_object(NULL, uu, NULL, PRI_MAIN, NC_RELOAD, end_dump, NULL, &dump_obj);
+		request_object(NULL, uu, NULL, PRI_MAIN, NC_RELOAD, ALLOW_ALL, end_dump, NULL, &dump_obj);
 		mem_free(uu);
 		if (wd) mem_free(wd);
 	}
@@ -375,6 +389,7 @@ static void init(void)
 /* Is called before gaphics driver init */
 static void initialize_all_subsystems(void)
 {
+	init_charset();
 	init_trans();
 	set_sigcld();
 	init_home();
@@ -411,6 +426,7 @@ static void terminate_all_subsystems(void)
 	abort_all_connections();
 
 	free_all_caches();
+	free_format_text_cache();
 	if (init_b) save_url_history();
 	free_history_lists();
 	free_term_specs();
@@ -426,9 +442,10 @@ static void terminate_all_subsystems(void)
 	free_strerror_buf();
 	shutdown_trans();
 	GF(shutdown_graphics());
-	terminate_osdep();
-	if (clipboard) mem_free(clipboard);
+	GF(free_dither());
+	if (clipboard) mem_free(clipboard), clipboard = NULL;
 	if (fg_poll_timer != -1) kill_timer(fg_poll_timer), fg_poll_timer = -1;
+	terminate_osdep();
 }
 
 int main(int argc, char *argv[])
@@ -436,6 +453,8 @@ int main(int argc, char *argv[])
 	g_argc = argc;
 	g_argv = (unsigned char **)argv;
 
+	init_heap();
+
 	init_os();
 
 	get_path_to_exe();
