@@ -6,8 +6,8 @@
 
 #include "links.h"
 
-static void objreq_end(struct status *, void *);
-static void object_timer(void *);
+static void objreq_end(struct status *, struct object_request *);
+static void object_timer(struct object_request *);
 
 
 static struct list_head requests = {&requests, &requests};
@@ -22,15 +22,13 @@ struct auth_dialog {
 	unsigned char passwd[MAX_UID_LEN];
 	unsigned char *realm;
 	int proxy;
-	tcount count;
 	unsigned char msg[1];
 };
 
 static inline struct object_request *find_rq(tcount c)
 {
 	struct object_request *rq;
-	struct list_head *lrq;
-	foreach(struct object_request, rq, lrq, requests) if (rq->count == c) return rq;
+	foreach(rq, requests) if (rq->count == c) return rq;
 	return NULL;
 }
 
@@ -78,13 +76,11 @@ static void auth_fn(struct dialog_data *dlg)
 
 static int auth_cancel(struct dialog_data *dlg, struct dialog_item_data *item)
 {
-	struct auth_dialog *a = dlg->dlg->udata;
-	struct object_request *rq = find_rq(a->count);
+	struct object_request *rq = find_rq((my_intptr_t)dlg->dlg->udata2);
 	if (rq) {
-		rq->hold = 0;
 		rq->state = O_OK;
-		if (rq->timer != NULL) kill_timer(rq->timer);
-		rq->timer = install_timer(0, object_timer, rq);
+		if (rq->timer != -1) kill_timer(rq->timer);
+		rq->timer = install_timer(0, (void (*)(void *))object_timer, rq);
 		if (!rq->ce) (rq->ce = rq->ce_internal)->refcount++;
 	}
 	cancel_dialog(dlg, item);
@@ -93,21 +89,22 @@ static int auth_cancel(struct dialog_data *dlg, struct dialog_item_data *item)
 
 static int auth_ok(struct dialog_data *dlg, struct dialog_item_data *item)
 {
-	struct auth_dialog *a = dlg->dlg->udata;
-	struct object_request *rq = find_rq(a->count);
+	struct object_request *rq = find_rq((my_intptr_t)dlg->dlg->udata2);
 	if (rq) {
+		struct auth_dialog *a = dlg->dlg->udata;
 		struct session *ses;
+		struct conv_table *ct;
 		int net_cp;
 		unsigned char *uid, *passwd;
 		get_dialog_data(dlg);
-		ses = list_struct(dlg->win->term->windows.prev, struct window)->data;
-		get_convert_table(rq->ce_internal->head, term_charset(dlg->win->term), ses->ds.assume_cp, &net_cp, NULL, ses->ds.hard_assume);
-		uid = convert(term_charset(dlg->win->term), net_cp, a->uid, NULL);
-		passwd = convert(term_charset(dlg->win->term), net_cp, a->passwd, NULL);
+		ses = ((struct window *)dlg->win->term->windows.prev)->data;
+		get_convert_table(rq->ce_internal->head, dlg->win->term->spec->charset, ses->ds.assume_cp, &net_cp, NULL, ses->ds.hard_assume);
+		ct = get_translation_table(dlg->win->term->spec->charset, net_cp);
+		uid = convert_string(ct, a->uid, (int)strlen(cast_const_char a->uid), NULL);
+		passwd = convert_string(ct, a->passwd, (int)strlen(cast_const_char a->passwd), NULL);
 		add_auth(rq->url, a->realm, uid, passwd, a->proxy);
 		mem_free(uid);
 		mem_free(passwd);
-		rq->hold = 0;
 		change_connection(&rq->stat, NULL, PRI_CANCEL);
 		load_url(rq->url, rq->prev_url, &rq->stat, rq->pri, NC_RELOAD, 0, 0, 0);
 	}
@@ -119,46 +116,46 @@ static int auth_window(struct object_request *rq, unsigned char *realm)
 {
 	unsigned char *host, *port;
 	struct dialog *d;
-	struct auth_dialog * volatile a;	/* avoid gcc-10 warning */
+	struct auth_dialog *a;
 	struct terminal *term;
+	struct conv_table *ct;
 	unsigned char *urealm;
 	struct session *ses;
-	int net_cp;
-	if (!(term = find_terminal(rq->term))) return -1;
-	ses = list_struct(term->windows.prev, struct window)->data;
-	get_convert_table(rq->ce_internal->head, term_charset(term), ses->ds.assume_cp, &net_cp, NULL, ses->ds.hard_assume);
+	foreach(term, terminals) if (rq->term == term->count) goto ok;
+	return -1;
+	ok:
+	ses = ((struct window *)term->windows.prev)->data;
+	ct = get_convert_table(rq->ce_internal->head, term->spec->charset, ses->ds.assume_cp, NULL, NULL, ses->ds.hard_assume);
 	if (rq->ce_internal->http_code == 407) {
-		unsigned char *h = get_proxy_string(rq->url);
-		if (!h) h = cast_uchar "";
-		host = display_host(term, h);
+		host = get_proxy_string(rq->url);
+		if (!host) host = cast_uchar "";
+		host = stracpy(host);
 	} else {
-		unsigned char *h = get_host_name(rq->url);
-		if (!h) return -1;
-		host = display_host(term, h);
-		mem_free(h);
+		host = get_host_name(rq->url);
+		if (!host) return -1;
 		if ((port = get_port_str(rq->url))) {
 			add_to_strn(&host, cast_uchar ":");
 			add_to_strn(&host, port);
 			mem_free(port);
 		}
 	}
-	urealm = convert(term_charset(term), net_cp, realm, NULL);
-	d = mem_alloc(sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct auth_dialog) + strlen(cast_const_char get_text_translation(TEXT_(T_ENTER_USERNAME), term)) + strlen(cast_const_char urealm) + 1 + strlen(cast_const_char get_text_translation(TEXT_(T_AT), term)) + strlen(cast_const_char host));
+	urealm = convert_string(ct, realm, (int)strlen(cast_const_char realm), NULL);
+	d = mem_alloc(sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct auth_dialog) + strlen(cast_const_char _(TEXT_(T_ENTER_USERNAME), term)) + strlen(cast_const_char urealm) + 1 + strlen(cast_const_char _(TEXT_(T_AT), term)) + strlen(cast_const_char host));
 	memset(d, 0, sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct auth_dialog));
 	a = (struct auth_dialog *)((unsigned char *)d + sizeof(struct dialog) + 5 * sizeof(struct dialog_item));
-	strcpy(cast_char a->msg, cast_const_char get_text_translation(TEXT_(T_ENTER_USERNAME), term));
+	strcpy(cast_char a->msg, cast_const_char _(TEXT_(T_ENTER_USERNAME), term));
 	strcat(cast_char a->msg, cast_const_char urealm);
 	if (*host) {
 		strcat(cast_char a->msg, "\n");
-		strcat(cast_char a->msg, cast_const_char get_text_translation(TEXT_(T_AT), term));
+		strcat(cast_char a->msg, cast_const_char _(TEXT_(T_AT), term));
 		strcat(cast_char a->msg, cast_const_char host);
 	}
 	mem_free(host);
 	mem_free(urealm);
 	a->proxy = rq->ce_internal->http_code == 407;
 	a->realm = stracpy(realm);
-	a->count = rq->count;
 	d->udata = a;
+	d->udata2 = (void *)(my_intptr_t)rq->count;
 	if (rq->ce_internal->http_code == 401) d->title = TEXT_(T_AUTHORIZATION_REQUIRED);
 	else d->title = TEXT_(T_PROXY_AUTHORIZATION_REQUIRED);
 	d->fn = auth_fn;
@@ -184,117 +181,6 @@ static int auth_window(struct object_request *rq, unsigned char *realm)
 	return 0;
 }
 
-#ifdef HAVE_SSL_CERTIFICATES
-
-struct cert_dialog {
-	tcount term;
-	unsigned char *host;
-	int bl;
-	int state;
-};
-
-static void cert_action(struct object_request *rq, int yes)
-{
-	if (yes > 0) {
-		rq->hold = 0;
-		change_connection(&rq->stat, NULL, PRI_CANCEL);
-		load_url(rq->url, rq->prev_url, &rq->stat, rq->pri, NC_CACHE, 0, 0, 0);
-	} else {
-		rq->hold = 0;
-		rq->dont_print_error = 1;
-		rq->state = O_FAILED;
-		if (rq->timer != NULL) kill_timer(rq->timer);
-		rq->timer = install_timer(0, object_timer, rq);
-	}
-}
-
-static void cert_forall(struct cert_dialog *cs, int yes)
-{
-	struct object_request *rq;
-	struct list_head *lrq;
-	if (yes > 0) {
-		add_blacklist_entry(cs->host, cs->bl);
-		del_blacklist_entry(cs->host, BL_AVOID_INSECURE);
-	}
-	if (yes < 0) {
-		add_blacklist_entry(cs->host, BL_AVOID_INSECURE);
-		del_blacklist_entry(cs->host, BL_IGNORE_CERTIFICATE);
-		del_blacklist_entry(cs->host, BL_IGNORE_DOWNGRADE);
-		del_blacklist_entry(cs->host, BL_IGNORE_CIPHER);
-	}
-	foreach(struct object_request, rq, lrq, requests) if (rq->term == cs->term && rq->hold == HOLD_CERT && rq->stat.state == cs->state) {
-		unsigned char *host = get_host_name(rq->url);
-		if (!strcmp(cast_const_char host, cast_const_char cs->host)) cert_action(rq, yes);
-		mem_free(host);
-	}
-}
-
-static void cert_yes(void *data)
-{
-	cert_forall((struct cert_dialog *)data, 1);
-}
-
-static void cert_no(void *data)
-{
-	cert_forall((struct cert_dialog *)data, 0);
-}
-
-static void cert_never(void *data)
-{
-	cert_forall((struct cert_dialog *)data, -1);
-}
-
-static int cert_compare(void *data1, void *data2)
-{
-	struct cert_dialog *cs1 = (struct cert_dialog *)data1;
-	struct cert_dialog *cs2 = (struct cert_dialog *)data2;
-	return !strcmp(cast_const_char cs1->host, cast_const_char cs2->host) && cs1->state == cs2->state;
-}
-
-static int cert_window(struct object_request *rq)
-{
-	struct terminal *term;
-	unsigned char *h, *host, *title, *text;
-	struct cert_dialog *cs;
-	struct memory_list *ml;
-	if (!(term = find_terminal(rq->term))) return -1;
-	h = get_host_name(rq->url);
-	if (get_blacklist_flags(h) & BL_AVOID_INSECURE) {
-		mem_free(h);
-		return -1;
-	}
-	cs = mem_alloc(sizeof(struct cert_dialog));
-	cs->term = rq->term;
-	cs->host = h;
-	cs->state = rq->stat.state;
-	if (rq->stat.state == S_INVALID_CERTIFICATE) {
-		title = TEXT_(T_INVALID_CERTIFICATE);
-		text = TEXT_(T_DOESNT_HAVE_A_VALID_CERTIFICATE);
-		cs->bl = BL_IGNORE_CERTIFICATE;
-	} else if (rq->stat.state == S_DOWNGRADED_METHOD) {
-		title = TEXT_(T_DOWNGRADED_METHOD);
-		text = TEXT_(T_USES_DOWNGRADED_METHOD);
-		cs->bl = BL_IGNORE_DOWNGRADE;
-	} else {
-		title = TEXT_(T_INSECURE_CIPHER);
-		text = TEXT_(T_USES_INSECURE_CIPHER);
-		cs->bl = BL_IGNORE_CIPHER;
-	}
-	host = display_host(term, h);
-	ml = getml(cs, h, host, NULL);
-	if (find_msg_box(term, title, cert_compare, cs)) {
-		freeml(ml);
-		return 0;
-	}
-	msg_box(term, ml,
-		title,
-		AL_CENTER, TEXT_(T_THE_SERVER_), host, text, MSG_BOX_END,
-		(void *)cs, 3, TEXT_(T_NO), cert_no, B_ESC, TEXT_(T_YES), cert_yes, B_ENTER, TEXT_(T_NEVER), cert_never, 0);
-	return 0;
-}
-
-#endif
-
 /* prev_url is a pointer to previous url or NULL */
 /* prev_url will NOT be deallocated */
 void request_object(struct terminal *term, unsigned char *url, unsigned char *prev_url, int pri, int cache, int allow_flags, void (*upcall)(struct object_request *, void *), void *data, struct object_request **rqp)
@@ -304,7 +190,7 @@ void request_object(struct terminal *term, unsigned char *url, unsigned char *pr
 	rq->state = O_WAITING;
 	rq->refcount = 1;
 	rq->term = term ? term->count : 0;
-	rq->stat.end = objreq_end;
+	rq->stat.end = (void (*)(struct status *, void *))objreq_end;
 	rq->stat.data = rq;
 	rq->orig_url = stracpy(url);
 	rq->url = stracpy(url);
@@ -312,8 +198,10 @@ void request_object(struct terminal *term, unsigned char *url, unsigned char *pr
 	rq->cache = cache;
 	rq->upcall = upcall;
 	rq->data = data;
-	rq->timer = NULL;
-	rq->last_update = get_time() - STAT_UPDATE_MAX;
+	rq->timer = -1;
+	rq->z = (uttime)get_time() - STAT_UPDATE_MAX;
+	rq->last_update = rq->z;
+	rq->last_bytes = 0;
 	if (rq->prev_url) mem_free(rq->prev_url);
 	rq->prev_url = stracpy(prev_url);
 	if (rqp) *rqp = rq;
@@ -337,32 +225,23 @@ static void set_ce_internal(struct object_request *rq)
 	}
 }
 
-static void objreq_end(struct status *stat, void *data)
+static void objreq_end(struct status *stat, struct object_request *rq)
 {
-	struct object_request *rq = (struct object_request *)data;
-
 	set_ce_internal(rq);
 
 	if (stat->state < 0) {
-#ifdef HAVE_SSL_CERTIFICATES
-		if (!stat->ce && rq->state == O_WAITING && (stat->state == S_INVALID_CERTIFICATE || stat->state == S_DOWNGRADED_METHOD || stat->state == S_INSECURE_CIPHER) && ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
-			if (!cert_window(rq)) {
-				rq->hold = HOLD_CERT;
-				rq->redirect_cnt = 0;
-				goto tm;
-			}
-		}
-#endif
 		if (stat->ce && rq->state == O_WAITING && stat->ce->redirect) {
 			if (rq->redirect_cnt++ < MAX_REDIRECTS) {
 				int cache, allow_flags;
-				unsigned char *u, *pos;
+				unsigned char *u, *p, *pos;
 				change_connection(stat, NULL, PRI_CANCEL);
 				u = join_urls(rq->url, stat->ce->redirect);
+				u = translate_hashbang(u);
 				if ((pos = extract_position(u))) {
 					if (rq->goto_position) mem_free(rq->goto_position);
 					rq->goto_position = pos;
 				}
+				if (!http_options.bug_302_redirect && !stat->ce->redirect_get && (p = cast_uchar strchr(cast_const_char u, POST_CHAR))) add_to_strn(&u, p);
 				cache = rq->cache;
 				if (cache < NC_RELOAD && (!strcmp(cast_const_char u, cast_const_char rq->url) || !strcmp(cast_const_char u, cast_const_char rq->orig_url) || rq->redirect_cnt >= MAX_CACHED_REDIRECTS)) cache = NC_RELOAD;
 				allow_flags = get_allow_flags(rq->url);
@@ -394,7 +273,6 @@ static void objreq_end(struct status *stat, void *data)
 			}
 			mem_free(user);
 			if (!auth_window(rq, realm)) {
-				rq->hold = HOLD_AUTH;
 				rq->redirect_cnt = 0;
 				mem_free(realm);
 				goto tm;
@@ -412,35 +290,33 @@ static void objreq_end(struct status *stat, void *data)
 		if (!rq->ce) (rq->ce = stat->ce)->refcount++;
 	}
 	tm:
-	if (rq->timer != NULL) kill_timer(rq->timer);
-	rq->timer = install_timer(0, object_timer, rq);
+	if (rq->timer != -1) kill_timer(rq->timer);
+	rq->timer = install_timer(0, (void (*)(void *))object_timer, rq);
 }
 
-static void object_timer(void *rq_)
+static void object_timer(struct object_request *rq)
 {
-	struct object_request *rq = (struct object_request *)rq_;
 	off_t last;
 
-	rq->timer = NULL;
-
 	set_ce_internal(rq);
 
 	last = rq->last_bytes;
 	if (rq->ce) rq->last_bytes = rq->ce->length;
-	if (rq->stat.state < 0 && !rq->hold && (!rq->ce_internal || !rq->ce_internal->redirect || rq->stat.state == S_CYCLIC_REDIRECT)) {
+	rq->timer = -1;
+	if (rq->stat.state < 0 && (!rq->ce_internal || (!rq->ce_internal->redirect && rq->ce_internal->http_code != 401 && rq->ce_internal->http_code != 407) || rq->stat.state == S_CYCLIC_REDIRECT)) {
 		if (rq->ce_internal && rq->stat.state != S_CYCLIC_REDIRECT) {
 			rq->state = rq->stat.state != S__OK ? O_INCOMPLETE : O_OK;
 		} else rq->state = O_FAILED;
 	}
 	if (rq->stat.state != S_TRANS) {
 		if (rq->stat.state >= 0)
-			rq->timer = install_timer(STAT_UPDATE_MAX, object_timer, rq);
-		rq->last_update = get_time() - STAT_UPDATE_MAX;
+			rq->timer = install_timer(STAT_UPDATE_MAX, (void (*)(void *))object_timer, rq);
+		rq->last_update = rq->z;
 		if (rq->upcall) rq->upcall(rq, rq->data);
 	} else {
-		uttime ct = get_time();
-		uttime t = ct - rq->last_update;
-		rq->timer = install_timer(STAT_UPDATE_MIN, object_timer, rq);
+		ttime ct = get_time();
+		ttime t = (uttime)ct - (uttime)rq->last_update;
+		rq->timer = install_timer(STAT_UPDATE_MIN, (void (*)(void *))object_timer, rq);
 		if (t >= STAT_UPDATE_MAX || (t >= STAT_UPDATE_MIN && rq->ce && rq->last_bytes > last)) {
 			rq->last_update = ct;
 			if (rq->upcall) rq->upcall(rq, rq->data);
@@ -455,7 +331,7 @@ void release_object_get_stat(struct object_request **rqq, struct status *news, i
 	*rqq = NULL;
 	if (--rq->refcount) return;
 	change_connection(&rq->stat, news, pri);
-	if (rq->timer != NULL) kill_timer(rq->timer);
+	if (rq->timer != -1) kill_timer(rq->timer);
 	if (rq->ce_internal) rq->ce_internal->refcount--;
 	if (rq->ce) rq->ce->refcount--;
 	mem_free(rq->orig_url);
@@ -474,12 +350,10 @@ void release_object(struct object_request **rqq)
 void detach_object_connection(struct object_request *rq, off_t pos)
 {
 	if (rq->state == O_WAITING || rq->state == O_FAILED) {
-		internal_error("detach_object_connection: no data received");
+		internal("detach_object_connection: no data received");
 		return;
 	}
-	if (rq->refcount == 1) {
-		detach_connection(&rq->stat, pos, 0, 1);
-	}
+	if (rq->refcount == 1) detach_connection(&rq->stat, pos);
 }
 
 void clone_object(struct object_request *rq, struct object_request **rqq)
