@@ -26,7 +26,7 @@
 #if defined(__EMX__)
 #define OS2
 #elif defined(_WIN32) || defined(__CYGWIN__)
-#define WIN
+#define WIN32
 #elif defined(__INTERIX)
 #define INTERIX
 #elif defined(__BEOS__) || defined(__HAIKU__)
@@ -60,6 +60,8 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define SHARED_CONFIG_DIR "/etc/"
 #ifdef HAVE_SYS_UN_H
 #define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
 #endif
 #define ASSOC_BLOCK
 #define ASSOC_CONS_XWIN
@@ -79,11 +81,10 @@ static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define DOS_FS_8_3
 #define NO_FILE_SECURITY
 #define NO_FORK_ON_EXIT
-#define USE_AF_UNIX
 #define ASSOC_CONS_XWIN
 #define DISABLE_SMB
 
-#elif defined(WIN)
+#elif defined(WIN32)
 
 static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define NEWLINE "\r\n"
@@ -106,16 +107,12 @@ static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #endif
 #if defined(HAVE_SYS_UN_H) && !defined(_UWIN)
 #define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
 #endif
-#if defined(_UWIN)
+#if defined(_UWIN) || 1
 #define USE_WIN32_HEAP
 #endif
-#ifdef USE_WIN32_HEAP
-#define NO_FORK_ON_EXIT
-#endif
-#ifndef __CYGWIN__
-#define OS_NO_SYSTEM_CHARSET
-#endif
 
 #elif defined(INTERIX)
 
@@ -132,10 +129,11 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define SHARED_CONFIG_DIR "/etc/"
 #ifdef HAVE_SYS_UN_H
 #define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
 #endif
 #define ASSOC_BLOCK
 #define ASSOC_CONS_XWIN
-#define OS_NO_SYSTEM_CHARSET
 
 #elif defined(BEOS)
 
@@ -152,10 +150,7 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NO_CTRL_Z
 #define SHARED_CONFIG_DIR "/etc/"
 #define NO_FORK_ON_EXIT
-#define USE_AF_UNIX
 #define ASSOC_BLOCK
-#define OS_NO_SYSTEM_CHARSET
-#define OS_NO_SYSTEM_LANGUAGE
 
 #include <sys/time.h>
 #include <sys/types.h>
@@ -178,10 +173,7 @@ static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define NO_CTRL_Z
 #define NO_FILE_SECURITY
 #define NO_FORK_ON_EXIT
-#define USE_AF_UNIX
 #define DISABLE_SMB
-#define OS_NO_SYSTEM_CHARSET
-#define OS_NO_SYSTEM_LANGUAGE
 
 #elif defined(ATHEOS)
 
@@ -196,7 +188,6 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define DEFAULT_SHELL "/bin/sh"
 #define GETSHELL getenv("SHELL")
 #define SHARED_CONFIG_DIR "/etc/"
-#define NO_FORK_ON_EXIT
 #define ASSOC_BLOCK
 
 #elif defined(SPAD)
@@ -209,11 +200,14 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define GETSHELL "LIB.:/SHELL.EXE"
 #define NO_CTRL_Z
 #define SHARED_CONFIG_DIR "ETC.:/"
-#define NO_FORK_ON_EXIT
+#ifdef HAVE_SYS_UN_H
+#define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
+#endif
 #define ASSOC_BLOCK
 #define ASSOC_CONS_XWIN
-#define OS_NO_SYSTEM_CHARSET
-#define OS_NO_SYSTEM_LANGUAGE
+#define NO_FORK_ON_EXIT
 
 #elif defined(OPENVMS)
 
@@ -228,13 +222,12 @@ static inline int dir_sep(unsigned char x) { return x == '/'; }
 #ifndef HAVE_PTHREADS
 #define HAVE_PTHREADS
 #endif
-#define NO_CTRL_Z
-#define NO_FORK_ON_EXIT
+#define DONT_USE_AF_UNIX
 #define ASSOC_BLOCK
 #define ASSOC_CONS_XWIN
+#define NO_CTRL_Z
+#define NO_FORK_ON_EXIT
 #define DISABLE_SMB
-#define OS_NO_SYSTEM_CHARSET
-#define OS_NO_SYSTEM_LANGUAGE
 
 #elif defined(DOS)
 
@@ -250,6 +243,7 @@ static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define DOS_FS
 #define DOS_FS_8_3
 #define NO_FILE_SECURITY
+#define DONT_USE_AF_UNIX
 #define NO_FORK_ON_EXIT
 #define DISABLE_SMB
 
