@@ -8,6 +8,10 @@
 #ifdef G
 #include "links.h"
 
+#ifdef HAVE_ENDIAN_H
+#include <endian.h>
+#endif /* #ifdef HAVE_ENDIAN_H */
+
 #ifdef REPACK_16
 #undef REPACK_16
 #endif /* #ifdef REPACK_16 */
@@ -26,11 +30,6 @@
 
 /* Decoder structs */
 
-struct png_decoder{
-	png_structp png_ptr;
-	png_infop info_ptr;
-};
-
 /* Warning for from-web PNG images */
 static void img_my_png_warning(png_structp a, png_const_charp b)
 {
@@ -59,7 +58,7 @@ static void png_info_callback(png_structp png_ptr, png_infop info_ptr)
 	color_type=png_get_color_type(png_ptr, info_ptr);
 	if (color_type == PNG_COLOR_TYPE_PALETTE)
 		png_set_expand(png_ptr);
-	if (color_type == PNG_COLOR_TYPE_GRAY &&
+    	if (color_type == PNG_COLOR_TYPE_GRAY &&
 		bit_depth < 8) png_set_expand(png_ptr);
 	if (png_get_valid(png_ptr, info_ptr,
 		PNG_INFO_tRNS)){
@@ -96,8 +95,8 @@ static void png_info_callback(png_structp png_ptr, png_infop info_ptr)
 		gamma=sRGB_gamma;
 	}
 	else
-	{
-		if (!png_get_gAMA(png_ptr, info_ptr, &gamma)){
+ 	{
+  		if (!png_get_gAMA(png_ptr, info_ptr, &gamma)){
 			gamma=sRGB_gamma;
 		}
 	}
@@ -223,7 +222,7 @@ error:
 	png_set_progressive_read_fn(png_ptr, NULL,
 				    png_info_callback, &png_row_callback,
 				    png_end_callback);
-	if (setjmp(png_jmpbuf(png_ptr))) goto error;
+   	if (setjmp(png_jmpbuf(png_ptr))) goto error;
 	decoder=mem_alloc(sizeof(*decoder));
 	decoder->png_ptr=png_ptr;
 	decoder->info_ptr=info_ptr;
@@ -234,33 +233,22 @@ void png_restart(struct cached_image *cimg, unsigned char *data, int length)
 {
 	png_structp png_ptr;
 	png_infop info_ptr;
-	volatile int h;
 
 #ifdef DEBUG
 	if (!cimg->decoder)
-		internal_error("decoder NULL in png_restart\n");
-
+		internal("decoder NULL in png_restart\n");
 #endif /* #ifdef DEBUG */
-	h = close_std_handle(2);
 	png_ptr=((struct png_decoder *)(cimg->decoder))->png_ptr;
 	info_ptr=((struct png_decoder *)(cimg->decoder))->info_ptr;
 	end_callback_hit=0;
-	if (setjmp(png_jmpbuf(png_ptr))) {
-		restore_std_handle(2, h);
+	if (setjmp(png_jmpbuf(png_ptr))){
 		img_end(cimg);
 		return;
 	}
 	png_process_data(png_ptr, info_ptr, data, length);
-	restore_std_handle(2, h);
 	if (end_callback_hit) img_end(cimg);
 }
 
-void png_destroy_decoder(struct cached_image *cimg)
-{
-	struct png_decoder *decoder = (struct png_decoder *)cimg->decoder;
-	png_destroy_read_struct(&decoder->png_ptr, &decoder->info_ptr, NULL);
-}
-
 void add_png_version(unsigned char **s, int *l)
 {
 	add_to_str(s, l, cast_uchar "PNG (");
@@ -269,7 +257,7 @@ void add_png_version(unsigned char **s, int *l)
 #else
 	add_to_str(s, l, cast_uchar PNG_LIBPNG_VER_STRING);
 #endif
-	add_chr_to_str(s, l, ')');
+	add_to_str(s, l, cast_uchar ")");
 }
 
 #endif /* #ifdef G */
