@@ -43,32 +43,60 @@ struct timer {
 static struct list_head timers = {&timers, &timers};
 
 
-ttime get_time(void)
+#ifndef OPENVMS
+void portable_sleep(unsigned msec)
 {
 	struct timeval tv;
-	gettimeofday(&tv, NULL);
-	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
+	int rs;
+	tv.tv_sec = msec / 1000;
+	tv.tv_usec = msec % 1000 * 1000;
+	EINTRLOOP(rs, select(0, NULL, NULL, NULL, &tv));
 }
+#endif
 
 int can_write(int fd)
 {
 	fd_set fds;
 	struct timeval tv = {0, 0};
+	int rs;
 	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_write: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		fatal_exit("too big handle %d", fd);
+	}
 	FD_SET(fd, &fds);
-	return select(fd + 1, NULL, &fds, NULL, &tv);
+	EINTRLOOP(rs, select(fd + 1, NULL, &fds, NULL, &tv));
+	if (rs < 0) fatal_exit("ERROR: select for write (%d) failed: %s", fd, strerror(errno));
+	return rs;
+}
+
+int can_read_timeout(int fd, int sec)
+{
+	fd_set fds;
+	struct timeval tv;
+	int rs;
+	tv.tv_sec = sec;
+	tv.tv_usec = 0;
+	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_read: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		fatal_exit("too big handle %d", fd);
+	}
+	FD_SET(fd, &fds);
+	EINTRLOOP(rs, select(fd + 1, &fds, NULL, NULL, &tv));
+	if (rs < 0) fatal_exit("ERROR: select for read (%d) failed: %s", fd, strerror(errno));
+	return rs;
 }
 
 int can_read(int fd)
 {
-	fd_set fds;
-	struct timeval tv = {0, 0};
-	FD_ZERO(&fds);
-	FD_SET(fd, &fds);
-	return select(fd + 1, &fds, NULL, NULL, &tv);
+	return can_read_timeout(fd, 0);
 }
 
-long select_info(int type)
+
+unsigned long select_info(int type)
 {
 	int i = 0, j;
 	struct cache_entry *ce;
@@ -81,7 +109,7 @@ long select_info(int type)
 			foreach(ce, timers) i++;
 			return i;
 		default:
-			internal("cache_info: bad request");
+			internal("select_info_info: bad request");
 	}
 	return 0;
 }
@@ -132,7 +160,12 @@ void check_bottom_halves(void)
 #ifdef DEBUG_CALLS
 	fprintf(stderr, "call: bh %p\n", fn);
 #endif
-	pr(fn(data)) {free_list(bottom_halves); return;};
+	pr(fn(data)) {
+#ifdef OOPS
+		free_list(bottom_halves);
+		return;
+#endif
+	};
 #ifdef DEBUG_CALLS
 	fprintf(stderr, "bh done\n");
 #endif
@@ -140,14 +173,14 @@ void check_bottom_halves(void)
 }
 
 #define CHK_BH if (!list_empty(bottom_halves)) check_bottom_halves()
-		
+
 static ttime last_time;
 
 static void check_timers(void)
 {
-	ttime interval = get_time() - last_time;
+	ttime interval = (uttime)get_time() - (uttime)last_time;
 	struct timer * volatile t;	/* volatile because of setjmp */
-	foreach(t, timers) t->interval -= interval;
+	foreach(t, timers) t->interval -= (uttime)interval;
 	/*ch:*/
 	foreach(t, timers) if (t->interval <= 0) {
 		struct timer *tt;
@@ -155,8 +188,10 @@ static void check_timers(void)
 		fprintf(stderr, "call: timer %p\n", t->func);
 #endif
 		pr(t->func(t->data)) {
+#ifdef OOPS
 			del_from_list((struct timer *)timers.next);
 			return;
+#endif
 		}
 #ifdef DEBUG_CALLS
 		fprintf(stderr, "timer done\n");
@@ -167,7 +202,7 @@ static void check_timers(void)
 		mem_free(t);
 		t = tt;
 	} else break;
-	last_time += interval;
+	last_time += (uttime)interval;
 }
 
 int install_timer(ttime t, void (*func)(void *), void *data)
@@ -207,15 +242,13 @@ void *get_handler(int fd, int tp)
 	if (fd < 0)
 		internal("get_handler: handle %d", fd);
 	if (fd >= (int)FD_SETSIZE) {
-		error("too big handle %d");
-		fatal_tty_exit();
-		exit(RET_FATAL);
+		fatal_exit("too big handle %d", fd);
 		return NULL;
 	}
 	switch (tp) {
-		case H_READ:	return threads[fd].read_func;
-		case H_WRITE:	return threads[fd].write_func;
-		case H_ERROR:	return threads[fd].error_func;
+		case H_READ:	return (void *)threads[fd].read_func;
+		case H_WRITE:	return (void *)threads[fd].write_func;
+		case H_ERROR:	return (void *)threads[fd].error_func;
 		case H_DATA:	return threads[fd].data;
 	}
 	internal("get_handler: bad type %d", tp);
@@ -224,12 +257,14 @@ void *get_handler(int fd, int tp)
 
 void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void *data)
 {
+#ifdef __GNU__
+	/* GNU Hurd has bugs w.r.t. exceptions */
+	error_func = NULL;
+#endif
 	if (fd < 0)
 		internal("set_handlers: handle %d", fd);
 	if (fd >= (int)FD_SETSIZE) {
-		error("too big handle %d");
-		fatal_tty_exit();
-		exit(RET_FATAL);
+		fatal_exit("too big handle %d", fd);
 		return;
 	}
 	threads[fd].read_func = read_func;
@@ -262,7 +297,32 @@ void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *),
 	}
 }
 
+void clear_events(int h, int blocking)
+{
+#if !defined(O_NONBLOCK) && !defined(FIONBIO)
+	blocking = 1;
+#endif
+	while (blocking ? can_read(h) : 1) {
+		unsigned char c[64];
+		int rd;
+		EINTRLOOP(rd, (int)read(h, c, sizeof c));
+		if (rd != sizeof c) break;
+	}
+}
+
+#if defined(NSIG) && NSIG > 32
+#define NUM_SIGNALS	NSIG
+#else
 #define NUM_SIGNALS	32
+#endif
+
+#ifndef NO_SIGNAL_HANDLERS
+
+static void clear_events_ptr(void *handle)
+{
+	clear_events((int)(my_intptr_t)handle, 0);
+}
+
 
 struct signal_handler {
 	void (*fn)(void *);
@@ -270,81 +330,141 @@ struct signal_handler {
 	int critical;
 };
 
-static int signal_mask[NUM_SIGNALS];
-static struct signal_handler signal_handlers[NUM_SIGNALS];
+static volatile int signal_mask[NUM_SIGNALS];
+static volatile struct signal_handler signal_handlers[NUM_SIGNALS];
 
 static int signal_pipe[2];
 
-static void signal_break(void *data)
-{
-	char c;
-	while (can_read(signal_pipe[0])) read(signal_pipe[0], &c, 1);
-}
-
 SIGNAL_HANDLER static void got_signal(int sig)
 {
+	void (*fn)(void *);
 	int sv_errno = errno;
 		/*fprintf(stderr, "ERROR: signal number: %d\n", sig);*/
+
+#if !defined(HAVE_SIGACTION)
+	do_signal(sig, got_signal);
+#endif
+
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		/*error("ERROR: bad signal number: %d", sig);*/
 		goto ret;
 	}
-	if (!signal_handlers[sig].fn) goto ret;
+	fn = signal_handlers[sig].fn;
+	if (!fn) goto ret;
 	if (signal_handlers[sig].critical) {
-		signal_handlers[sig].fn(signal_handlers[sig].data);
+		fn(signal_handlers[sig].data);
 		goto ret;
 	}
 	signal_mask[sig] = 1;
 	ret:
-	if (can_write(signal_pipe[1])) write(signal_pipe[1], "x", 1);
+	if (can_write(signal_pipe[1])) {
+		int wr;
+		EINTRLOOP(wr, (int)write(signal_pipe[1], "", 1));
+	}
 	errno = sv_errno;
 }
 
+#ifdef HAVE_SIGACTION
 static struct sigaction sa_zero;
+#endif
+
+#endif
 
 void install_signal_handler(int sig, void (*fn)(void *), void *data, int critical)
 {
+#if defined(NO_SIGNAL_HANDLERS)
+#elif defined(HAVE_SIGACTION)
+	int rs;
 	struct sigaction sa = sa_zero;
+	/*debug("install (%d) -> %p,%d", sig, fn, critical);*/
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
 	if (!fn) sa.sa_handler = SIG_IGN;
-	else sa.sa_handler = (void *)got_signal;
-#ifdef HAVE_SIGFILLSET
+	else sa.sa_handler = (void (*)(int))got_signal;
 	sigfillset(&sa.sa_mask);
-#else
-	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
-#endif
 	sa.sa_flags = SA_RESTART;
-	if (!fn) sigaction(sig, &sa, NULL);
+	if (!fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
 	signal_handlers[sig].fn = fn;
 	signal_handlers[sig].data = data;
 	signal_handlers[sig].critical = critical;
-	if (fn) sigaction(sig, &sa, NULL);
+	if (fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
+#else
+	if (!fn) do_signal(sig, SIG_IGN);
+	signal_handlers[sig].fn = fn;
+	signal_handlers[sig].data = data;
+	signal_handlers[sig].critical = critical;
+	if (fn) do_signal(sig, got_signal);
+#endif
 }
 
 void interruptible_signal(int sig, int in)
 {
+#if defined(HAVE_SIGACTION) && !defined(NO_SIGNAL_HANDLERS)
 	struct sigaction sa = sa_zero;
+	int rs;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
 	if (!signal_handlers[sig].fn) return;
-	sa.sa_handler = (void *)got_signal;
-#ifdef HAVE_SIGFILLSET
+	sa.sa_handler = (void (*)(int))got_signal;
 	sigfillset(&sa.sa_mask);
-#else
-	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
-#endif
 	if (!in) sa.sa_flags = SA_RESTART;
-	sigaction(sig, &sa, NULL);
+	EINTRLOOP(rs, sigaction(sig, &sa, NULL));
+#endif
+}
+
+static sigset_t sig_old_mask;
+static int sig_unblock = 0;
+
+void block_signals(int except1, int except2)
+{
+	int rs;
+	sigset_t mask;
+	sigfillset(&mask);
+#ifdef HAVE_SIGDELSET
+	if (except1) sigdelset(&mask, except1);
+	if (except2) sigdelset(&mask, except2);
+#ifdef SIGILL
+	sigdelset(&mask, SIGILL);
+#endif
+#ifdef SIGABRT
+	sigdelset(&mask, SIGABRT);
+#endif
+#ifdef SIGFPE
+	sigdelset(&mask, SIGFPE);
+#endif
+#ifdef SIGSEGV
+	sigdelset(&mask, SIGSEGV);
+#endif
+#ifdef SIGBUS
+	sigdelset(&mask, SIGBUS);
+#endif
+#else
+	if (except1 || except2) return;
+#endif
+	EINTRLOOP(rs, do_sigprocmask(SIG_BLOCK, &mask, &sig_old_mask));
+	if (!rs) sig_unblock = 1;
+}
+
+void unblock_signals(void)
+{
+	int rs;
+	if (sig_unblock) {
+		EINTRLOOP(rs, do_sigprocmask(SIG_SETMASK, &sig_old_mask, NULL));
+		sig_unblock = 0;
+	}
 }
 
 static int check_signals(void)
 {
-	int i, r = 0;
+	int r = 0;
+#ifndef NO_SIGNAL_HANDLERS
+	int i;
 	for (i = 0; i < NUM_SIGNALS; i++)
 		if (signal_mask[i]) {
 			signal_mask[i] = 0;
@@ -352,7 +472,11 @@ static int check_signals(void)
 #ifdef DEBUG_CALLS
 				fprintf(stderr, "call: signal %d -> %p\n", i, signal_handlers[i].fn);
 #endif
-				pr(signal_handlers[i].fn(signal_handlers[i].data)) return 1;
+				pr(signal_handlers[i].fn(signal_handlers[i].data)) {
+#ifdef OOPS
+					return 1;
+#endif
+}
 #ifdef DEBUG_CALLS
 				fprintf(stderr, "signal done\n");
 #endif
@@ -360,16 +484,20 @@ static int check_signals(void)
 			CHK_BH;
 			r = 1;
 		}
+#endif
 	return r;
 }
 
+#ifdef SIGCHLD
 static void sigchld(void *p)
 {
+	pid_t pid;
 #ifndef WNOHANG
-	wait(NULL);
+	EINTRLOOP(pid, wait(NULL));
 #else
-	while (waitpid(-1, NULL, WNOHANG) > 0)
-		;
+	do {
+		EINTRLOOP(pid, waitpid(-1, NULL, WNOHANG));
+	} while (pid > 0);
 #endif
 }
 
@@ -377,30 +505,42 @@ void set_sigcld(void)
 {
 	install_signal_handler(SIGCHLD, sigchld, NULL, 1);
 }
+#else
+void set_sigcld(void)
+{
+}
+#endif
 
 int terminate_loop = 0;
 
 void select_loop(void (*init)(void))
 {
 	struct stat st;
-	if (stat(".", &st) && getenv("HOME")) chdir(getenv("HOME"));
+	int rs;
+	EINTRLOOP(rs, stat(".", &st));
+	if (rs && getenv("HOME"))
+		EINTRLOOP(rs, chdir(getenv("HOME")));
+#if !defined(NO_SIGNAL_HANDLERS)
+#if defined(HAVE_SIGACTION)
 	memset(&sa_zero, 0, sizeof sa_zero);
-	memset(signal_mask, 0, sizeof signal_mask);
-	memset(signal_handlers, 0, sizeof signal_handlers);
+#endif
+	memset((void *)signal_mask, 0, sizeof signal_mask);
+	memset((void *)signal_handlers, 0, sizeof signal_handlers);
+#endif
 	FD_ZERO(&w_read);
 	FD_ZERO(&w_write);
 	FD_ZERO(&w_error);
 	w_max = 0;
 	last_time = get_time();
 	ignore_signals();
+#if !defined(NO_SIGNAL_HANDLERS)
 	if (c_pipe(signal_pipe)) {
-		error("ERROR: can't create pipe for signal handling");
-		retval = RET_FATAL;
-		return;
+		fatal_exit("ERROR: can't create pipe for signal handling");
 	}
-	fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK);
-	fcntl(signal_pipe[1], F_SETFL, O_NONBLOCK);
-	set_handlers(signal_pipe[0], signal_break, NULL, NULL, NULL);
+	set_nonblock(signal_pipe[0]);
+	set_nonblock(signal_pipe[1]);
+	set_handlers(signal_pipe[0], clear_events_ptr, NULL, NULL, (void *)(my_intptr_t)signal_pipe[0]);
+#endif
 	init();
 	CHK_BH;
 	while (!terminate_loop) {
@@ -411,8 +551,6 @@ void select_loop(void (*init)(void))
 		check_signals();
 		check_timers();
 		check_timers();
-		check_timers();
-		check_timers();
 		if (!F) redraw_all_terminals();
 #ifdef OS_BAD_SIGNALS
 	/* Cygwin has buggy signals that sometimes don't interrupt select.
@@ -422,13 +560,13 @@ void select_loop(void (*init)(void))
 		tm = &tv;
 #endif
 		if (!list_empty(timers)) {
-			ttime tt = ((struct timer *)timers.next)->interval + 1;
+			ttime tt = (uttime)((struct timer *)timers.next)->interval + 1;
 			if (tt < 0) tt = 0;
 #ifdef OS_BAD_SIGNALS
 			if (tt < 1000)
 #endif
 			{
-				tv.tv_sec = tt / 1000;
+				tv.tv_sec = tt / 1000 < MAXINT ? (int)(tt / 1000) : MAXINT;
 				tv.tv_usec = (tt % 1000) * 1000;
 				tm = &tv;
 			}
@@ -463,9 +601,7 @@ void select_loop(void (*init)(void))
 			fprintf(stderr, "select intr\n");
 #endif
 			if (errno != EINTR) {
-				error("ERROR: select failed: %s", strerror(errno));
-				fatal_tty_exit();
-				exit(RET_FATAL);
+				fatal_exit("ERROR: select failed: %s", strerror(errno));
 			}
 			continue;
 		}
