@@ -56,7 +56,7 @@ static int svgalib_timer_id;
 
 /*------------------------STRUCTURES-------------------------------------------*/
 struct modeline{
-	unsigned char *name;
+	char *name;
 	int number;
 };
 
@@ -580,7 +580,7 @@ static void redraw_mouse(void);
  * Good white purity
  * Correct rounding and dithering prediction
  * And this is the cabbala:
- * 063 021 063 
+ * 063 021 063
  * 009 009 021
  * 255 085 255
  * 036 036 084
@@ -617,19 +617,18 @@ static void generate_palette(struct irgb *palette)
 		}else{
 			/* palette_depth==8 */
 			for (a=0;a<256;a++,palette++){
-				palette->r=((a>>5)&7)*36;
-				palette->g=((a>>2)&7)*36;
-				palette->b=(a&3)*84;
+				palette->r=(((a>>5)&7)*255+3)/7;
+				palette->g=(((a>>2)&7)*255+3)/7;
+				palette->b=(a&3)*85;
 			}
 		}
 	}
-			
 }
 
 static void set_palette(struct irgb *palette)
 {
 	int r,g,b,c;
-	
+
 	for (c=0;c<vga_colors;c++){
 		r=palette[c].r;
 		g=palette[c].g;
@@ -654,10 +653,11 @@ static void svga_shutdown_driver(void)
 #ifndef __SPAD__
 	kill_timer(svgalib_timer_id);
 #endif
+	vga_setmode_retry:
 	if (vga_setmode(TEXT) < 0) {
-		error("ERROR: vga_setmode failed");
-		fatal_tty_exit();
-		exit(RET_FATAL);
+		if (out_of_memory(0, NULL, 0))
+			goto vga_setmode_retry;
+		fatal_exit("ERROR: vga_setmode failed");
 	}
 	svgalib_free_trm(svgalib_kbd);
 	if (svga_driver_param)mem_free(svga_driver_param);
@@ -674,14 +674,14 @@ static void svga_unregister_bitmap(struct bitmap *bmp)
 }
 
 #define SYNC if (do_sync) vga_accel(ACCEL_SYNC);
-	
+
 /* This assures that x, y, xs, ys, data will be sane according to clipping
  * rectangle. If nothing lies within this rectangle, the current function
  * returns. The data pointer is automatically advanced by this macro to reflect
  * the right position to start with inside the bitmap. */
 #define	CLIP_PREFACE \
 	int xs=hndl->x,ys=hndl->y;\
-	char *data=hndl->data;\
+	unsigned char *data=hndl->data;\
 	int mouse_hidden;\
 \
  	TEST_INACTIVITY\
@@ -708,7 +708,7 @@ static inline void draw_bitmap_accel(struct graphics_device *dev,
 	struct bitmap* hndl, int x, int y)
 {
 	CLIP_PREFACE
-	
+
 	if (xs*fb_pixelsize==hndl->skip) vga_accel(ACCEL_PUTIMAGE,x,y,xs,ys,data);
 	else for(;ys;ys--){
 		vga_accel(ACCEL_PUTIMAGE,x,y,xs,1,data);
@@ -731,7 +731,7 @@ static inline void paged_memcpy(int lina, unsigned char *src, int len)
 	int page=lina>>16;
 	int paga=lina&0xffff;
 	int remains;
-	
+
 	my_setpage(page);
 	remains=65536-paga;
 	again:
@@ -778,8 +778,6 @@ static inline void draw_bitmap_paged(struct graphics_device *dev, struct bitmap*
 		scr_start+=vga_linewidth;
 	}
 	END_MOUSE
-		
-		
 }
 
 static inline void draw_bitmap_linear(struct graphics_device *dev,struct bitmap* hndl, int x, int y)
@@ -813,20 +811,20 @@ static inline void draw_bitmap_linear(struct graphics_device *dev,struct bitmap*
 	if (bottom>dev->clip.y2) bottom=dev->clip.y2;\
 	TEST_MOUSE(left,right,top,bottom)
 
-	
+
 static void fill_area_accel_box(struct graphics_device *dev, int left, int top, int right, int bottom, long color)
 {
 	FILL_CLIP_PREFACE
-	
+
 	vga_accel(ACCEL_SETFGCOLOR,color);
 	vga_accel(ACCEL_FILLBOX,left,top,right-left,bottom-top);
 	END_MOUSE
 }
-	
+
 static void fill_area_accel_lines(struct graphics_device *dev, int left, int top, int right, int bottom, long color)
 {
 	int y;
-	
+
 	FILL_CLIP_PREFACE
 	vga_accel(ACCEL_SETFGCOLOR,color);
 	for (y=top;y<bottom;y++) vga_accel(ACCEL_DRAWLINE,left,y,right-1,y);
@@ -864,7 +862,7 @@ static void pixel_set_paged(int lina, int len, void * color)
 
 	memcpy(color0,color,vga_bytes);
 	memcpy(color0+vga_bytes,color,vga_bytes-1);
-	
+
 	my_setpage(page);
 	again:
 	if (remains>=len){
@@ -906,7 +904,7 @@ static void fill_area_paged(struct graphics_device *dev, int left, int top, int
 	}
 	END_MOUSE
 }
-	
+
 #define HLINE_CLIP_PREFACE \
 	int mouse_hidden;\
 	TEST_INACTIVITY\
@@ -915,7 +913,7 @@ static void fill_area_paged(struct graphics_device *dev, int left, int top, int
 	if (left<dev->clip.x1) left=dev->clip.x1;\
 	if (right>dev->clip.x2) right=dev->clip.x2;\
 	TEST_MOUSE (left,right,y,y+1)
-	
+
 #define VLINE_CLIP_PREFACE \
 	int mouse_hidden;\
 	TEST_INACTIVITY\
@@ -924,7 +922,7 @@ static void fill_area_paged(struct graphics_device *dev, int left, int top, int
 	if (top<dev->clip.y1) top=dev->clip.y1;\
 	if (bottom>dev->clip.y2) bottom=dev->clip.y2;\
 	TEST_MOUSE(x,x+1,top,bottom)
-	
+
 static void draw_hline_accel_line(struct graphics_device *dev, int left, int y, int right, long color)
 {
 	HLINE_CLIP_PREFACE
@@ -965,7 +963,7 @@ static void draw_hline_linear(struct graphics_device *dev, int left, int y, int
 {
 	unsigned char *dest;
 	HLINE_CLIP_PREFACE
-	SYNC	
+	SYNC
 	dest=my_graph_mem+y*vga_linewidth+left*vga_bytes;
 	pixel_set(dest,(right-left)*vga_bytes,&color);
 	END_MOUSE
@@ -989,7 +987,7 @@ static void draw_hline_paged(struct graphics_device *dev, int left, int y, int r
 {
 	int dest;
 	int len;
-	HLINE_CLIP_PREFACE	
+	HLINE_CLIP_PREFACE
 	SYNC
 	len=(right-left)*vga_bytes;
 
@@ -1002,7 +1000,7 @@ static void draw_hline_paged(struct graphics_device *dev, int left, int y, int r
 static void draw_vline_paged_1(struct graphics_device *dev, int x, int top, int bottom, long color)
 {
 	int dest,n, page,paga,remains;
-	int byte=*(char *)&color;
+	int byte=*(unsigned char *)&color;
 	VLINE_CLIP_PREFACE;
 	SYNC
 	dest=top*vga_linewidth+x;
@@ -1031,7 +1029,7 @@ static void draw_vline_paged_1(struct graphics_device *dev, int x, int top, int
 		goto again;
 	}
 }
-			
+
 #ifdef t2c
 /* Works only for pixel length 2 */
 static void draw_vline_paged_2(struct graphics_device *dev, int x, int top, int bottom, long color)
@@ -1067,7 +1065,7 @@ static void draw_vline_paged_2(struct graphics_device *dev, int x, int top, int
 	}
 }
 #endif /* #ifdef t2c */
-			
+
 #ifdef t4c
 /* Works only for pixel length 4 */
 static void draw_vline_paged_4(struct graphics_device *dev, int x, int top, int bottom, long color)
@@ -1104,7 +1102,7 @@ static void draw_vline_paged_4(struct graphics_device *dev, int x, int top, int
 	}
 }
 #endif /*t4c*/
-			
+
 /* Works only for pixel lengths power of two */
 static void draw_vline_paged_aligned(struct graphics_device *dev, int x, int top, int bottom, long color)
 {
@@ -1137,7 +1135,7 @@ static void draw_vline_paged_aligned(struct graphics_device *dev, int x, int top
 		goto again;
 	}
 }
-			
+
 /* Works for any pixel length */
 static void draw_vline_paged(struct graphics_device *dev, int x, int top, int bottom, long color)
 {
@@ -1164,7 +1162,7 @@ static void draw_vline_paged(struct graphics_device *dev, int x, int top, int bo
 	if (remains<vga_bytes){
 		memcpy(my_graph_mem+paga,&color,remains);
 		vga_setpage(++page);
-		memcpy(my_graph_mem,(char *)&color+remains,vga_bytes-remains);
+		memcpy(my_graph_mem,(unsigned char *)&color+remains,vga_bytes-remains);
 		lina+=vga_linewidth;
 		n--;
 		if (!n) goto end;
@@ -1191,7 +1189,7 @@ end:
 		goto again;
 	}
 }
-		
+
 #define HSCROLL_CLIP_PREFACE \
 	int mouse_hidden;\
 	TEST_INACTIVITY_0\
@@ -1199,8 +1197,8 @@ end:
 	if (sc>(dev->clip.x2-dev->clip.x1)||-sc>(dev->clip.x2-dev->clip.x1))\
 		return 1;\
 	TEST_MOUSE (dev->clip.x1,dev->clip.x2,dev->clip.y1,dev->clip.y2)
-		
-							
+
+
 /* When sc is <0, moves the data left. Scrolls the whole clip window */
 static int hscroll_accel(struct graphics_device *dev, struct rect_set **ignore, int sc)
 {
@@ -1225,7 +1223,7 @@ static int hscroll_accel(struct graphics_device *dev, struct rect_set **ignore,
 	if (!sc) return 0;\
 	if (sc>dev->clip.y2-dev->clip.y1||-sc>dev->clip.y2-dev->clip.y1) return 1;\
 	TEST_MOUSE (dev->clip.x1, dev->clip.x2, dev->clip.y1, dev->clip.y2)
-	
+
 /* Positive sc means data move down */
 static int vscroll_accel(struct graphics_device *dev, struct rect_set **ignore, int sc)
 {
@@ -1249,7 +1247,7 @@ static int hscroll_scansegment(struct graphics_device *dev, struct rect_set **ig
 	int y;
 	int len;
 	HSCROLL_CLIP_PREFACE
-	SYNC	
+	SYNC
 	if (sc>0){
 		/* Right */
 		len=dev->clip.x2-dev->clip.x1-sc;
@@ -1359,7 +1357,7 @@ static inline void get_row(unsigned char *bptr, int lina, int len)
 	int page=lina>>16;
 	int paga=lina&0xffff;
 	int remains;
-	
+
 	my_setpage(page);
 	remains=65536-paga;
 	again:
@@ -1418,7 +1416,7 @@ static int hscroll_paged(struct graphics_device *dev, struct rect_set **ignore,
 	int len;
 
 	HSCROLL_CLIP_PREFACE
-	SYNC	
+	SYNC
 	if (sc>0){
 		len=(dev->clip.x2-dev->clip.x1-sc)*vga_bytes;
 		src=vga_linewidth*dev->clip.y1+dev->clip.x1*vga_bytes;
@@ -1444,27 +1442,12 @@ static int hscroll_paged(struct graphics_device *dev, struct rect_set **ignore,
 	return 1;
 }
 
-static void svga_set_clip_area(struct graphics_device *dev, struct rect *r)
-{
-	memcpy(&dev->clip, r, sizeof(struct rect));
-	if (dev->clip.x1>=dev->clip.x2||dev->clip.y2<=dev->clip.y1||dev->clip.y2<=0||dev->clip.x2<=0||dev->clip.x1>=xsize
-			||dev->clip.y1>=ysize){
-		/* Empty region */
-		dev->clip.x1=dev->clip.x2=dev->clip.y1=dev->clip.y2=0;
-	}else{
-		if (dev->clip.x1<0) dev->clip.x1=0;
-		if (dev->clip.x2>xsize) dev->clip.x2=xsize;
-		if (dev->clip.y1<0) dev->clip.y1=0;
-		if (dev->clip.y2>ysize) dev->clip.y2=ysize;
-	}
-}
-
 /* For modes where video memory is not directly accessible through svgalib */
 static inline void fill_area_drawscansegment(struct graphics_device *dev, int left, int top, int right, int bottom, long color)
 {
 	int xs;
 	int col=*(unsigned char *)&color;
-	
+
 	FILL_CLIP_PREFACE
 	SYNC
 	xs=right-left;
@@ -1552,7 +1535,6 @@ static void svga_draw_bitmaps(struct graphics_device *dev, struct bitmap **hndls
 		n--;
 		hndls++;
 	}
-		
 }
 #endif
 
@@ -1566,6 +1548,7 @@ static void alloc_scroll_buffer(void)
 
 static void setup_functions(void)
 {
+	svga_driver.flags &= ~GD_DONT_USE_SCROLL;
 
 	if (accel_avail&ACCELFLAG_SETMODE){
 		do_sync=1;
@@ -1604,13 +1587,13 @@ static void setup_functions(void)
 		}else{
 			svga_driver.draw_bitmap=draw_bitmap_paged;
 		}
-		
+
 		if (accel_avail&ACCELFLAG_FILLBOX) svga_driver.fill_area=fill_area_accel_box;
 		else if (accel_avail&ACCELFLAG_DRAWLINE) svga_driver.fill_area=fill_area_accel_lines;
 		else if (vga_linear) svga_driver.fill_area=fill_area_linear;
 		else if (mode_x) svga_driver.fill_area=fill_area_drawscansegment;
 		else svga_driver.fill_area=fill_area_paged;
-		
+
 		if (accel_avail&ACCELFLAG_DRAWLINE){
 			svga_driver.draw_hline=draw_hline_accel_line;
 			svga_driver.draw_vline=draw_vline_accel_line;
@@ -1721,8 +1704,9 @@ void dump_mode_info_into_file(vga_modeinfo* i)
 }
 #endif
 
-static void svgalib_key_in(void *p, struct event *ev, int size)
+static void svgalib_key_in(void *p, unsigned char *ev_, int size)
 {
+	struct event *ev = (struct event *)(void *)ev_;
 	if (size != sizeof(struct event)) return;
 	if (ev->ev == EV_ABORT) terminate_loop = 1;
 	if (ev->ev != EV_KBD) return;
@@ -1764,7 +1748,7 @@ static void mouse_event_handler(int button, int dx, int dy, int dz, int drx, int
 	int moved,old_mouse_x,old_mouse_y;
 	void (*mh)(struct graphics_device *, int, int, int);
 	struct graphics_device *cd=current_virtual_device;
-	
+
 	mh=cd?cd->mouse_handler:NULL;
 	old_mouse_x=mouse_x;
 	old_mouse_y=mouse_y;
@@ -1780,13 +1764,13 @@ static void mouse_event_handler(int button, int dx, int dy, int dz, int drx, int
 	redraw_mouse();
 
 	moved=(old_mouse_x!=mouse_x||old_mouse_y!=mouse_y);
-	
+
 	/* Test movement without buttons */
 	if (!(mouse_buttons & BUTTON_MASK) && moved) {
 		mouse_aggregate_flag=1;
 		mouse_aggregate_action=B_MOVE;
 	}
-			
+
 	/* Test presses */
 	if ((button&MOUSE_LEFTBUTTON)&&!(mouse_buttons&MOUSE_LEFTBUTTON)){
 		mouse_aggregate_flush();
@@ -1839,7 +1823,7 @@ static void mouse_event_handler(int button, int dx, int dy, int dz, int drx, int
 		mouse_aggregate_flush();
 		/*if (mh) mh(cd,mouse_x, mouse_y,B_SIXTH|B_UP);*/
 	}
-	
+
 	if (drx < 0 && mh) mh(cd, mouse_x, mouse_y, B_MOVE | B_WHEELUP);
 	if (drx > 0 && mh) mh(cd, mouse_x, mouse_y, B_MOVE | B_WHEELDOWN);
 
@@ -1926,7 +1910,7 @@ static void render_mouse_arrow(void)
 {
 	int x,y, reg0, reg1;
 	unsigned char *mouse_ptr=mouse_buffer;
-	unsigned *arrow_ptr=arrow;
+	const unsigned *arrow_ptr=arrow;
 
 	for (y=arrow_height;y;y--){
 		reg0=*arrow_ptr;
@@ -1952,7 +1936,7 @@ static void place_mouse(void)
 	bmp.x=arrow_width;
 	bmp.y=arrow_height;
 	bmp.skip=arrow_width*fb_pixelsize;
-	bmp.data=mouse_buffer;	
+	bmp.data=mouse_buffer;
 	{
 		struct graphics_device * current_graphics_device_backup;
 
@@ -2043,7 +2027,7 @@ static inline void place_mouse_composite(void)
 	if (vga_colors>=256)
 		vga_waitretrace();
 	*/
-	
+
 	if (mouse_bottom>ysize) mouse_bottom=ysize;
 	if (background_bottom>ysize) background_bottom=ysize;
 
@@ -2062,7 +2046,7 @@ static inline void place_mouse_composite(void)
 				,background_top,background_length*fb_pixelsize);
 			background_ptr+=skip;
 		}
-			
+
 	}else if (background_top>mouse_top){
 		/* Draw the mouse */
 		mouse_length=mouse_right>xsize
@@ -2085,10 +2069,10 @@ static inline void place_mouse_composite(void)
 			mouse_ptr+=skip;
 			background_ptr+=skip;
 		}
-			
+
 	}else{
 		int l1, l2, l3;
-		
+
 		/* Draw mouse, background */
 		mouse_length=mouse_right>xsize?xsize-mouse_left:arrow_width;
 		background_length=background_right-mouse_right;
@@ -2149,8 +2133,8 @@ static inline void redraw_mouse_sophisticated(void)
 	background_y=new_background_y;
 }
 
-static void redraw_mouse(void){
-	
+static void redraw_mouse(void)
+{
 	if (flags) return; /* We are not drawing */
 	if (mouse_x!=background_x||mouse_y!=background_y){
 		if (RECTANGLES_INTERSECT(
@@ -2212,13 +2196,11 @@ static void setup_mode(int mode)
 	mode_x=!!(i->flags&IS_MODEX);
 	vga_linear=!!(i->flags&IS_LINEAR);
 	/*
-	if (!vga_linear && i->flags&CAPABLE_LINEAR && 0<=vga_setlinearaddressing()) vga_linear=1; 
+	if (!vga_linear && i->flags&CAPABLE_LINEAR && 0<=vga_setlinearaddressing()) vga_linear=1;
 	*/
 	my_graph_mem=vga_getgraphmem();
 	svga_driver.x = xsize=i->width;
 	svga_driver.y = ysize=i->height;
-	aspect_native=(196608*xsize+(ysize<<1))/(ysize<<2);
-	aspect=aspect_native*bfu_aspect+0.5;
 	vga_colors=i->colors;
 	if (xsize==320&&ysize==200&&vga_colors==256) vga_linear=1; /* The mode
 		does not need to page :-) */
@@ -2304,7 +2286,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	if (!param || !*param) goto not_found;
 	svga_driver_param=stracpy(param);
 	for (j=0;(size_t)j<sizeof(modes)/sizeof(*modes);j++)
-		if (!strcasecmp(modes[j].name,param)) goto found;
+		if (!strcasecmp(cast_const_char modes[j].name,cast_const_char param)) goto found;
 	j = 1;
 	not_found:
 	{
@@ -2312,24 +2294,23 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 		int l = 0;
 		int f = 0;
 		if (j) {
-			add_to_str(&m, &l, "Video mode ");
+			add_to_str(&m, &l, cast_uchar "Video mode ");
 			add_to_str(&m, &l, param);
 			add_to_str(&m, &l,
-				j == 1 ? " not supported by svgalib" :
-				j == 2 ? " not supported by your video card" :
-				" could not be set");
-			add_to_str(&m, &l, ".\n");
-		} else add_to_str(&m, &l, "There is no default video mode.\n");
+				j == 1 ? cast_uchar " not supported by svgalib" :
+				j == 2 ? cast_uchar " not supported by your video card" :
+				cast_uchar " could not be set");
+			add_to_str(&m, &l, cast_uchar ".\n");
+		} else add_to_str(&m, &l, cast_uchar "There is no default video mode.\n");
 		for (j=0;(size_t)j<sizeof(modes)/sizeof(*modes);j++) if (vga_hasmode(modes[j].number)) {
-			if (f) add_to_str(&m, &l, ", ");
-			else f = 1, add_to_str(&m, &l, "The following modes are supported:\n");
-			add_to_str(&m, &l, modes[j].name);
+			if (f) add_to_str(&m, &l, cast_uchar ", ");
+			else f = 1, add_to_str(&m, &l, cast_uchar "The following modes are supported:\n");
+			add_to_str(&m, &l, cast_uchar modes[j].name);
 		}
-		if (f) add_to_str(&m, &l, "\nUse -mode switch to set video mode.\n");
-		else add_to_str(&m, &l, "There are no supported video modes. Links can't run on svgalib.\n");
+		if (f) add_to_str(&m, &l, cast_uchar "\nUse -mode switch to set video mode.\n");
+		else add_to_str(&m, &l, cast_uchar "There are no supported video modes. Links can't run on svgalib.\n");
 		if(svga_driver_param)mem_free(svga_driver_param),svga_driver_param=NULL;
 		return m;
-		
 	}
 	found:
 	if (!vga_hasmode(modes[j].number)) {
@@ -2339,7 +2320,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	if (init_virtual_devices(&svga_driver, NUMBER_OF_DEVICES))
 	{
 		if(svga_driver_param)mem_free(svga_driver_param),svga_driver_param=NULL;
-		return stracpy("Allocation of virtual devices failed.\n");
+		return stracpy(cast_uchar "Allocation of virtual devices failed.\n");
 	}
 	if ((vga_getmousetype()&MOUSE_TYPE_MASK)==MOUSE_NONE)
 		{
@@ -2354,7 +2335,10 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	svgalib_timer_id=install_timer(100,vtswitch_handler,NULL);
 	if (vga_runinbackground_version()>=1) vga_runinbackground(1);
 #endif
+	vga_setmode_retry:
 	if (vga_setmode(modes[j].number) < 0) {
+		if (out_of_memory(0, NULL, 0))
+			goto vga_setmode_retry;
 #ifndef __SPAD__
 		kill_timer(svgalib_timer_id);
 #endif
@@ -2365,7 +2349,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	}
 	vga_mode=modes[j].number;
 	setup_mode(modes[j].number);
-	svgalib_kbd = handle_svgalib_keyboard((void (*)(void *, unsigned char *, int))svgalib_key_in);
+	svgalib_kbd = handle_svgalib_keyboard(svgalib_key_in);
 
 	if (mouse_works){
 		if ((unsigned)arrow_area > (unsigned)MAXINT / fb_pixelsize) overalloc();
@@ -2380,13 +2364,14 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 		background_x=mouse_x=xsize>>1;
 		background_y=mouse_y=ysize>>1;
 		show_mouse();
-		mouse_seteventhandler(mouse_event_handler);
+		mouse_seteventhandler((void *)mouse_event_handler);
 	}else{
 		global_mouse_hidden=1;
 		/* To ensure hide_mouse and show_mouse will do nothing */
 	}
-	signal(SIGPIPE, SIG_IGN);
-	signal(SIGTSTP, SIG_IGN);
+	ignore_signals();
+	errno = 0;
+	while (signal(SIGTSTP, SIG_IGN) == SIG_ERR && errno == EINTR) errno = 0;
 	install_signal_handler(SIGINT, (void (*)(void *))svga_ctrl_c, svgalib_kbd, 0);
 	return NULL;
 }
@@ -2399,7 +2384,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 static int svga_get_filled_bitmap(struct bitmap *dest, long color)
 {
 	int n;
-	
+
 	if (dest->x && (unsigned)dest->x * (unsigned)dest->y / (unsigned)dest->x != (unsigned)dest->y) overalloc();
 	if ((unsigned)dest->x * (unsigned)dest->y > MAXINT / fb_pixelsize) overalloc();
 	n=dest->x*dest->y*fb_pixelsize;
@@ -2441,10 +2426,11 @@ static int vga_block(struct graphics_device *dev)
 			/* mouse_close(); This is not necessary as it is
 				handled by vga_setmode(TEXT). */
 		}
+		vga_setmode_retry:
 		if (vga_setmode(TEXT) < 0) {
-			error("ERROR: vga_setmode failed");
-			fatal_tty_exit();
-			exit(RET_FATAL);
+			if (out_of_memory(0, NULL, 0))
+				goto vga_setmode_retry;
+			fatal_exit("ERROR: vga_setmode failed");
 		}
 	}
 	flags|=2;
@@ -2462,15 +2448,16 @@ static int vga_unblock(struct graphics_device *dev)
 	flags&=~2;
 	if (!flags) current_virtual_device=backup_virtual_device;
 	vga_setmousesupport(1);
+	vga_setmode_retry:
 	if (vga_setmode(vga_mode) < 0) {
-		error("ERROR: vga_setmode failed");
-		fatal_tty_exit();
-		exit(RET_FATAL);
+		if (out_of_memory(0, NULL, 0))
+			goto vga_setmode_retry;
+		fatal_exit("ERROR: vga_setmode failed");
 	}
 	setup_mode(vga_mode);
 	if (mouse_works){
 		show_mouse();
-		mouse_seteventhandler(mouse_event_handler);
+		mouse_seteventhandler((void *)mouse_event_handler);
 	}
 	svgalib_unblock_itrm(svgalib_kbd);
 	if (current_virtual_device) current_virtual_device->redraw_handler(current_virtual_device
@@ -2480,7 +2467,7 @@ static int vga_unblock(struct graphics_device *dev)
 
 static void *svga_prepare_strip(struct bitmap *bmp, int top, int lines)
 {
-	return ((char *)bmp->data)+bmp->skip*top;
+	return ((unsigned char *)bmp->data)+bmp->skip*top;
 }
 
 
@@ -2489,11 +2476,7 @@ static void svga_commit_strip(struct bitmap *bmp, int top, int lines)
 	return;
 }
 
-/* This is a nasty hack */
-#undef select
-
-int vga_select(int  n,  fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
-			      struct timeval *timeout)
+int vga_select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
 {
 	int retval,i;
 
@@ -2535,12 +2518,15 @@ int vga_select(int  n,  fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
 }
 
 struct graphics_driver svga_driver={
-	"svgalib",
+	cast_uchar "svgalib",
 	svga_init_driver,
 	init_virtual_device,
 	shutdown_virtual_device,
 	svga_shutdown_driver,
+	dummy_emergency_shutdown,
 	svga_get_driver_param,
+	NULL,
+	NULL,
 	svga_get_empty_bitmap,
 	/*svga_get_filled_bitmap,*/
 	svga_register_bitmap,
@@ -2555,7 +2541,7 @@ struct graphics_driver svga_driver={
 	NULL, /* draw_vline */
 	NULL, /* hscroll */
 	NULL, /* vscroll */
-	svga_set_clip_area,
+	generic_set_clip_area,
 	vga_block, /* block */
 	vga_unblock, /* unblock */
 	NULL, /* set_title */
